module rif_module
  type rifgridstruct
     character(len=24) :: hdate
     character(len=9) :: field
     character(len=25) :: units
     character(len=46) :: desc
     real :: layer1   ! Top of a layer definition
     real :: layer2   ! Bottom of a layer definition
     real :: xfcst
     real :: xlvl
     integer :: iproj
     integer :: nx
     integer :: ny
     real :: startlat, startlon ! Used for all IPROJ
     real :: deltalat, deltalon ! Used for IPROJ = 0
     real :: dxkm, dykm             ! Used for IPROJ = 1,3,5
     real :: xlonc              ! Used for IPROJ = 3,5
     real :: truelat1           ! Used for IPROJ = 1,3,5
     real :: truelat2           ! Used for IPROJ = 3
     real, pointer, dimension(:,:) :: data  ! The data array

     real :: gx1, gy1, gkappa
  end type rifgridstruct

contains

  subroutine rif_lltoxy(lat, lon, x, y, rif)
    implicit none
    real, intent(in) :: lat, lon
    real, intent(out) :: x, y
    type(rifgridstruct), intent(in) :: rif
    real, parameter :: grrth = 6370.949
    real :: r
    real, external :: sind
    real, external :: cosd
    real, external :: tand

    if (rif%iproj == 0) then  ! Cylindrical Equidistant grid
       x = 1. + ((lon - rif%startlon) / rif%deltalon)
       y = 1. + ((lat - rif%startlat) / rif%deltalat)
    else if (rif%iproj == 3) then ! Lambert Conformal
       r = grrth/(rif%dxkm*rif%gkappa)*sind(90.-rif%truelat1) * &
            (tand(45.-lat/2.)/tand(45.-rif%truelat1/2.)) ** rif%gkappa
       x =  r*sind(rif%gkappa*(lon-rif%xlonc)) - rif%gx1 + 1.
       y = -r*cosd(rif%gkappa*(lon-rif%xlonc)) - rif%gy1 + 1.
       
    else if (rif%iproj == 5) then ! Polar Stereographic
       r = grrth/(rif%dxkm*rif%gkappa) * tand((90.-lat)/2.) * &
            (1.+sind(rif%truelat1))
       x = ( r * sind(lon-rif%xlonc)) - rif%gx1 + 1.
       y = (-r * cosd(lon-rif%xlonc)) - rif%gy1 + 1.
       
    else
       write(*,'("Unrecoginzed projection: ", I2)') rif%iproj
       write(*,'("ERROR EXIT in RIF_LLTOXY",/)')
       stop
    endif

  end subroutine rif_lltoxy

  subroutine rlccone (fsplat,ssplat,sign,confac)
    implicit none
    real, intent(in) :: fsplat,ssplat
    integer, intent(in) :: sign
    real, intent(out) :: confac
    real, external :: sind
    real, external :: cosd
    real, external :: tand
    if (abs(fsplat-ssplat).lt.1.E-3) then
       confac = sind(fsplat)
    else
       confac = log10(cosd(fsplat))-log10(cosd(ssplat))
       confac = confac/(log10(tand(45.-float(sign)*fsplat/2.))- &
            log10(tand(45.-float(sign)*ssplat/2.)))
    endif
  end subroutine rlccone

  subroutine rif_xytoll(x, y, lat, lon, rif)
    implicit none
    real, intent(in) :: x, y
    real, intent(out) :: lat, lon
    type(rifgridstruct), intent(in) :: rif

    real, parameter :: grrth = 6370.949

    real :: r, y1
    integer :: iscan, jscan
    real, parameter :: pi = 3.14159265358979
    real, parameter :: degran = pi/180.
    real, parameter :: raddeg = 180./pi
    real :: gt
    real :: dtrue1, dtrue2, dlat1
    real, external :: sind
    real, external :: cosd

    if (rif%iproj == 0) then  ! Cylindrical Equidistant grid

       lat = rif%startlat + rif%deltalat*(y-1.)
       lon = rif%startlon + rif%deltalon*(x-1.)

    elseif (rif%iproj == 1) then ! Mercator grid

       gt = grrth * cosd(rif%truelat1)
       lon = rif%startlon + (rif%dxkm*(x-1.)/gt)*raddeg
       y1 = gt*alog((1.+sind(rif%startlat))/cosd(rif%startlat))/rif%dykm
       lat = 90. - 2. * atan(exp(-rif%dykm*(y+y1-1.)/gt))*raddeg

    elseif (rif%iproj == 5) then  ! Polar Stereographic grid
       dtrue1 = rif%truelat1*degran
       dlat1 = rif%startlat*degran

       r = sqrt((x-1.+rif%gx1)**2 + (y-1+rif%gy1)**2)
       lat = 90. - 2.*atan2((r*rif%dxkm),(grrth*(1.+sin(dtrue1))))*raddeg
       lon = atan2((x-1.+rif%gx1),-(y-1.+rif%gy1))*raddeg + rif%xlonc
       if (lon > 180) lon = lon - 360.

    elseif (rif%iproj == 3) then  ! Lambert-conformal grid
       dtrue1 = rif%truelat1*degran
       dtrue2 = rif%truelat2*degran
       dlat1  = rif%startlat*degran

       r = sqrt((x-1.+rif%gx1)**2 + (y-1+rif%gy1)**2)
       lat = 90. - 2.*atan(tan((45.*degran)-dtrue1/2.)* &
            ((r*rif%gkappa*rif%dxkm)/(grrth*sin((90.*degran)-dtrue1)))**(1./rif%gkappa))&
            *raddeg
       lon = atan2((x-1.+rif%gx1),-(y-1.+rif%gy1))*raddeg/rif%gkappa + rif%xlonc

    else

       write(*,'("Unrecoginzed projection: ", A)') rif%iproj
       write(*,'("Abort in RIF_XYTOLL",/)')
       stop

    endif
  end subroutine rif_xytoll

  subroutine read_RIF_data(iunit, rif, ierr)
    implicit none
    integer :: iunit
    type(rifgridstruct) :: rif
    integer :: ierr
!==================

    ierr = 0

    allocate(rif%data(rif%nx, rif%ny))

    read(iunit, iostat=ierr) rif%data
    if (ierr /= 0) then
       stop "READ_RIF:  Read record 4 (array)"
    endif

  end subroutine read_RIF_data

  subroutine skip_RIF_data(iunit, rif, ierr)
    implicit none
    integer :: iunit
    type(rifgridstruct) :: rif
    integer :: ierr
    real :: xdum
!==================

    ierr = 0

    read(iunit, iostat=ierr) xdum
    if (ierr /= 0) then
       stop "SKIP_RIF"
    endif

  end subroutine skip_RIF_data

  subroutine read_RIF_header(fname, iunit, rif, ierr)
    implicit none
    character(len=*) :: fname
    integer :: iunit
    type(rifgridstruct) :: rif
    integer :: ierr

    logical :: opened
    integer :: i, ifv
    real :: rb
    real, parameter :: grrth = 6370.949
    real, external :: sind, cosd, tand
!=============================================================================

    ierr = 0

    inquire(file=trim(fname), opened=opened)

    if ( .not. opened) then
! The file is not currently opened.  Find a free unit number and open it
! with that unit number.
       FINDUNIT : do i=10, 255
          inquire(unit=i, opened=opened)
          if (.not. opened) then
             iunit = i
             exit FINDUNIT
          endif
       enddo FINDUNIT

       open(iunit, file=trim(fname), form='unformatted', status='old', &
            action='read', iostat=ierr)
       if (ierr /= 0) then
          print*, 'Filename = ', trim(fname)
          stop "READ_RIF:  OPEN"
       endif
    else
! The file is already opened.  Find the unit number.
       inquire(file=trim(fname), number=iunit)
       if (iunit == -1) then
          print*, 'File unit number not found.  This is a problem.'
          print*, 'Filename = ', trim(fname)
          stop "READ_RIF:  INQUIRE"
       endif
    endif

    read(iunit, iostat=ierr) ifv
    if (ierr < 0) then
       close(iunit)
       ierr = 1
       return
    endif
    if (ierr /= 0) stop "READ_RIF:  Read record 1"
    if (ifv /= 3) stop "READ_RIF:  IFV Value"

    read(iunit, iostat=ierr) rif%hdate, rif%xfcst, &
         rif%field, &
         rif%units, rif%desc, rif%xlvl, &
         rif%nx, rif%ny, rif%iproj
    if (ierr /= 0) stop "READ_RIF:  Read record 2"

    if (rif%iproj == 0) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%deltalat, rif%deltalon
       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=0)"
    else if (rif%iproj == 1) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%dxkm, rif%dykm, rif%truelat1
       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=1)"
    else if (rif%iproj == 3) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%dxkm, rif%dykm, rif%xlonc, rif%truelat1, &
            rif%truelat2

       call rlccone(rif%truelat1, rif%truelat2, 1, &
            rif%gkappa)
       rb = grrth/(rif%dxkm*rif%gkappa)*&
            sind(90.-rif%truelat1) * &
            (tand(45.-rif%startlat/2.)/&
            tand(45.-rif%truelat1/2.)) ** rif%gkappa
       rif%gx1 =  rb*sind(rif%gkappa*&
            (rif%startlon-rif%xlonc))
       rif%gy1 = -rb*cosd(rif%gkappa*&
            (rif%startlon-rif%xlonc))


       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=3)"

    else if (rif%iproj == 5) then

       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%dxkm, rif%dykm, rif%xlonc, rif%truelat1
       rif%gkappa = 1.0
       rb= grrth/rif%dxkm * tand((90.-rif%startlat)/2.) * &
            (1.+sind(rif%truelat1))
       rif%gx1 = rb * sind(rif%startlon-rif%xlonc)
       rif%gy1 = -rb * cosd(rif%startlon-rif%xlonc)

       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=5)"

    else

       stop "READ_RIF:  IPROJ"

    endif
    
  end subroutine read_RIF_header

  subroutine read_RIF(fname, rif, ierr)
    implicit none
    character(len=*) :: fname
    type(rifgridstruct) :: rif
    logical :: opened
    integer :: i, iunit
    integer :: ifv
    integer :: ierr
    real :: rb
    real, parameter :: grrth = 6370.949
    real, external :: sind, cosd, tand

    ierr = 0

    inquire(file=trim(fname), opened=opened)

    if ( .not. opened) then
! The file is not currently opened.  Find a free unit number and open it
! with that unit number.
       FINDUNIT : do i=10, 255
          inquire(unit=i, opened=opened)
          if (.not. opened) then
             iunit = i
             exit FINDUNIT
          endif
       enddo FINDUNIT

       open(iunit, file=trim(fname), form='unformatted', status='old', &
            action='read', iostat=ierr)
       if (ierr /= 0) then
          print*, 'Filename = ', trim(fname)
          stop "READ_RIF:  OPEN"
       endif
    else
! The file is already opened.  Find the unit number.
       inquire(file=trim(fname), number=iunit)
       if (iunit == -1) then
          print*, 'File unit number not found.  This is a problem.'
          print*, 'Filename = ', trim(fname)
          stop "READ_RIF:  INQUIRE"
       endif
    endif

    read(iunit, iostat=ierr) ifv
    if (ierr < 0) then
       close(iunit)
       ierr = 1
       return
    endif
    if (ierr /= 0) stop "READ_RIF:  Read record 1"
    if (ifv /= 3) stop "READ_RIF:  IFV Value"

    read(iunit, iostat=ierr) rif%hdate, rif%xfcst, &
         rif%field, &
         rif%units, rif%desc, rif%xlvl, &
         rif%nx, rif%ny, rif%iproj
    if (ierr /= 0) stop "READ_RIF:  Read record 2"

    allocate(rif%data(rif%nx, rif%ny))

    if (rif%iproj == 0) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%deltalat, rif%deltalon
       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=0)"
    else if (rif%iproj == 1) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%dxkm, rif%dykm, rif%truelat1
       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=1)"
    else if (rif%iproj == 3) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%dxkm, rif%dykm, rif%xlonc, rif%truelat1, &
            rif%truelat2

       call rlccone(rif%truelat1, rif%truelat2, 1, &
            rif%gkappa)
       rb = grrth/(rif%dxkm*rif%gkappa)*&
            sind(90.-rif%truelat1) * &
            (tand(45.-rif%startlat/2.)/&
            tand(45.-rif%truelat1/2.)) ** rif%gkappa
       rif%gx1 =  rb*sind(rif%gkappa*&
            (rif%startlon-rif%xlonc))
       rif%gy1 = -rb*cosd(rif%gkappa*&
            (rif%startlon-rif%xlonc))


       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=3)"
    else if (rif%iproj == 5) then
       read(iunit, iostat=ierr) rif%startlat, rif%startlon, &
            rif%dxkm, rif%dykm, rif%xlonc, rif%truelat1
       rif%gkappa = 1.0
       rb= grrth/rif%dxkm * tand((90.-rif%startlat)/2.) * &
            (1.+sind(rif%truelat1))
       rif%gx1 = rb * sind(rif%startlon-rif%xlonc)
       rif%gy1 = -rb * cosd(rif%startlon-rif%xlonc)

       if (ierr /= 0) stop "READ_RIF:  Read record 3 (IPROJ=5)"
    else
       stop "READ_RIF:  IPROJ"
    endif

    read(iunit, iostat=ierr) rif%data
    if (ierr /= 0) then
       stop "READ_RIF:  Read record 4 (array)"
    endif

  end subroutine read_RIF

  subroutine write_RIF(fname, rif, ierr)
! Write a single RIF slab to output file.
    implicit none
    character(len=*) :: fname
    type(rifgridstruct) :: rif
    logical :: opened
    integer :: i, iunit
    integer, parameter :: ifv = 3
    integer :: ierr
    real :: rb
    real, parameter :: grrth = 6370.949

    ierr = 0

    inquire(file=trim(fname), opened=opened)

    if ( .not. opened) then

       FINDUNIT : do i=10, 255
          inquire(unit=i, opened=opened)
          if (.not. opened) then
             iunit = i
             exit FINDUNIT
          endif
       enddo FINDUNIT

       open(iunit, file=trim(fname), form='unformatted', status='unknown', &
            action='write', iostat=ierr)
       if (ierr /= 0) then
          print*, 'Filename = ', trim(fname)
          stop "WRITE_RIF:  OPEN"
       endif

    else
       inquire(file=trim(fname), number=iunit)
    endif

    write(iunit) ifv

    write(iunit) rif%hdate, rif%xfcst, rif%field, &
         rif%units, rif%desc, rif%xlvl, &
         rif%nx, rif%ny, rif%iproj

    if (rif%iproj == 0) then
       write(iunit) rif%startlat, rif%startlon, rif%deltalat, rif%deltalon
    else if (rif%iproj == 1) then
       write(iunit) rif%startlat, rif%startlon, rif%dxkm, rif%dykm, rif%truelat1
    else if (rif%iproj == 3) then
       write(iunit) rif%startlat, rif%startlon, rif%dxkm, rif%dykm, &
            rif%xlonc, rif%truelat1, rif%truelat2
    else if (rif%iproj == 5) then
       write(iunit) rif%startlat, rif%startlon, rif%dxkm, rif%dykm, &
            rif%xlonc, rif%truelat1
    else
       print*, 'iproj = ', rif%iproj
       stop "WRITE_RIF:  IPROJ"
    endif

    write(iunit) rif%data
  end subroutine write_RIF

end module rif_module

