subroutine read_wrfinput_file(flnm, wrfinput, ierr)
  ! 
  ! Reads the wrfinput file identified by <flnm>, and fills out 
  ! the <wrfinput> structure with the appropriate values.
  !
  use module_wrfinputfile
  implicit none

  character(len=*),     intent(in)  :: flnm
  type (wrfinput_type), intent(out) :: wrfinput
  integer, intent(out)              :: ierr
  integer                           :: ncid
  integer                           :: iret
  integer                           :: dimid

! Open the NetCDF file.
  iret = nf90_open(flnm, NF90_NOWRITE, ncid)
  call error_handler(iret, "Problem with wrfinput file: "//flnm)

! Find out about dimensions in the file.

  iret = nf90_inq_dimid(ncid, "west_east", dimid)
  call error_handler(iret, "STOP:  Problem finding west_east dimension")
  iret = nf90_inquire_dimension(ncid, dimid, len=wrfinput%idim)
  call error_handler(iret, "STOP:  Problem finding west_east dimension")

  iret = nf90_inq_dimid(ncid, "south_north", dimid)
  call error_handler(iret, "STOP:  Problem finding south_north dimension")
  iret = nf90_inquire_dimension(ncid, dimid, len=wrfinput%jdim)
  call error_handler(iret, "STOP:  Problem finding south_north dimension")

  ! Grid id.  Check for the string "GRID_ID" or the string "grid_id"
  iret = nf90_get_att(ncid, NF90_GLOBAL,"GRID_ID", wrfinput%grid_id)
  if (iret /= 0) then
     iret = nf90_get_att(ncid, NF90_GLOBAL,"grid_id", wrfinput%grid_id)
     call error_handler(iret, "STOP:  nf90_get_att:  'grid_id' or 'GRID_ID' not found.")
  endif

  ! Projection information:

  iret = nf90_get_att(ncid, NF90_GLOBAL,"MAP_PROJ", wrfinput%map_proj)
  call error_handler(iret, "STOP:  nf90_get_att:  map projection problem")
      
  iret = nf90_get_att(ncid, NF90_GLOBAL, "STAND_LON", wrfinput%LoV)
  call error_handler(iret, "STOP:  nf90_get_att:  STAND_LON")

  iret = nf90_get_att(ncid, NF90_GLOBAL, "TRUELAT1" , wrfinput%Latin1)
  call error_handler(iret, "STOP:  nf90_get_att:  TRUELAT1")

  iret = nf90_get_att(ncid, NF90_GLOBAL, "TRUELAT2" , wrfinput%Latin2)
  call error_handler(iret, "STOP:  nf90_get_att:  TRUELAT2")

  iret = nf90_get_att(ncid, NF90_GLOBAL, "DX" , wrfinput%Dx)
  call error_handler(iret, "STOP:  nf90_get_att:  DX")

  iret = nf90_get_att(ncid, NF90_GLOBAL, "DY" , wrfinput%Dy)
  call error_handler(iret, "STOP:  nf90_get_att:  DY")

  iret = nf90_get_att(ncid, NF90_GLOBAL, "ISWATER" , wrfinput%iswater)
  call error_handler(iret, "STOP:  nf90_get_att:  ISWATER")

  iret = nf90_get_att(ncid, NF90_GLOBAL, "MMINLU" , wrfinput%landuse_dataset)
  call error_handler(iret, "STOP:  nf90_get_att:  MMINLU")

  ! Get Latitude
  allocate(wrfinput%lat(wrfinput%idim, wrfinput%jdim))
  call get_2d("XLAT", ncid, wrfinput%lat, wrfinput%idim, wrfinput%jdim)
  wrfinput%La1 = wrfinput%lat(1,1)
  wrfinput%La2 = wrfinput%lat(wrfinput%idim,wrfinput%jdim)

  ! Get Longitude
  allocate(wrfinput%lon(wrfinput%idim, wrfinput%jdim))
  call get_2d("XLONG", ncid, wrfinput%lon, wrfinput%idim, wrfinput%jdim)
  wrfinput%Lo1 = wrfinput%lon(1,1)
  wrfinput%Lo2 = wrfinput%lon(wrfinput%idim,wrfinput%jdim)

  ! Get Terrain
  allocate(wrfinput%ter(wrfinput%idim, wrfinput%jdim))
  call get_2d("HGT", ncid, wrfinput%ter, wrfinput%idim, wrfinput%jdim)

  ! Get Land Use categories 
  allocate(wrfinput%use(wrfinput%idim, wrfinput%jdim))
  call get_2d("IVGTYP", ncid, wrfinput%use, wrfinput%idim, wrfinput%jdim)

!KWM  ! Get deep soil Temperature field
!KWM  allocate(wrfinput%tmn(wrfinput%idim, wrfinput%jdim))
!KWM  call get_2d("TMN", ncid, wrfinput%tmn, wrfinput%idim, wrfinput%jdim)

  ! Close the file and get out of here

  iret = nf90_close(ncid)
  call error_handler(iret, "STOP:  Problem closing file??")

  ierr = 0
  print*, "Done with subroutine read_wrfinput_file"

end subroutine read_wrfinput_file

subroutine get_2d(name, ncid, array, idim, jdim)
  !
  ! From the NetCDF unit <ncid>, read the variable named <name> with  
  ! dimensions <idim> and <jdim>, filling the pre-dimensioned array <array>
  !
  use module_wrfinputfile
  implicit none
  character(len=*),           intent(in)  :: name
  integer,                    intent(in)  :: ncid
  integer,                    intent(in)  :: idim
  integer,                    intent(in)  :: jdim
  real, dimension(idim,jdim), intent(out) :: array
  ! Local:
  integer                                 :: ierr
  integer                                 :: varid

  ierr = nf90_inq_varid(ncid,  name,  varid)
  ! If the variable is "XLAT", and "XLAT" is not found, look for "XLAT_M"
  ! If the variable is "XLAT_M", and "XLAT_M" is not found, look for "XLAT"
  ! If the variable is "XLONG", and "XLONG" is not found, look for "XLONG_M"
  ! If the variable is "XLONG_M", and "XLONG_M" is not found, look for "XLONG"
  if (name == "XLAT") then
     if (ierr /= 0) then
        ierr = nf90_inq_varid(ncid,  "XLAT_M",  varid)
     endif
  else if (name == "XLAT_M") then
     if (ierr /= 0) then
        ierr = nf90_inq_varid(ncid,  "XLAT",  varid)
     endif
  else  if (name == "XLONG") then
     if (ierr /= 0) then
        ierr = nf90_inq_varid(ncid,  "XLONG_M",  varid)
     endif
  else if (name == "XLONG_M") then
     if (ierr /= 0) then
        ierr = nf90_inq_varid(ncid,  "XLONG",  varid)
     endif
  endif
  call error_handler(ierr, "STOP:  READ_WRFINPUT: Problem finding variable '"//trim(name)//"' in wrfinput file.")


  ierr = nf90_get_var(ncid, varid, array)
  call error_handler(ierr, "STOP:  READ_WRFINPUT:  Problem retrieving variable '"//trim(name)//"' from wrfinput file.")

end subroutine get_2d
