program Noah_hrldas_driver

  USE module_hrldas_netcdf_io
  USE module_Noahlsm_utility
  USE module_sf_noahlsm
  USE module_sf_urban
  USE module_date_utilities

  IMPLICIT NONE

  character(len=9), parameter :: version = "v20080425"
  integer :: LDASIN_VERSION

  ! Dummy parameterized dimension for maximum number of soil levels allowed
  INTEGER, PARAMETER :: NSOLDX = 100

  ! ZSOIL, set through the namelist, is the BOTTOM of each soil layer (m)
  REAL, DIMENSION(NSOLDX) :: ZSOIL

  REAL :: T1V,TH2V

!-----------------------------------------------------------------
!  Dimensions from the input file:
!-----------------------------------------------------------------

!     IX: x-direction points, usually along latitude
!     JX: y-direction points, usually along longitude
  INTEGER :: IX
  INTEGER :: JX

!
! Attributes from LDASIN input file (or HRLDAS_CONSTANTS_FILE, as the case may be)
  REAL              :: DX
  REAL              :: DY
  REAL              :: TRUELAT1
  REAL              :: TRUELAT2
  REAL              :: CEN_LON
  INTEGER           :: MAPPROJ
  REAL              :: LAT1
  REAL              :: LON1
  INTEGER           :: ISWATER

!-----------------------------------------------------------------
!  DECLARE VARIABLES FOR GRIDDED SIMULATION
!-----------------------------------------------------------------

! the following parameters are read from a namelist file

! setup model configuration
!   KHOUR: number of hours for the simulation
  INTEGER :: KHOUR
  INTEGER :: KDAY
! INTEGER :: NSOIL: number of soil layers for the LSM integration
! REAL    :: DT:    time-step in seconds
! REAL    :: ZLVL:  Level (m) at which atmospheric fields are taken to be valid

! Unaccounted for variables when compiling w/ implicit none F90

!KWM  REAL   :: VEGMIN,VEGMAX
  REAL   :: SNOFAC,ALBED
  REAL   :: RHO,CHKFF
!KWM  REAL   :: T14,AET,FUP,S
!KWM  REAL   :: F, H
!KWM  REAL   :: SNOFLX, SNMAX, SNOEVP

  INTEGER   :: i,j,k,ns,ierr
  INTEGER   :: NTIME
  REAL      :: FRACTION

! Gridded fields

  INTEGER, allocatable, DIMENSION(:,:)   :: VEGTYP,SOLTYP
  REAL,    allocatable, DIMENSION(:,:)   :: TERRAIN, LATITUDE, LONGITUDE
  REAL,    allocatable, DIMENSION(:,:)   :: T2,XLONG,U,V,PRES,SHORT,PRCP1
  REAL,    allocatable, DIMENSION(:,:)   :: FPAR, ZNT, LAI, TBOT_2D, FPAR_SAVE, LAI_SAVE
  REAL,    allocatable, DIMENSION(:,:)   :: CMC,SNODEP,WEASD,T1
  REAL,    allocatable, DIMENSION(:,:)   :: ETPX,ETAX,GRDFLX,CHX,RUNOFF1X,RUNOFF2X
  REAL,    allocatable, DIMENSION(:,:)   :: ETAKIN, QFX
  REAL,    allocatable, DIMENSION(:,:)   :: RUNOFF3X,EDIRX,ECX,ETTX,SNMAXX,RCX,HX
  REAL,    allocatable, DIMENSION(:,:)   :: Q2X,SFCSPDX,ALBEDX,SMCMAX1,SMCWLT1
  REAL,    allocatable, DIMENSION(:,:)   :: SMCREF1,FX,RESX
  REAL,    allocatable, DIMENSION(:,:)   :: SNOFLXX,SNOEVPX,ACSNOM,ESNOW2D,ACRAIN
  REAL,    allocatable, DIMENSION(:,:)   :: DRIP2D, DEWFALL, SOILMX, EMISS, XLAI2D
!KWM  REAL,    allocatable, DIMENSION(:,:)   :: ETPNDX, SFCHEAD, INFXS1, PDDUM2, PCPDRP, SFCWATR2

  REAL,    allocatable, DIMENSION(:,:,:) :: SMC,STC,SH2OX
  REAL,    allocatable, DIMENSION(:,:,:) :: ZSOILX 

  ! Min/Max values from the 12-monthly Green Vegetation Fraction:
  REAL,    allocatable, DIMENSION(:,:)   :: GVFMIN, GVFMAX



  real :: albbrd
!---------------------------------------------------------------------
! NEW VARIABLES ADDED DURING NOAH F90 UPGRADE CALL to SFLX
!---------------------------------------------------------------------
  REAL :: EMISSI
  REAL :: EMBRD


! INTENT(IN) to SFLX:
  REAL    ::            FFROZP      ! Fraction of precipitation that is frozen
  REAL    ::            SOLNET      ! Net solar radiaton (W m-2)
  INTEGER, PARAMETER :: ICE = 0     ! Sea-ice flag  (=1: sea-ice, =0: land)
  REAL    :: DT       ! Timestep (s) (DT should not exceed 3600 S, recommend 1800 s or less)
  REAL    :: ZLVL     ! Height (m) above ground of atmospheric forcing variables
  INTEGER :: NSOIL    ! Number of soil layers
  REAL, ALLOCATABLE, DIMENSION(:) :: SLDPTH ! The THICKNESS (m) of each soil layer 
  CHARACTER(LEN=4) :: LLANDUSE ! (=USGS, using USGS landuse classification)
  CHARACTER(LEN=4) :: LSOIL    ! (=STAS, using FAO/STATSGO soil texture classification)

  REAL :: LWDN      ! LW downward radiation (W m-2; positive, not net longwave)
  REAL :: SOLDN     ! Solar downward radiation (W m-2; positive, not net solar)
  REAL :: SFCPRS    ! Pressure (Pa) at height ZLVL m above ground
  REAL :: PRCP      ! Precip rate (kg m-2 s-1) (NOTE: this is a rate)
  REAL :: SFCTMP    ! Air temperature (K) at height ZLVL m above ground
  REAL :: Q2        ! Specific Humidity (kg kg-1) at height ZLVL m above ground 
  REAL :: SFCSPD    ! Wind speed (m s-1) at height ZLVL m above ground
  REAL :: PRCPRAIN  ! Liquid-precipitation rate (KG M-2 S-1) (not used)
  REAL :: TH2       ! Air potential temperature (K) at height ZLVL m above ground
  REAL :: Q2SAT     ! Saturation specific humidity (kg kg-1) at height ZLVL m above ground
  REAL :: DQSDT2    ! Slope of saturation specific humidity curve  (kg kg-1 K-1) at T=SFCTMP
  INTEGER :: VEGTYPX! Vegetation type (integer index)
  INTEGER :: SOILTYP! Soil type (integer index)
  INTEGER ::SLOPETYP! Class of sfc slope (integer index)
  REAL :: SHDMIN    ! Minimum areal fractional coverage of green vegetation
  !         (dimensionless fraction 0.0-1.0) <= SHDFAC
  REAL :: SHDMAX    !UNDOCUMENTED IN SFLX
  REAL :: TBOT      ! Bottom soil temperature (local yearly-mean of surfacc air temperature)


! INTENT(INOUT) to/from SFLX:
  REAL :: SNCOVR  ! Fractional snow cover (dimensionless fraction, 0.0-1.0)
  REAL :: COSZ    ! Solar zenith angle
  REAL :: DECLIN  ! Solar declension (rad)
  REAL :: SOLARDIRECT ! Direct component of downward solar radiation (W m-2) (not used)
  REAL :: Z0      ! Time varying roughness length (m) as function of snow depth
  REAL :: CMCX    ! Canopy moisture content (m)
  REAL :: T1X     ! Ground/Canopy/Snowpack effective skin temperature (K)
  REAL, allocatable, DIMENSION(:)  :: STC1    ! Soil temp (K)                            
  REAL, allocatable, DIMENSION(:)  :: SMC1    ! Total soil moisture content (volumetric fraction)
  REAL, allocatable, DIMENSION(:)  :: SH2O    ! Unfrozen soil moisture content (volumetric fraction)
  !          NOTE: Frozen soil moisture = SMC - SH2O
  REAL :: SNOWH   ! Actual snow depth (m)                                      
  REAL :: SNEQV   ! Liquid water-equivalent snow depth (m)                     
  !       NOTE: snow density = SNEQV/SNOWH                         
  REAL :: ALBEDO  ! Surface albedo including snow effect (unitless fraction)   
  !      =snow-free albedo (ALB) when SNEQV=0, or                 
  !      =FCT(MSNOALB,ALB,VEGTYP,SHDFAC,SHDMIN) when SNEQV>0      
  REAL :: CH      ! Surface exchange coefficient for heat and moisture         
  !   (m s-1); NOTE: CH is technically a conductance since     
  !   it has been multiplied by wind speed.                    
  REAL :: CM      ! Surface exchange coefficient for momentum (m s-1); NOTE:   
  !   CM is technically a conductance since it has been        
  !   multiplied by wind speed.                                

! INTENT(OUT) from SFLX:
  REAL :: ETT     ! Total plant transpiration (W m-2)
  REAL :: ETA     ! Actual latent heat flux (W m-2: negative, if up from surface)     
!KWM ???  Is the note about the sign of ETA right ???? !KWM
  REAL :: SHEAT   ! Sensible heat flux (W m-2: negative, if upward from surface)
  REAL :: ETA_KINEMATIC ! Actual latent heat flux (kg m/s)
  REAL :: FDOWN   ! Radiation forcing at the surface (W m-2) = SOLDN*(1-alb)+LWDN
  REAL :: EC      ! Canopy water evaporation (W m-2)
  REAL :: EDIR    ! Direct soil evaporation (W m-2)
  REAL, allocatable, DIMENSION(:) :: ET      ! Plant transpiration from a particular root (soil) layer (W m-2)
  REAL :: ESNOW   ! Sublimation from (or deposition to if <0) snowpack (W m-2)
  REAL :: DRIP    ! Through-fall of precip and/or dew in excess of canopy
  ! water-holding capacity (m)
  REAL :: DEW     ! Dewfall (or Frostfall for T<273.15) (m)

  REAL :: BETA    ! Ratio of actual to potential evap (dimensionless)
  REAL :: ETP     ! Potential Evaporation (w m-2)
  REAL :: SSOIL   ! Soil heat flux (W m-2: negative if downward from surface)
  REAL :: FLX1    ! Precip-snow sfc (W m-2)
  REAL :: FLX2    ! Freezing rain latent heat flux (W m-2)      
  REAL :: FLX3    ! Phase-change heat flux from snowmelt (W m-2)
  REAL :: SNOMLT  ! Snow melt (m) (water equivalent)
  REAL :: RUNOFF1 ! Surface runoff (m s-1), not infiltrating the surface
  REAL :: RUNOFF2 ! Subsurface runoff (m s-1), drainage out bottom of last
  !       soil layer (baseflow).  Note: RUNOFF2 is actually 
  !       the sum of RUNOFF2 and RUNOFF3
  REAL :: RUNOFF3 ! Numerical trunctation in excess of porosity (SMCMAX)
  ! for a given soil layer at the end of a time step (m s-1).
  REAL :: RC      ! Canopy resistance (s m-1)
  REAL :: PC      ! Plant coefficient (dimensionless fraction, 0.0-1.0) 
  ! where PC*ETP = actual transpiration
  REAL :: RCS     ! Incoming solar RC factor (dimensionless)
  REAL :: RCT     ! Air temperature RC factor (dimensionless)
  REAL :: RCQ     ! Atmos. vapor pressure deficit RC factor (dimensionless)
  REAL :: RCSOIL  ! Soil moisture RC factor (dimensionless)
  REAL :: SOILW   ! Available soil moisture in root zone (dimensionless fraction
  !      between SMCWLT and SMCMAX)
  REAL :: SOILM   ! Total soil column moisture content (frozen+unfrozen) (m)
  REAL :: Q1      ! Effective specific humidity at surface (kg kg-1), used for
  !      diagnosing the specific humidity at 2 meter for 
  !      coupled model

  real ::  snoalb, frzx
  logical :: local = .TRUE.

  character(len=4) :: MMINSL

! Timing:
  integer :: clock_count_1 = 0
  integer :: clock_count_2 = 0
  integer :: clock_rate    = 0
  real    :: timing_sum    = 0.0

#ifdef _HRLDAS_URBAN_
! ----------------------------------------------------------------------
! DECLARATIONS START - urban
! ----------------------------------------------------------------------

  INTEGER :: KROOF
  INTEGER :: KWALL
  INTEGER :: KROAD
! input variables surface_driver --> lsm
  INTEGER, parameter  :: num_roof_layers = 4
  INTEGER, parameter  :: num_wall_layers = 4
  INTEGER, parameter  :: num_road_layers = 4

  REAL, DIMENSION(1:num_roof_layers)  :: DZR
  REAL, DIMENSION(1:num_wall_layers)  :: DZB
  REAL, DIMENSION(1:num_road_layers)  :: DZG

! input variables lsm --> urban
  INTEGER :: UTYPE_URB ! urban type [urban=1, suburban=2, rural=3]
  REAL :: TA_URB       ! potential temp at 1st atmospheric level [K]
  REAL :: QA_URB       ! mixing ratio at 1st atmospheric level  [kg/kg]
  REAL :: UA_URB       ! wind speed at 1st atmospheric level    [m/s]
  REAL :: U1_URB       ! u at 1st atmospheric level             [m/s]
  REAL :: V1_URB       ! v at 1st atmospheric level             [m/s]
  REAL :: SSG_URB      ! downward total short wave radiation    [W/m/m]
  REAL :: LLG_URB      ! downward long wave radiation           [W/m/m]
  REAL :: RAIN_URB     ! precipitation                          [mm/h]
  REAL :: RHOO_URB     ! air density                            [kg/m^3]
  REAL :: ZA_URB       ! first atmospheric level                [m]
  REAL :: DELT_URB     ! time step                              [s]
  REAL :: SSGD_URB     ! downward direct short wave radiation   [W/m/m]
  REAL :: SSGQ_URB     ! downward diffuse short wave radiation  [W/m/m]
  REAL :: XLAT_URB     ! latitude                               [deg]
  REAL :: OMG_URB      ! hour angle
  REAL :: ZNT_URB      ! roughness length                       [m]
  REAL :: TR_URB
  REAL :: TB_URB
  REAL :: TG_URB
  REAL :: TC_URB
  REAL :: QC_URB
  REAL :: UC_URB
  REAL :: XXXR_URB
  REAL :: XXXB_URB
  REAL :: XXXG_URB
  REAL :: XXXC_URB
  REAL, DIMENSION(1:num_roof_layers) :: TRL_URB  ! roof layer temp [K]
  REAL, DIMENSION(1:num_wall_layers) :: TBL_URB  ! wall layer temp [K]
  REAL, DIMENSION(1:num_road_layers) :: TGL_URB  ! road layer temp [K]
  LOGICAL :: LSOLAR_URB
! state variable surface_driver <--> lsm <--> urban
  REAL, ALLOCATABLE, DIMENSION(:,:) :: TR_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: TB_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: TG_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: TC_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: QC_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: UC_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: XXXR_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: XXXB_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: XXXG_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: XXXC_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: SH_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: LH_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: G_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: RN_URB2D
!
  REAL, ALLOCATABLE, DIMENSION(:,:) :: TS_URB2D

  REAL, ALLOCATABLE, DIMENSION( :, :, :) :: TRL_URB3D
  REAL, ALLOCATABLE, DIMENSION( :, :, :) :: TBL_URB3D
  REAL, ALLOCATABLE, DIMENSION( :, :, :) :: TGL_URB3D

! output variable lsm --> surface_driver
  REAL, ALLOCATABLE, DIMENSION(:,:) :: PSIM_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: PSIH_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: GZ1OZ0_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: U10_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: V10_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: TH2_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: Q2_URB2D
!
!m     REAL, ALLOCATABLE, DIMENSION(:,:) :: AKHS_URB2D 
  REAL, ALLOCATABLE, DIMENSION(:,:) :: AKMS_URB2D
!
  REAL, ALLOCATABLE, DIMENSION(:,:) :: UST_URB2D
  REAL, ALLOCATABLE, DIMENSION(:,:) :: FRC_URB2D
  INTEGER, ALLOCATABLE, DIMENSION(:,:) :: UTYPE_URB2D

! output variables urban --> lsm
  REAL :: TS_URB     ! surface radiative temperature    [K]
  REAL :: QS_URB     ! surface humidity                 [-]
  REAL :: SH_URB     ! sensible heat flux               [W/m/m]
  REAL :: LH_URB     ! latent heat flux                 [W/m/m]
  REAL :: LH_KINEMATIC_URB ! latent heat flux, kinetic  [kg/m/m/s]
  REAL :: SW_URB     ! upward short wave radiation flux [W/m/m]
  REAL :: ALB_URB    ! time-varying albedo            [fraction]
  REAL :: LW_URB     ! upward long wave radiation flux  [W/m/m]
  REAL :: G_URB      ! heat flux into the ground        [W/m/m]
  REAL :: RN_URB     ! net radiation                    [W/m/m]
  REAL :: PSIM_URB   ! shear f for momentum             [-]
  REAL :: PSIH_URB   ! shear f for heat                 [-]
  REAL :: GZ1OZ0_URB   ! shear f for heat                 [-]
  REAL :: U10_URB    ! wind u component at 10 m         [m/s]
  REAL :: V10_URB    ! wind v component at 10 m         [m/s]
  REAL :: TH2_URB    ! potential temperature at 2 m     [K]
  REAL :: Q2_URB     ! humidity at 2 m                  [-]
  REAL :: CHS_URB
  REAL :: CHS2_URB
  REAL :: UST_URB

! ----------------------------------------------------------------------
! DECLARATIONS END - urban
! ----------------------------------------------------------------------

#endif

        real :: CFACTR, CMCMAX, RSMAX, TOPT, REFKDT, KDT, SBETA, SHDFAC, RSMIN
        real :: RGL, HS, ZBOT, PSISAT, SLOPE, SNUP, SALP, BEXP, DKSAT, DWSAT
        real :: SMCMAX, SMCWLT, SMCREF, SMCDRY, F1, QUARTZ, FXEXP, Z0BRD, CZIL
        real :: XLAI, CSOIL, ALB, PTU
        real, dimension(NSOLD) :: RTDIS
        integer :: NROOT

!---------------------------------------------------------------------
!  DECLARE/Initialize constants
!---------------------------------------------------------------------

  REAL, PARAMETER :: R=287.04
  REAL, PARAMETER :: CPHEAT=1004.5
  INTEGER, PARAMETER :: LAND=1

  character(len=19) :: olddate, newdate, startdate
  CHARACTER(len=256) :: inflnm, outflnm, inflnm_template
  character :: hgrid
  integer   :: igrid
  logical   :: lexist

  logical            :: restart_flag

! NAMELIST:

  CHARACTER(len=256) :: indir
  character(len=256) :: hrldas_constants_file = " "
  character(len=256) :: external_fpar_filename_template = " "
  character(len=256) :: external_lai_filename_template = " "
  character(len=13)  :: restart_hdate = " "
  integer            :: split_output_count = 1
  integer            :: restart_frequency_hours
  integer            :: output_timestep
  integer            :: subwindow_xstart = 1
  integer            :: subwindow_ystart = 1
  integer            :: subwindow_xend = 0
  integer            :: subwindow_yend = 0



  integer :: START_YEAR, START_MONTH, START_DAY, START_HOUR, START_MIN
  integer :: noah_timestep = -999
  integer :: forcing_timestep = -999

  namelist / NOAHLSM_OFFLINE/ INDIR, NSOIL, ZSOIL, FORCING_TIMESTEP, NOAH_TIMESTEP, &
       START_YEAR, START_MONTH, START_DAY, START_HOUR, START_MIN, &
       RESTART_HDATE, RESTART_FREQUENCY_HOURS, OUTPUT_TIMESTEP, &
       SPLIT_OUTPUT_COUNT, & 
       KHOUR, KDAY, ZLVL, HRLDAS_CONSTANTS_FILE, &
       external_fpar_filename_template, external_lai_filename_template, &
       subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend
#ifdef _HRLDAS_URBAN_
  INTEGER  ::  ucmcall                        !urban
  REAL     ::  zlvl_urban
  namelist / URBAN_OFFLINE / UCMCALL, ZLVL_URBAN
#endif
!KWM       , VEGMIN, VEGMAX

  integer, parameter :: iunit = 10
  integer, parameter :: ounit = 11

  LSOIL = 'STAS'

  ! Initialize namelist variables to dummy values, so we can tell
  ! if they have not been set properly.

  nsoil = 4
  zsoil = -1.E36
  zsoil(1:4) = (/ -0.1, -0.4, -1.0, -2.0 /)

  dt = -999.

  start_year = -999
  start_month = -999
  start_day = -999
  start_hour = -999
  start_min = -999
  khour = -999
  kday = -999

  zlvl = 2.0
!KWM  vegmin = -999.
!KWM  vegmax = -999.

  output_timestep = 0
  restart_frequency_hours = 0

  open(30, file="namelist.hrldas", form="FORMATTED")
  read(30, NOAHLSM_OFFLINE, iostat=ierr)
  if (ierr /= 0) then
     write(*,'(/," ***** ERROR: Problem reading namelist NOAHLSM_OFFLINE",/)')
     rewind(30)
     read(30, NOAHLSM_OFFLINE)
     stop " ***** ERROR: Problem reading namelist NOAHLSM_OFFLINE"
  endif
#ifdef _HRLDAS_URBAN_
  zlvl_urban = 15.0 ! Default value
  read(30, URBAN_OFFLINE, iostat=ierr)
  if (ierr /= 0) then
     write(*,'(/," ***** ERROR: Problem reading namelist URBAN_OFFLINE",/)')
     rewind(30)
     read(30, URBAN_OFFLINE)
     stop " ***** ERROR: Problem reading namelist URBAN_OFFLINE"
  endif
#endif

  close(30)

  if ((khour < 0) .and. (kday < 0)) then
     write(*, '(" ***** Namelist error: ************************************")')
     write(*, '(" ***** ")')
     write(*, '(" *****      Either KHOUR or KDAY must be defined.")')
     write(*, '(" ***** ")')
     stop
  else if (( khour < 0 ) .and. (kday > 0)) then
     khour = kday * 24
  else if ((khour > 0) .and. (kday > 0)) then
     write(*, '("Namelist warning:  KHOUR and KDAY both defined.")')
  else
     ! all is well.  KHOUR defined
  endif

  if (forcing_timestep < 0) then
        write(*, *)
        write(*, '(" ***** Namelist error: *****************************************")')
        write(*, '(" ***** ")')
        write(*, '(" *****       FORCING_TIMESTEP needs to be set greater than zero.")')
        write(*, '(" ***** ")')
        write(*, *)
        stop
  endif

  if (noah_timestep < 0) then
        write(*, *)
        write(*, '(" ***** Namelist error: *****************************************")')
        write(*, '(" ***** ")')
        write(*, '(" *****       NOAH_TIMESTEP needs to be set greater than zero.")')
        write(*, '(" *****                     900 seconds is recommended.       ")')
        write(*, '(" ***** ")')
        write(*, *)
        stop
  endif

  !
  ! Check that OUTPUT_TIMESTEP fits into NOAH_TIMESTEP:
  !
  if (output_timestep /= 0) then
     if (mod(output_timestep, noah_timestep) > 0) then
        write(*, *)
        write(*, '(" ***** Namelist error: *********************************************************")')
        write(*, '(" ***** ")')
        write(*, '(" *****       OUTPUT_TIMESTEP should set to an integer multiple of NOAH_TIMESTEP.")')
        write(*, '(" *****            OUTPUT_TIMESTEP = ", I12, " seconds")') output_timestep
        write(*, '(" *****            NOAH_TIMESTEP   = ", I12, " seconds")') noah_timestep
        write(*, '(" ***** ")')
        write(*, *)
        stop
     endif
  endif

  !
  ! Check that RESTART_FREQUENCY_HOURS fits into NOAH_TIMESTEP:
  !
  if (restart_frequency_hours /= 0) then
     if (mod(restart_frequency_hours*3600, noah_timestep) > 0) then
        write(*, *)
        write(*, '(" ***** Namelist error: ******************************************************")')
        write(*, '(" ***** ")')
        write(*, '(" *****       RESTART_FREQUENCY_HOURS (converted to seconds) should set to an ")')
        write(*, '(" *****       integer multiple of NOAH_TIMESTEP.")')
        write(*, '(" *****            RESTART_FREQUENCY_HOURS = ", I12, " hours:  ", I12, " seconds")') &
             restart_frequency_hours, restart_frequency_hours*3600
        write(*, '(" *****            NOAH_TIMESTEP           = ", I12, " seconds")') noah_timestep
        write(*, '(" ***** ")')
        write(*, *)
        stop
     endif
  endif


  dt = real(noah_timestep)

  write(olddate,'(I4.4,"-",I2.2,"-",I2.2,"_",I2.2,":",I2.2,":",I2.2)') &
       start_year, start_month, start_day, start_hour, start_min, 0

  startdate = olddate

  call read_hrldas_hdrinfo(hrldas_constants_file, ix, jx, &
       subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend, &
       iswater, isurban, llanduse, dx, dy, truelat1, truelat2, cen_lon, lat1, lon1, &
       igrid, mapproj)
  write(hgrid,'(I1)') igrid
  print*, 'llanduse = ', llanduse

  if (nsoil < 0) then
     stop " ***** ERROR: NSOIL must be set in the namelist."
  endif

  write(olddate,'(I4.4,"-",I2.2,"-",I2.2,"_",I2.2,":",I2.2,":",I2.2)') &
       start_year, start_month, start_day, start_hour, start_min, 0

  startdate = olddate

!......................... end of model configuration (later in a namelist) .....

  MMINSL='STAS'
  CALL SOIL_VEG_GEN_PARM( LLANDUSE, MMINSL )

!----------------------------------------------------------------------
! Allocate arrays for our gridded domain, now that we know the size
!----------------------------------------------------------------------
  allocate( VEGTYP   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SOLTYP   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( TERRAIN  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( LATITUDE (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( LONGITUDE(SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( T2       (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( XLONG    (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( U        (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( V        (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( PRES     (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SHORT    (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( PRCP1    (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( FPAR     (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ZNT      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( LAI      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( CMC      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SNODEP   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( WEASD    (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( T1       (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ETPX     (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ETAX     (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ETAKIN   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( GRDFLX   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( CHX      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( RUNOFF1X (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( RUNOFF2X (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( RUNOFF3X (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( EDIRX    (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ECX      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ETTX     (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SNMAXX   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( RCX      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( HX       (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( QFX      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( Q2X      (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SFCSPDX  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ALBEDX   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SMCMAX1  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SMCWLT1  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SMCREF1  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ACSNOM   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ACRAIN   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( ESNOW2D  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( SNOFLXX  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SNOEVPX  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( FX       (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( RESX     (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( DRIP2D   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( DEWFALL  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( SOILMX   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( EMISS    (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( XLAI2D   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( GVFMIN   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( GVFMAX   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
  allocate( TBOT_2D  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( ETPNDX   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( SFCHEAD  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( INFXS1   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( PDDUM2   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( PCPDRP   (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
!KWM  allocate( SFCWATR2 (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )

  allocate( SMC  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND,NSOIL) ) 
  allocate( STC  (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND,NSOIL) )
  allocate( SH2OX(SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND,NSOIL) )
  allocate( ZSOILX(SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND,NSOIL) )

  allocate( SH2O   (NSOIL) )
  allocate( SLDPTH (NSOIL) )
  allocate( SMC1   (NSOIL) )
  allocate( STC1   (NSOIL) )
  allocate( ET     (NSOIL) )

#ifdef _HRLDAS_URBAN_
  call urban_param_init(DZR,DZB,DZG,num_roof_layers)
  ALLOCATE(TR_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(TB_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(TG_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(TC_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(QC_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(UC_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(XXXR_URB2D( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(XXXB_URB2D( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(XXXG_URB2D( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(XXXC_URB2D( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(SH_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(LH_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(G_URB2D   ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(RN_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(TS_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(TRL_URB3D ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND, 1:NUM_ROOF_LAYERS ) )
  ALLOCATE(TBL_URB3D ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND, 1:NUM_WALL_LAYERS ) )
  ALLOCATE(TGL_URB3D ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND, 1:NUM_ROAD_LAYERS ) )
  ALLOCATE(PSIM_URB2D   ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(PSIH_URB2D   ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(GZ1OZ0_URB2D ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(U10_URB2D    ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(V10_URB2D    ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(TH2_URB2D    ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(Q2_URB2D     ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(AKMS_URB2D   ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(UST_URB2D    ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(FRC_URB2D    ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )
  ALLOCATE(UTYPE_URB2D  ( SUBWINDOW_XSTART:SUBWINDOW_XEND, SUBWINDOW_YSTART:SUBWINDOW_YEND ) )

#endif

!----------------------------------------------------------------------
! Initialize gridded domain
!----------------------------------------------------------------------

  ! SLDPTH is the thickness of each layer
  SLDPTH(1) = -ZSOIL(1)
  do i = 2, nsoil
     sldpth(i) = zsoil(i-1)-zsoil(i)
  enddo

  ETAX=0.0 ! -999.9
  ETAKIN=0.0 ! -999.9
  ETPX=-999.9
  CHX=-999.9
  FX=-999.9
  HX=-999.9
  RESX=-999.9
  GRDFLX=-999.9
  CMC=0.0
  T2=-999.9
  T1=-999.9
  PRCP1=-999.9

  RUNOFF1X=0.0 
  RUNOFF2X=0.0 
  RUNOFF3X=0.0 
  EDIRX=0.0
  ETTX=0.0
!FC  ETPNDX=-999.9
  SNOEVPX=-999.9
  SNODEP=-999.9

  STC=-999.9
  SH2OX=-999.9
  SMC=-999.9

  ECX=0.0
  ACSNOM  = 0.0
  ACRAIN  = 0.0
  ESNOW2D = 0.0
  DRIP2D = 0.0
  DEWFALL = 0.0
  SOILMX = -999.
  XLAI2D = 0.
  GVFMIN = -999.
  GVFMAX = -999.

!---------------------------------------------------------------------
! Initialize static surface data
!---------------------------------------------------------------------


!   OK     VEGTYP     VEGETATION TYPE (INTEGER INDEX)
!   OK     SOILTYP    SOIL TYPE (INTEGER INDEX)
!   OK     SLOPETYP   CLASS OF SFC SLOPE (INTEGER INDEX)
!          SHDMIN     MINIMUM AREAL FRACTIONAL COVERAGE OF GREEN VEGETATION
!                (FRACTION= 0.0-1.0) <= SHDFAC
!   N/U    PTU        PHOTO THERMAL UNIT (PLANT PHENOLOGY FOR ANNUALS/CROPS)
!                (NOT YET USED, BUT PASSED TO REDPRM FOR FUTURE USE IN
!                VEG PARMS)
!   OK     ALB        BACKROUND SNOW-FREE SURFACE ALBEDO (FRACTION), FOR JULIAN
!                DAY OF YEAR (USUALLY FROM TEMPORAL INTERPOLATION OF
!                MONTHLY MEAN VALUES' CALLING PROG MAY OR MAY NOT
!                INCLUDE DIURNAL SUN ANGLE EFFECT)
!   OK     SNOALB     UPPER BOUND ON MAXIMUM ALBEDO OVER DEEP SNOW (E.G. FROM
!                ROBINSON AND KUKLA, 1985, J. CLIM. & APPL. METEOR.)
!   OK     TBOT       BOTTOM SOIL TEMPERATURE (LOCAL YEARLY-MEAN SFC AIR
!                TEMPERATURE)
!   ??     Z0BRD      Background fixed roughness length (M)
!   ??     Z0         Time varying roughness length (M) as function of snow depth




!----------------------------------------------------------------------
! Read Landuse Type and Soil Texture and Other Information
!----------------------------------------------------------------------

  CALL READLAND_HRLDAS(hrldas_constants_file,                           &
       subwindow_xstart, subwindow_xend,                                &
       subwindow_ystart, subwindow_yend,                                &
       ISWATER, VEGTYP, SOLTYP, TERRAIN, TBOT_2D, LATITUDE, LONGITUDE)

!----------------------------------------------------------------------
! Initialize Model State
!----------------------------------------------------------------------

  SLOPETYP = 2

  if (restart_hdate /= " ") then
     restart_flag = .TRUE.

     if (trim(restart_hdate) == "LATEST") then
        ! Find the latest RESTART file
        call geth_newdate(olddate(1:13), olddate(1:13), khour+24)
        RLOOP : do
           inquire (file="RESTART."//olddate(1:4)//olddate(6:7)//olddate(9:10)//&
                &olddate(12:13)//"_DOMAIN"//trim(hgrid), exist=lexist)
           if ( .not. lexist ) then
              call geth_newdate(olddate(1:13), olddate(1:13), -1)
              if (olddate(1:13) < startdate(1:13) ) then
                 write(*, *)
                 write(*, '(" ***** RESTART error: **************************************************")')
                 write(*, '(" ***** ")')
                 write(*, '(" *****       You have requested to restart from the latest restart file,")')
                 write(*, '(" *****       but no restart file was found in the run directory.")')
                 write(*, '(" ***** ")')
                 write(*, *)
                 stop " ***** ERROR EXIT:  Cannot find a restart file"
              endif
           else
              restart_hdate(1:13) = olddate(1:13)
              print*, 'Found restart date ', restart_hdate
              exit RLOOP
           endif
        enddo RLOOP
     endif

     olddate = restart_hdate(1:13)//":00:00"
     call read_restart(igrid, olddate,                                                  &
          subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, nsoil,  &
#ifdef _HRLDAS_URBAN_
          ucmcall,                                                                      &
          ts_urb2d, tr_urb2d, tb_urb2d, tg_urb2d, tc_urb2d, qc_urb2d, uc_urb2d,         &
          trl_urb3d, tbl_urb3d, tgl_urb3d, &
          xxxr_urb2d, xxxb_urb2d, xxxg_urb2d, xxxc_urb2d,                               &
          frc_urb2d, utype_urb2d,                                                       &
#endif
     stc, smc, sh2ox, qfx, snodep, weasd, fx, etpx, t1, etax, etakin, cmc, grdflx, &
          chx, runoff1x, runoff2x, runoff3x, edirx, ecx, ettx, rcx, hx, smcmax1,        &
          smcref1, smcwlt1, albedx, acrain, acsnom, esnow2d, drip2d, dewfall, soilmx)

     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "EMISS", emiss)
     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "FPAR", fpar)
     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "ZNT", znt)
     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "GVFMIN", gvfmin)
     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "GVFMAX", gvfmax)

     allocate(fpar_save(SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "FPAR_SAVE", fpar_save, ierr)
     if (ierr /= 0) deallocate(fpar_save)

     allocate(lai_save(SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
     call get_from_restart(olddate, igrid, subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, "LAI_SAVE",  lai_save,  ierr)
     if (ierr /= 0) deallocate(lai_save)

  else

     restart_flag = .FALSE.
     inflnm = trim(indir)//"/"//&
          startdate(1:4)//startdate(6:7)//startdate(9:10)//startdate(12:13)//&
          ".LDASIN_DOMAIN"//hgrid

     CALL READINIT_HRLDAS(inflnm,                                              &
          subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend,  &
          NSOIL, SLDPTH, OLDDATE, LDASIN_VERSION, SMC, STC, SH2OX, CMC, T1, &
          WEASD, SNODEP)
     ! *** Read lai, fveg etc.
     CALL READVEG_HRLDAS(inflnm,                                               &
          subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend,  &
          OLDDATE, VEGTYP, FPAR, LAI, GVFMIN, GVFMAX)

     SOILMX = -1.0 * SMC (:,:,1)* ZSOIL (1)
     DO K = 2,NSOIL
        where (SMC(:,:,K) > -1.E25)
           SOILMX = SOILMX + SMC(:,:,K) * (ZSOIL (K-1) - ZSOIL (K))
        elsewhere
           SOILMX = -1.E36
        endwhere
           
     ENDDO
     where (SOILMX > -1.E25) SOILMX = SOILMX * 1.E3 ! Convert from m to mm
     
     HX = 0.0
     QFX = 0.0
     ZNT = -1.0

  endif

#ifdef _HRLDAS_URBAN_
  if (ucmcall == 1) then
     ! Initialize urban stuff.
     if (restart_hdate(1:1) == " ") then
        write(*, '("Initialize urban var.")')
        call urban_var_init(T1, STC, STC(:,:,4),VEGTYP,                                 &
             ISURBAN,                                                                   &
             SUBWINDOW_XSTART, SUBWINDOW_XEND, SUBWINDOW_YSTART, SUBWINDOW_YEND, NSOIL, &
             restart_flag,                                                              &
             XXXR_URB2D,XXXB_URB2D,XXXG_URB2D,XXXC_URB2D,                               & ! inout ! OUT ???
             TR_URB2D,TB_URB2D,TG_URB2D,TC_URB2D,QC_URB2D,                              & ! inout ! OUT ???
             TRL_URB3D,TBL_URB3D,TGL_URB3D,                                             & ! inout ! OUT ???
             SH_URB2D,LH_URB2D,G_URB2D,RN_URB2D,                                        & ! inout ! OUT ???
             TS_URB2D, FRC_URB2D, UTYPE_URB2D)                                            ! inout ! OUT ???
     else
        write(*, '("Restart, so we do not initialize urban var")')
     endif
  endif
#endif

!------------------------------------------------------------------------
! Begin Time Loop
!------------------------------------------------------------------------

  NTIME=(KHOUR+1)*3600./nint(dt)
  ! Start a timer
  call system_clock(count=clock_count_1)

  KLOOP : DO K=1,NTIME

!--------------------------------------------------------------------------------
! Output for restart BEFORE the processing for this particular time has begun, 
! so that upon restart, we're ready to go on this time step.
!--------------------------------------------------------------------------------

     if (restart_frequency_hours > 0) then
        if ((k > 1) .and. (mod(k-1, restart_frequency_hours) == 0)) then
           print*, 'Write restart at '//olddate(1:13)
           call write_restart(version, igrid, llanduse, olddate, startdate,                   & 
                subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1,         &
                nsoil, dx, dy, truelat1, truelat2, mapproj, lat1, lon1, cen_lon,              &
                iswater, vegtyp,                                                              &
#ifdef _HRLDAS_URBAN_
                ucmcall,                                                                      &
                ts_urb2d, tr_urb2d, tb_urb2d, tg_urb2d, tc_urb2d, qc_urb2d, uc_urb2d,         &
                trl_urb3d, tbl_urb3d, tgl_urb3d, &
                xxxr_urb2d, xxxb_urb2d, xxxg_urb2d, xxxc_urb2d,                               &
                frc_urb2d, utype_urb2d,                                                       &
#endif
           stc, smc, sh2ox, qfx, snodep, weasd, fx, etpx, t1, etax, etakin, cmc, grdflx, &
                chx, runoff1x, runoff2x, runoff3x, edirx, ecx, ettx, rcx, hx, smcmax1,        &
                smcref1, smcwlt1, albedx, acrain, acsnom, esnow2d, drip2d, dewfall, soilmx,   &
                fpar, znt)

           call add_to_restart(olddate, igrid, iswater, &
                subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, vegtyp,  &
                "EMISS", "fraction", "Emissivity (including possible snow-cover effect)", emiss)

           call add_to_restart(olddate, igrid, iswater, &
                subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, vegtyp,  &
                "GVFMIN", "fraction", "Minimum Green Vegetation Fraction", gvfmin)

           call add_to_restart(olddate, igrid, iswater, &
                subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, vegtyp,  &
                "GVFMAX", "fraction", "Maximum Green Vegetation Fraction", gvfmax)

           if (allocated(FPAR_SAVE)) then
              call add_to_restart(olddate, igrid, iswater, &
                   subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, vegtyp,  &
                   "FPAR_SAVE", "fraction", "Remembered FPAR field from external files", FPAR_SAVE)
           endif

           if (allocated(LAI_SAVE)) then
              call add_to_restart(olddate, igrid, iswater, &
                   subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, vegtyp,  &
                   "LAI_SAVE", "index", "Remembered LAI field from external files", LAI_SAVE)
           endif

        endif
     endif


!---------------------------------------------------------------------------------
! Read the forcing data.
!---------------------------------------------------------------------------------

! For HRLDAS, we're assuming (for now) that each time period is in a 
! separate file.  So we can open a new one right now.

     inflnm = trim(indir)//"/"//&
          olddate(1:4)//olddate(6:7)//olddate(9:10)//olddate(12:13)//&
          ".LDASIN_DOMAIN"//hgrid

     ! Build a filename template
     inflnm_template = trim(indir)//"/<date>.LDASIN_DOMAIN"//hgrid

     CALL READFORC_HRLDAS(inflnm_template, forcing_timestep, olddate,  &
          subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend,    &
          T2,Q2X,U,V,PRES,XLONG,SHORT,PRCP1,LAI,FPAR)

! Every 24 hours, update the snow field from analyses.
     if ( OLDDATE(12:13) == "00") then
        CALL READSNOW_HRLDAS(inflnm,                                             &
             subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend, &
             OLDDATE,WEASD,SNODEP)
     endif

!----------------------------------------------------------------------------------
! Read possible external maps.  There's got to be some good way to make this
! more general.
!----------------------------------------------------------------------------------

     if (external_fpar_filename_template /= " ") then
        call read_additional(external_fpar_filename_template, olddate, "fPAR", &
             subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend, FPAR, ierr)
        if (ierr == 0) then
           if ( .not. allocated(FPAR_SAVE)) then
              allocate( FPAR_SAVE (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
           endif
           FPAR_SAVE = FPAR
        else
           if (allocated(FPAR_SAVE)) then
              FPAR = FPAR_SAVE
           endif
        endif
     endif

     if (external_lai_filename_template /= " ") then
        call read_additional(external_lai_filename_template,  olddate, "LAI",  &
             subwindow_xstart, subwindow_xend, subwindow_ystart, subwindow_yend, LAI, ierr)
        if (ierr == 0) then
           if ( .not. allocated(LAI_SAVE)) then
              allocate( LAI_SAVE (SUBWINDOW_XSTART:SUBWINDOW_XEND,SUBWINDOW_YSTART:SUBWINDOW_YEND) )
           endif
           LAI_SAVE = LAI
        else
           if (allocated(LAI_SAVE)) then
              LAI = LAI_SAVE
           endif
        endif
     endif

!------------------------------------------------------------------------
! Spatial Loop to Convert Gridded data to single point for 1-D SFLX call
!------------------------------------------------------------------------


     JLOOP : DO J=subwindow_ystart,subwindow_yend
        ILOOP : DO I=subwindow_xstart, subwindow_xend
           if (gvfmin(i,j) < -1.E25) cycle ILOOP

           IF((VEGTYP(I,J).GT.0).AND.(VEGTYP(I,J).NE.ISWATER)) THEN
              VEGTYPX=VEGTYP(I,J)
              SOILTYP=SOLTYP(I,J)
              SFCTMP=T2(I,J)
              SFCSPD=SQRT(U(I,J)*U(I,J)+V(I,J)*V(I,J))
              SFCPRS=PRES(I,J)
              Q2=Q2X(I,J)/(1.0+Q2X(I,J)) ! Convert mixing ratio to specific humidity
              PRCP=PRCP1(I,J)
              if (prcp < 0.0) prcp = 0.0 ! Just in case
              SOLDN=SHORT(I,J)

              !
              ! The following module variables are set in subroutine REDPRM.
              ! If you want other than the default values based on land/soil
              ! categories, reset them after the call to REDPRM.
              !
              !    -- SOIL PARAMETERS:
              !
              !           CSOIL
              !           BEXP
              !           DKSAT
              !           DWSAT
              !           F1
              !           PSISAT
              !           QUARTZ
              !           SMCDRY
              !           SMCMAX
              !           SMCREF
              !           SMCWLT
              !
              !    -- "UNIVERSAL" PARAMETERS:
              !
              !           ZBOT
              !           SALP
              !           SBETA
              !           REFDK
              !           FRZK
              !           FXEXP
              !           REFKDT
              !           PTU
              !           KDT
              !           CZIL
              !           SLOPE
              !           FRZFACT
              ! 
              !    -- VEGETATION PARAMETERS:
              !
              !           TOPT
              !           RGL
              !           RSMAX
              !           RSMIN
              !           HS
              !           XLAI
              !           CMCMAX
              !           CFACTR
              !           NROOT
              !           SNUP
              !           ALB
              !           Z0BRD
              !           SHDFAC
              !           SNOALB_NOAH
              !           RTDIS
              !

#ifdef _HRLDAS_URBAN_
              IF ( UCMCALL == 1 ) THEN
                 IF ( ( VEGTYP(I,J) == ISURBAN  ) .or. ( VEGTYP(I,J) == 31 ) .or. &
                      ( VEGTYP(I,J) == 32 ) .or. ( VEGTYP(I,J) == 33 ) ) THEN
                    ! VEGTYPX = 5   ! HARD WIRED CATEGORY IS A PROBLEM.  NEEDS CORRECTION
                    VEGTYPX = 10   ! HARD WIRED CATEGORY IS A PROBLEM.  NEEDS CORRECTION
                 ENDIF
              ELSE
                 IF ( ( VEGTYP(I,J) == ISURBAN  ) .or. ( VEGTYP(I,J) == 31 ) .or. &
                      ( VEGTYP(I,J) == 32 ) .or. ( VEGTYP(I,J) == 33 ) ) THEN
                    VEGTYPX = ISURBAN
           	 ENDIF
              ENDIF
#endif

              CALL REDPRM (VEGTYPX,SOILTYP,SLOPETYP,CFACTR,CMCMAX,RSMAX,TOPT,   &
                   REFKDT,KDT,SBETA, SHDFAC,RSMIN,RGL,HS,ZBOT,FRZX,            &
                   PSISAT,SLOPE,SNUP,SALP,BEXP,DKSAT,DWSAT,                    &
                   SMCMAX,SMCWLT,SMCREF,SMCDRY,F1,QUARTZ,FXEXP,                &
                   RTDIS,SLDPTH,ZSOIL,NROOT,NSOIL,Z0BRD,CZIL,XLAI,             &
                   CSOIL,ALBBRD,PTU,LLANDUSE,LSOIL,.True.)

              offline_cfactr = cfactr
              offline_cmcmax = cmcmax
              offline_rsmax  = rsmax
              offline_topt   = topt
              offline_refkdt = refkdt
              offline_kdt    = kdt
              offline_sbeta  = sbeta
              offline_rsmin  = rsmin
              offline_rgl    = rgl
              offline_hs     = hs
              offline_zbot   = zbot
              offline_frzx   = frzx
              offline_psisat = psisat
              offline_slope  = slope
              offline_snup   = snup
              offline_salp   = salp
              offline_bexp   = bexp
              offline_dksat  = dksat
              offline_dwsat  = dwsat
              offline_smcmax = smcmax
              offline_smcwlt = smcwlt
              offline_smcref = smcref
              offline_smcdry = smcdry
              offline_f1     = f1
              offline_quartz = quartz
              offline_fxexp  = fxexp
              offline_rtdis  = rtdis
              offline_nroot  = nroot
              offline_czil   = czil
              offline_csoil  = csoil
              offline_ptu    = ptu

              SHDFAC = FPAR(I,J)
              if (VEGTYPX==BARE) SHDFAC=0.0; !KWM Added 20060920 to be consistent with REDPRM
              IF(VEGTYPX==ISURBAN)THEN
                 !urban change
                 SHDFAC=0.05
                 RSMIN=400.0
                 SMCMAX = 0.45
                 SMCREF = 0.42
                 SMCWLT = 0.40
              ENDIF

              if (gvfmax(i,j) == gvfmin(i,j)) then
                 fraction = 0.0
              else
                 fraction = (SHDFAC-GVFMIN(I,J))/(GVFMAX(I,J)-GVFMIN(I,J))
              endif
              fraction = max(fraction,0.)
              fraction = min(fraction,1.)
              embrd  = emimintbl(vegtypx)+fraction*(emimaxtbl(vegtypx)-emimintbl(vegtypx))
              z0brd  =  z0mintbl(vegtypx)+fraction*( z0maxtbl(vegtypx)- z0mintbl(vegtypx))
              albbrd = albmaxtbl(vegtypx)+fraction*(albmintbl(vegtypx)-albmaxtbl(vegtypx))

              if (allocated(LAI_SAVE)) then
                 xlai = LAI(I,J) !EXTERNAL
                 if (xlai==0) xlai=0.05
              else
                 xlai   = laimintbl(vegtypx)+fraction*(laimaxtbl(vegtypx)-laimintbl(vegtypx))
              endif

              !
              !  First time step, initialize EMISS from EMBRD.  Subsequent time steps will use
              !  the EMISS value calculcated in the previous time step.  This means that EMISS
              !  now has to go into the RESTART file.
              !

              if ( ( k==0 ) .and. ( .not. restart_flag ) ) then
                 emiss(i,j) = embrd
              endif

              LWDN = XLONG(I,J) * EMISS(I,J)

!   snow albedo
              alb = albbrd !?
              albedo = alb !?
              SNOFAC=MIN(SNODEP(I,J)*5.0, 1.0)
              ALBED=ALBEDO+(1.0-FPAR(I,J))*SNOFAC*(0.55-ALBEDO)
!   Net downward radiation
!              F= SRAD*(1.0-ALBED)+LWDN
!   
              SOLNET=SOLDN*(1.-ALBED) !KWM ???

              CH = 0.1
              CM = 0.1

              if (ZNT(I,J) < 0) then
                 Z0 = Z0BRD
              else
                 Z0 = ZNT(I,J)
              endif

!     Calculate a saturation specific humidity and slope of saturation specific humidity curve WRT Temperature

              CALL CALTMP(T1(I,J), SFCTMP, SFCPRS, ZLVL, Q2, TH2, T1V, TH2V, RHO)
              CALL CALHUM(SFCTMP, SFCPRS, Q2SAT, DQSDT2)

!     Calculate the surface exchange coefficients CM, CH
              ! Intent (IN) :: ZLVL, Z0, T1V, TH2V, SFCSPD, CZIL
              ! Intent (INOUT) :: CM
              CALL SFCDIF_off (ZLVL,Z0,T1V,TH2V,SFCSPD,CZIL,CM,CH)

              IF (Q2 .LT. 0.0) THEN
                 print*,'Q2<0','I=',I, 'J=',J
                 Q2=.1E-5
              ENDIF
              IF (Q2 .GT. Q2SAT) THEN
                 ! print*, 'Q2 .GT. Q2SAT', 'I=',I, 'J=',J
                 ! print*,'Q2=',Q2,'Q2SAT=',Q2SAT
                 Q2=Q2SAT*0.99
              ENDIF


              CHKFF = CH * CPHEAT * RHO           

              STC1(1:NSOIL)=STC(I,J,1:NSOIL)

              SMC1(1:NSOIL)=SMC(I,J,1:NSOIL)
              SH2O(1:NSOIL)=SH2OX(I,J,1:NSOIL)
              ZSOILX(I,J,1:NSOIL)=ZSOIL(1:NSOIL)
! *** diagnostics 
              SFCSPDX(I,J)=SFCSPD


!--- Other conversions 'History (State) Variables

              TBOT=TBOT_2D(I,J)
              CMCX=CMC(I,J)
              T1X=T1(I,J)
              SNOWH=SNODEP(I,J)

              SNEQV = WEASD(I,J)
              shdmin = 0.0

              esnow = 0.0
              if (T1X <= 273.15) then
                 ffrozp = 1.0
              else
                 ffrozp = 0.0
              endif

#ifdef _HRLDAS_URBAN_
              IF ( UCMCALL == 1 ) THEN
                 IF ( ( VEGTYP(I,J) == ISURBAN  ) .or. ( VEGTYP(I,J) == 31 ) .or. &
                      ( VEGTYP(I,J) == 32 ) .or. ( VEGTYP(I,J) == 33 ) ) THEN
                    ! VEGTYPX = 5   ! HARD WIRED CATEGORY IS A PROBLEM.  NEEDS CORRECTION
                    VEGTYPX = 10   ! HARD WIRED CATEGORY IS A PROBLEM.  NEEDS CORRECTION
                    SHDFAC = 0.8 
                    ALBEDO  =0.2
                    ALBBRD  =0.2
                    T1X= ( T1(I,J) -FRC_URB2D(I,J) * TS_URB2D (I,J) )/ (1-FRC_URB2D(I,J))
                 ENDIF
              ELSE
                 IF ( ( VEGTYP(I,J) == ISURBAN  ) .or. ( VEGTYP(I,J) == 31 ) .or. &
                      ( VEGTYP(I,J) == 32 ) .or. ( VEGTYP(I,J) == 33 ) ) THEN
                    VEGTYPX = ISURBAN
           	 ENDIF
              ENDIF
#endif



              ! The following variables are available here, to subroutine 
              ! SFLX, and to its deeper subroutines via use-association with 
              ! module noahlsm_globals.
              ! 
              !    ALB   : Surface albedo (Fraction 0.0 to 1.0)
              !    Z0BRD : Roughness Length (m)
              !    SHDFAC: Green Vegetation Fraction
              !          : Areal fractional coverage of green vegetation (fraction: 0.0-1.0)
              !    NROOT : Rooting depth (as the number of layers)
              !    RSMIN : Minimum Canopy Resistance (s m-1)
              !    RGL   : Parameter used in radiation stress function
              !    HS    : Parameter used in vapor pressure deficit function
              !    SNUP  : Threshold snow depth (in water equivalent m) that
              !            implies 100 percent snow cover
              !    XLAI  : Leaf Area Index
              !    SNOALB_NOAH: Upper bound on maximum albedo over deep snow (e.g., from
              !            Robinson and Kukla, 1985, J. Clim. & Appl. Meteor.)
              !    TOPT  : Optimum transpiration air temperature.
              !    CMCMAX: Maximum canopy water capacity
              !    CFACTR: Parameter used in the canopy inteception calculation
              !    RSMAX : Maximum stomatal resistance
              !    PTU   : Photo thermal unit (plant phenology for annuals/crops)
              !
              !    BEXP  : B parameter
              !    SMCDRY: Dry soil moisture threshold where direct evap from top 
              !            layer ends (volumetric)
              !    F1    : Soil thermal diffusivity/conductivity coef.
              !    SMCMAX: Porosity, i.e. saturated value of soil moisture (volumetric)
              !             -- Max soil moisture content (porosity)
              !             -- Saturation soil moisture content (from REDPRM)
              !    SMCREF: Soil moisture threshold where transpiration begins to 
              !            stress (volumetric)
              !             -- Reference soil moisture  (field capacity)
              !             -- Reference soil moisture (where soil water 
              !                deficit stress sets in)
              !    PSISAT: Saturated soil matric potential
              !    DKSAT : Saturated soil conductivity
              !    DWSAT : 
              !    SMCWLT: Wilting point soil moisture (volumetric)
              !    QUARTZ: Soil quartz content
              !
              !    SLOPE
              !    SBETA
              !    FXEXP
              !    CSOIL
              !    SALP  : Tuning parameter
              !    REFDK : Parameter in the surface runoff parameterization
              !    REFKDT: Parameter in the surface runoff parameterization
              !    FRZFACT: Frozen ground parameter
              !    ZBOT  : Depth (m) of lower boundary soil temperature
              !    CZIL  : Calculate roughness length of heat
              !    KDT   
              !
              ! Arguments to SFLX are:
              ! ----------------------------------------------------------------------
              ! 1. CONFIGURATION INFORMATION (C):
              ! ----------------------------------------------------------------------
              !    ICE   : Sea-ice flag (1=ice, 0=land)
              !    DT    : Timestep (sec)
              !    ZLVL  : Height (m) above ground of atmospheric forcing variables
              !    NSOIL : Number of soil layers.
              !    SLDPTH: Thickness of each soil layer (m)
              ! ----------------------------------------------------------------------
              ! 3. FORCING DATA (F):
              ! ----------------------------------------------------------------------
              !    LWDN  : LW downward radiation, not net (w m-2)
              !    SOLDN : Solar downward radiation, not net (w m-2)
              !    SFCPRS: Pressure  (Pa) at height ZLVL above ground
              !    PRCP  : Precipitation rate (kg m-2 s-1; or mm/s)
              !    SFCTMP: Air temperature (K) at height ZLVL above ground
              !    Q2    : Specific humidity (kg kg-1) at height ZLVL above ground
              !    SFCSPD: Wind speed (m s-1) at height ZLVL above ground
! UNUSED     !    COSZ  : Solar zenith angle
! UNUSED     !    PRCPRAIN: Liquid-precipitation rate (kg m-2 s-1) (Not used)
! UNUSED     !    SOLARDIRECT: Direct component of downward solar radiation (W m-2) (Not used)
              !    TH2   : Air potential temperature (K) at height ZLVL above ground
              ! ----------------------------------------------------------------------
              ! 4. OTHER FORCING (INPUT) DATA (I):
              ! ----------------------------------------------------------------------
              !    Q2SAT : Saturation specific humidity (kg kg-1) at height ZLVL above ground
              !    DQSDT2: Slope (kg kg-1 K-1) of sat specific humidity curve at T=SFCTMP
              ! ----------------------------------------------------------------------
              ! 5. CANOPY/SOIL CHARACTERISTICS (S):
              ! ----------------------------------------------------------------------
              !   VEGTYPX: Vegetation type (integer index)
              !   SHDMIN : Minimum areal fractional coverage of green vegetation (Fraction= 0.0-1.0)
              !   SHDMAX : Maximum areal fractional coverage of annual green vegetation (Not used)
              !   TBOT   : Bottom soil temperature (K) (local yearly-mean sfc air Temp.
              !   Z0     : Roughness length (m) modifed by snow depth as appropriate
              ! ----------------------------------------------------------------------
              ! 6. HISTORY (STATE) VARIABLES (H):
              ! ----------------------------------------------------------------------
              !   CMCX   : Canopy moisture content (m)
              !   T1X    : Ground/canopy/snowpack effective skin temperature (K)
              !   STC1   : Soil Temperature (K) :: dimension(NSOIL)
              !   SMC1   : Total soil moisture content (volumetric fraction) :: dimension(NSOIL)
              !   SH2O   : Unfrozen soil moisture content (volumetric fraction) :: dimension(NSOIL)
              !            NOTE: Frozen soil moisture = SMC - SH2O
              !   SNOWH  : Actual snow depth (m)
              !   SNEQV  : Liquid water-equivalent snow depth (m).  NOTE: Snow density = SNEQV/SNOWH
              !   ALBEDO : Surface albedo including snow effect (Fraction)
              !   CH     : Surface exchange coefficient for heat and moisture (m s-1)
              !            NOTE: CH is technically a conductance since it has been 
              !            multiplied by wind speed.
! UNUSED     !   CM     : Surface exchange coefficient for momentum (m s-1)
              !            NOTE: CM is technically a conductance since it has been
              !            multiplied by wind speed. (Not used)
              ! ----------------------------------------------------------------------
              ! 7. OUTPUT (O):
              ! ----------------------------------------------------------------------
              !   ETA    : Actual latent heat flux (W m-2)
              !   SHEAT  : Sensible heat flux (W m-2)
              !   ETA_KINEMATIC : Actual latent heat flux (kg m s-1)
              !   FDOWN  : Radiation forcing at the surface (W m-2)
              !   EC     : Canopy water evaporation (W m-2)
              !   EDIR   : Direct soil evaporation (W m-2)
              !   ET     : Plant transpiration from a particular root/soil layer (W m-2) :: dimension(NSOIL)
              !   ETT    : Total plant transpiration (W m-2)
              !   ESNOW  : Sublimation from (or deposition to if ESNOW<0) snowpack (W m-2)
              !   DRIP   : Through-fall of precip and/or dew in excess of canopy water-holding capacity (m)
              !   DEW    : Dewfall (or frostfall for T<273.15) (m)
              !   BETA   : Ratio of actual/potential evaporation (dimensionless)
              !   ETP    : Potential evaporation (W m-2)
              !   SSOIL  : Soil heat flux (W m-2)
              !   FLX1   : Precip-snow sfc (W m-2)
              !   FLX2   : Freezing rain latent heat flux (W m-2)
              !   FLX3   : Phase-change heat flux from snowmelt (W m-2)
              !   SNOMLT : Snow melt (m) (water equivalent)
              !   SNCOVR : Fractional snow cover (Fraction, 0.0 to 1.0)
              !   RUNOFF1: Surface runoff (m s-1), not infiltrating the surface
              !   RUNOFF2: Subsurface runoff (m s-1), drainage out bottom of deepest soil layer (baseflow)
              !   RUNOFF3: Numerical trunctation in excess of porosity
              !            for a given soil layer at the end of a time step (m s-1).
              !            NOTE: the above RUNOFF2 is actually the sum of RUNOFF2 and RUNOFF3
              !   RC     : Canopy resistance (s m-1)
              !   PC     : Plant coefficient (Fraction, 0.0 to 1.0) where PC*ETP = Actual transpiration
              !   RCS    : Incoming solar RC factor (dimensionless)
              !   RCT    : Air temperature RC factor (dimensionless)
              !   RCQ    : Atmos vapor pressure deficit RC factor (dimensionless)
              !   RCSOIL : Soil moisture RC factor (dimensionless)
              ! ----------------------------------------------------------------------
              ! 8. DIAGNOSTIC OUTPUT (D):
              ! ----------------------------------------------------------------------
              !   SOILW  : Available soil moisture in root zone (Fraction, between SMCWLT and SMCMAX)
              !   SOILM  : Total soil column moisture content (frozen+unfrozen) (m)
              !   Q1     : Effective specific humidity at surface (kg kg-1), used for
              !            diagnosing the specific humidity at 2 meter for coupled model

#ifdef _HRLDAS_URBAN_
              if ( ucmcall == 1) then
                 call calc_declin( olddate(1:13), latitude(i,j), longitude(i,j), cosz, omg_urb, declin )
              endif
#endif



              snoalb = MAXALB(VEGTYPX) !?????  albed !?????

              CALL SFLX (FFROZP,ICE,DT,ZLVL,NSOIL,SLDPTH,           &    !C
                   .FALSE., LLANDUSE, 'STAS',                       &    !C
                   LWDN,SOLDN,SOLNET,SFCPRS,PRCP,SFCTMP,Q2,SFCSPD,  &    !F
                   COSZ,PRCPRAIN, SOLARDIRECT,                      &    !F
                   TH2,Q2SAT,DQSDT2,                                &    !I
                   VEGTYPX,SOILTYP,SlOPETYP,SHDFAC,SHDMIN,SHDMAX,   &    !S  
                   ALB,SNOALB,TBOT,Z0BRD,Z0,EMISSI, EMBRD,          &    !S   
                   CMCX,T1X,STC1,SMC1,SH2O,SNOWH,SNEQV,ALBEDO,CH,CM,&    !H  
! ----------------------------------------------------------------------         
! OUTPUTS, DIAGNOSTICS, PARAMETERS BELOW GENERALLY NOT NECESSARY WHEN            
! COUPLED WITH E.G. A NWP MODEL (SUCH AS THE NOAA/NWS/NCEP MESOSCALE ETA         
! MODEL).  OTHER APPLICATIONS MAY REQUIRE DIFFERENT OUTPUT VARIABLES.            
! ----------------------------------------------------------------------         
                   ETA,SHEAT, ETA_KINEMATIC,FDOWN,                  &    !O  
                   EC,EDIR,ET,ETT,ESNOW,DRIP,DEW,                   &    !O  
                   BETA,ETP,SSOIL,                                  &    !O  
                   FLX1,FLX2,FLX3,                                  &    !O  
                   SNOMLT,SNCOVR,                                   &    !O  
                   RUNOFF1,RUNOFF2,RUNOFF3,                         &    !O  
                   RC,PC,RSMIN,XLAI,RCS,RCT,RCQ,RCSOIL,             &    !O  
                   SOILW,SOILM,Q1,                                  &    !D
                   SMCWLT,SMCDRY,SMCREF,SMCMAX,NROOT)                    !P  

!---------------------------------------------------------------------
! Begin Converting Data back to grid from 1-d and make units conversions
!---------------------------------------------------------------------


!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!     CONVERT ACTUAL EVAPOTRANSPIRATION VALUE FROM  KG M-2 S-1  TO
!     W M-2  (BY MULTIPLYING BY 2.501E+6) FOR USE IN ENERGY BALANCE.
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!KWM              T14 = T1X * T1X * T1X * T1X
!KWM              AET = ETA
!KWM              FUP = 5.67E-8 * T14
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!     CALCULATE RESIDUAL OF ALL SURFACE ENERGY BALANCE EQN TERMS.
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!KWM              S = -S
!FC              RESX(I,J) = F - H - S - AET - FUP - FLX1 - FLX2 - FLX3
!KWM              SNOFLX=FLX3
!  Convert ETA and ETP from W M-2 to KG M-2 S-1

!KWM              ETA = ETA/2.501E+6
              ETP = ETP/2.501E+6  

              QFX(I,J) = (EDIR+EC+ETT) + ESNOW ! in W m{-2}

!KWM Overwrite ETA for now
              ETA = ((EDIR+EC+ETT)/2.501E+6) + (ESNOW/2.83E6)

! Fill output variable arrays and doing output

              STC(I,J,1:NSOIL)=STC1(1:NSOIL)    ! Updated Soil Temperature
              SMC(I,J,1:NSOIL)=SMC1(1:NSOIL)    ! Updated Soil Moisture
              SH2OX(I,J,1:NSOIL)=SH2O(1:NSOIL)  ! Updated Soil Liquid Water
              SNODEP(I,J)=SNOWH                 ! Updated snow depth !KWM
              WEASD(I,J)=SNEQV      ! ( m of water )
              FX(I,J)=FDOWN
              ETPX(I,J)=ETP
              T1(I,J)=T1X
              ETAX(I,J)=ETAX(I,J)+ETA*dt ! kg m{-2} s{-1} to mm liquid
              ETAKIN(I,J)=ETA_KINEMATIC*dt
              CMC(I,J)=CMCX
              GRDFLX(I,J)=SSOIL
              CHX(I,J)=CH
              ZNT(I,J)=Z0
              RUNOFF1X(I,J)=RUNOFF1X(I,J)+RUNOFF1*dt*1.E3
              RUNOFF2X(I,J)=RUNOFF2X(I,J)+RUNOFF2*dt*1.E3
              RUNOFF3X(I,J)=RUNOFF3X(I,J)+RUNOFF3*dt*1.E3
              EDIRX(I,J)=EDIRX(I,J)+(EDIR/2.501E6)*dt ! (W m{-2} to kg m{-2} s{-2} to mm)
              ECX(I,J)=ECX(I,J)+(EC/2.501E6)*dt       ! (W m{-2} to kg m{-2} s{-2} to mm)
              ETTX(I,J)=ETTX(I,J)+(ETT/2.501E6)*dt    ! (W m{-2} to kg m{-2} s{-2} to mm)
!KWM              ETPNDX(I,J)=ETPND
!KWM              SNMAXX(I,J)=SNMAX
!KWM              SNOFLXX(I,J)=SNOFLX
!FC             SNOEVPX(I,J)=SNOEVP
              RCX(I,J)=RC
              HX(I,J)=SHEAT
              SMCMAX1(I,J)=SMCMAX
              SMCREF1(I,J)=SMCREF
              SMCWLT1(I,J)=SMCWLT
              ALBEDX(I,J)=ALBEDO
              ACRAIN(I,J)=ACRAIN(I,J)+PRCP*dt      ! (mm/s to mm)
              ACSNOM(I,J)=ACSNOM(I,J)+snomlt*1.E3  ! Accumulated snow melt (convert m to mm).
              ! ESNOW2D:  Accumulated snow sublimation (converted from W m{-2} to kg m{-2} s{-2} to mm)
              ESNOW2D(I,J)=ESNOW2D(I,J)+(ESNOW/2.83E6)*dt
              DRIP2D(I,J)=DRIP2D(I,J)+DRIP*1.E3   ! convert m to mm
              DEWFALL(I,J)=DEWFALL(I,J)+DEW*1.E3  ! convert m to mm
              SOILMX(I,J)= SOILM*1.E3             ! convert m to mm
              EMISS(I,J) = EMISSI
              XLAI2D(I,J) = XLAI

!DG Convert from point to grid (units of SFHEAD & INFXS (mm)
!DG Temp assignment for parking lot runoff
!KWM              SFCHEAD(I,J)=SFHEAD
!DG       INFXS1(I,J)=PRCP1(I,J)
!KWM              INFXS1(I,J)=INFXS
! End temp assignments
!KWM              PDDUM2(I,J)=PDDUM
!KWM              PCPDRP(I,J)=DRIP
!KWM              SFCWATR2(I,J)=SFCWATR

#ifdef _HRLDAS_URBAN_
              IF (UCMCALL == 1 ) THEN                                              ! Beginning of UCM CALL if block
!--------------------------------------
! URBAN CANOPY MODEL START - urban
!--------------------------------------
! Input variables lsm --> urban


                 IF ( ( VEGTYP(I,J) == ISURBAN  ) .or. ( VEGTYP(I,J) == 31 ) .or. &
                      ( VEGTYP(I,J) == 32 ) .or. ( VEGTYP(I,J) == 33 ) ) THEN
! Call urban

!
                    UTYPE_URB = UTYPE_URB2D(I,J) !urban type (low, high or industrial) 
                    TA_URB    = SFCTMP           ! [K]
                    QA_URB    = Q2X(I,J)         ! [kg/kg]
                    UA_URB    = SQRT(U(I,J)**2.+V(I,J)**2.)
                    U1_URB    = U(I,J)
                    V1_URB    = V(I,J)
                    IF(UA_URB < 1.) UA_URB=1.    ! [m/s]
                    SSG_URB   = SOLDN            ! [W/m/m]
                    SSGD_URB  = 0.8*SOLDN        ! [W/m/m]
                    SSGQ_URB  = SSG_URB-SSGD_URB ! [W/m/m]
                    LLG_URB   = LWDN             ! [W/m/m]
!KWM RAIN_URB  = RAINBL(I,J)      ! [mm]
                    RAIN_URB  = PRCP*dt          ! [mm]  !KWM
                    RHOO_URB  = SFCPRS / (287.04 * SFCTMP * (1.0+ 0.61 * Q2X(I,J))) ![kg/m/m/m]
                    ZA_URB    = ZLVL_URBAN       ! [m]
                    DELT_URB  = DT               ! [sec]
                    XLAT_URB  = LATITUDE(I,J)  ! [deg]
                    ZNT_URB   = ZNT(I,J)

                    LSOLAR_URB = .FALSE.

                    TR_URB = TR_URB2D(I,J)
                    TB_URB = TB_URB2D(I,J)
                    TG_URB = TG_URB2D(I,J)
                    TC_URB = TC_URB2D(I,J)
                    QC_URB = QC_URB2D(I,J)
                    UC_URB = UC_URB2D(I,J)

                    DO KROOF = 1,num_roof_layers
                       TRL_URB(KROOF) = TRL_URB3D(I,J,KROOF)
                    END DO
                    DO KWALL = 1,num_wall_layers
                       TBL_URB(KWALL) = TBL_URB3D(I,J,KWALL)
                    END DO
                    DO KROAD = 1,num_road_layers
                       TGL_URB(KROAD) = TGL_URB3D(I,J,KROAD)
                    END DO

                    XXXR_URB = XXXR_URB2D(I,J)
                    XXXB_URB = XXXB_URB2D(I,J)
                    XXXG_URB = XXXG_URB2D(I,J)
                    XXXC_URB = XXXC_URB2D(I,J)
!
!KWM            CHS_URB  = CHS(I,J)
!KWM            CHS2_URB = CHS2(I,J)
                    CHS_URB  = HS !KWM ?????
                    CHS2_URB = HS !KWM ?????
!

                    IF ( .FALSE. ) THEN

                       print*, 'BEFORE CALL URBAN'
                       print*,'num_roof_layers',num_roof_layers, 'num_wall_layers',  &
                            num_wall_layers,                                             &
                            'DZR',DZR,'DZB',DZB,'DZG',DZG,'UTYPE_URB',UTYPE_URB,'TA_URB', &
                            TA_URB,                                                      &
                            'QA_URB',QA_URB,'UA_URB',UA_URB,'U1_URB',U1_URB,'V1_URB',    &
                            V1_URB,                                                     &
                            'SSG_URB',SSG_URB,'SSGD_URB',SSGD_URB,'SSGQ_URB',SSGQ_URB,  &
                            'LLG_URB',LLG_URB,'RAIN_URB',RAIN_URB,'RHOO_URB',RHOO_URB,   &
                            'ZA_URB',ZA_URB, 'DECLIN',DECLIN,'COSZ',COSZ,                &
                            'OMG_URB',OMG_URB,'XLAT_URB',XLAT_URB,'DELT_URB',DELT_URB,   &
                            'ZNT_URB',ZNT_URB,'TR_URB',TR_URB, 'TB_URB',TB_URB,'TG_URB',&
                            TG_URB,'TC_URB',TC_URB,'QC_URB',QC_URB,'TRL_URB',TRL_URB,   &
                            'TBL_URB',TBL_URB,'TGL_URB',TGL_URB,'XXXR_URB',XXXR_URB,   &
                            'XXXB_URB',XXXB_URB,'XXXG_URB',XXXG_URB,'XXXC_URB',XXXC_URB,&
                            'QS_URB',QS_URB,'SH_URB',SH_URB,'LH_URB',   &
                            LH_URB, 'LH_KINEMATIC_URB',LH_KINEMATIC_URB,'SW_URB',SW_URB,&
                            'ALB_URB',ALB_URB,'LW_URB',LW_URB,'G_URB',G_URB,'RN_URB',   &
                            RN_URB, 'PSIM_URB',PSIM_URB,'PSIH_URB',PSIH_URB,          &
                            'U10_URB',U10_URB,'V10_URB',V10_URB,'TH2_URB',TH2_URB,      &
                            'Q2_URB',Q2_URB,'CHS_URB',CHS_URB,'CHS2_URB',CHS2_URB
                    endif



! Call urban

                    CALL urban(LSOLAR_URB,                                      & ! I
                         num_roof_layers,num_wall_layers,num_road_layers, & ! C
                         DZR,DZB,DZG,                                     & ! C
                         UTYPE_URB,TA_URB,QA_URB,UA_URB,U1_URB,V1_URB,SSG_URB, & ! I
                         SSGD_URB,SSGQ_URB,LLG_URB,RAIN_URB,RHOO_URB,     & ! I
                         ZA_URB,DECLIN,COSZ,OMG_URB,                      & ! I
                         XLAT_URB,DELT_URB,ZNT_URB,                       & ! I
                         CHS_URB, CHS2_URB,                               & ! I
                         TR_URB, TB_URB, TG_URB, TC_URB, QC_URB,UC_URB,   & ! H
                         TRL_URB,TBL_URB,TGL_URB,                         & ! H
                         XXXR_URB, XXXB_URB, XXXG_URB, XXXC_URB,          & ! H
                         TS_URB,QS_URB,SH_URB,LH_URB,LH_KINEMATIC_URB,    & ! O
                         SW_URB,ALB_URB,LW_URB,G_URB,RN_URB,PSIM_URB,PSIH_URB, & ! O
                         GZ1OZ0_URB,                                      & !O
                         U10_URB, V10_URB, TH2_URB, Q2_URB,               & ! O
                         UST_URB)                                           !O


                    IF ( .FALSE. ) THEN

                       print*, 'AFTER CALL URBAN'
                       print*,'num_roof_layers',num_roof_layers, 'num_wall_layers',  &
                            num_wall_layers,                                             &
                            'DZR',DZR,'DZB',DZB,'DZG',DZG,'UTYPE_URB',UTYPE_URB,'TA_URB', &
                            TA_URB,                                                      &
                            'QA_URB',QA_URB,'UA_URB',UA_URB,'U1_URB',U1_URB,'V1_URB',    &
                            V1_URB,                                                     &
                            'SSG_URB',SSG_URB,'SSGD_URB',SSGD_URB,'SSGQ_URB',SSGQ_URB,  &
                            'LLG_URB',LLG_URB,'RAIN_URB',RAIN_URB,'RHOO_URB',RHOO_URB,   &
                            'ZA_URB',ZA_URB, 'DECLIN',DECLIN,'COSZ',COSZ,                &
                            'OMG_URB',OMG_URB,'XLAT_URB',XLAT_URB,'DELT_URB',DELT_URB,   &
                            'ZNT_URB',ZNT_URB,'TR_URB',TR_URB, 'TB_URB',TB_URB,'TG_URB',&
                            TG_URB,'TC_URB',TC_URB,'QC_URB',QC_URB,'TRL_URB',TRL_URB,   &
                            'TBL_URB',TBL_URB,'TGL_URB',TGL_URB,'XXXR_URB',XXXR_URB,   &
                            'XXXB_URB',XXXB_URB,'XXXG_URB',XXXG_URB,'XXXC_URB',XXXC_URB,&
                            'TS_URB',TS_URB,'QS_URB',QS_URB,'SH_URB',SH_URB,'LH_URB',   &
                            LH_URB, 'LH_KINEMATIC_URB',LH_KINEMATIC_URB,'SW_URB',SW_URB,&
                            'ALB_URB',ALB_URB,'LW_URB',LW_URB,'G_URB',G_URB,'RN_URB',   &
                            RN_URB, 'PSIM_URB',PSIM_URB,'PSIH_URB',PSIH_URB,          &
                            'U10_URB',U10_URB,'V10_URB',V10_URB,'TH2_URB',TH2_URB,      &
                            'Q2_URB',Q2_URB,'CHS_URB',CHS_URB,'CHS2_URB',CHS2_URB
                    endif

! Overwrite fluxes and surface temperature
!m

!m            ALBEDO(I,J) = FRC_URB*ALB_URB+FRC_NAT*ALBEDOK   ![-]
!m            HFX(I,J) = FRC_URB*SH_URB+FRC_NAT*SHEAT         ![W/m/m]
!m            QFX(I,J) = FRC_URB*LH_KINEMATIC_URB &
!m                     + FRC_NAT*ETA_KINEMATIC                ![kg/m/m/s]
!m            LH(I,J) = FRC_URB*LH_URB+FRC_NAT*ETA            ![W/m/m]
!m            GRDFLX(I,J) = FRC_URB*G_URB+FRC_NAT*SSOIL       ![W/m/m]
!m            TSK(I,J) = FRC_URB*TS_URB+FRC_NAT*T1            ![K]
!m            QSFC(I,J)= FRC_URB*QS_URB+FRC_NAT*Q1            ![-]

!
                    TS_URB2D(I,J) = TS_URB
!

                    ALBEDX(I,J) = FRC_URB2D(I,J)*ALB_URB+(1-FRC_URB2D(I,J))*ALBEDO    ![-]
                    HX(I,J) = FRC_URB2D(I,J)*SH_URB+(1-FRC_URB2D(I,J))*SHEAT         ![W/m/m]
                    ETAKIN(I,J) = FRC_URB2D(I,J)*LH_KINEMATIC_URB &
                         + (1-FRC_URB2D(I,J))*ETA_KINEMATIC                ![kg/m/m/s]
                    QFX(I,J) = FRC_URB2D(I,J)*LH_URB+(1-FRC_URB2D(I,J))*ETA            ![W/m/m]
                    GRDFLX(I,J) = FRC_URB2D(I,J)*G_URB+(1-FRC_URB2D(I,J))*SSOIL       ![W/m/m]
                    T1(I,J) = FRC_URB2D(I,J)*TS_URB+(1-FRC_URB2D(I,J))*T1X            ![K]
                    Q2X(I,J)= FRC_URB2D(I,J)*QS_URB+(1-FRC_URB2D(I,J))*Q1            ![-]

                    IF(.FALSE.)THEN

                       print*, ' FRC_URB2D', FRC_URB2D,                        &
                            'ALB_URB',ALB_URB, 'ALBEDO',ALBEDO, &
                            'ALBEDX(I,J)',  ALBEDX(I,J),                  &
                            'SH_URB',SH_URB,'SHEAT',SHEAT, 'HX(I,J)',HX(I,J),  &
                            'LH_KINEMATIC_URB',LH_KINEMATIC_URB,'ETA_KINEMATIC',  &
                            ETA_KINEMATIC, 'QFX(I,J)',QFX(I,J),                  &
                            'LH_URB',LH_URB, 'ETA',ETA, 'ETAKIN(I,J)',ETAKIN(I,J),        &
                            'G_URB',G_URB,'SSOIL',SSOIL,'GRDFLX(I,J)', GRDFLX(I,J),&
                            'TS_URB',TS_URB,'T1X',T1X,'T1(I,J)',T1(I,J),          &
                            'QS_URB',QS_URB,'Q1',Q1,'Q2X(I,J)',Q2X(I,J) 
                    endif




! Renew Urban State Varialbes

                    TR_URB2D(I,J) = TR_URB
                    TB_URB2D(I,J) = TB_URB
                    TG_URB2D(I,J) = TG_URB
                    TC_URB2D(I,J) = TC_URB
                    QC_URB2D(I,J) = QC_URB
                    UC_URB2D(I,J) = UC_URB

                    DO KROOF = 1,num_roof_layers
                       TRL_URB3D(I,J,KROOF) = TRL_URB(KROOF)
                    END DO
                    DO KWALL = 1,num_wall_layers
                       TBL_URB3D(I,J,KWALL) = TBL_URB(KWALL)
                    END DO
                    DO KROAD = 1,num_road_layers
                       TGL_URB3D(I,J,KROAD) = TGL_URB(KROAD)
                    END DO
                    XXXR_URB2D(I,J) = XXXR_URB
                    XXXB_URB2D(I,J) = XXXB_URB
                    XXXG_URB2D(I,J) = XXXG_URB
                    XXXC_URB2D(I,J) = XXXC_URB

                    SH_URB2D(I,J)    = SH_URB
                    LH_URB2D(I,J)    = LH_URB
                    G_URB2D(I,J)     = G_URB
                    RN_URB2D(I,J)    = RN_URB
                    PSIM_URB2D(I,J)  = PSIM_URB
                    PSIH_URB2D(I,J)  = PSIH_URB
                    GZ1OZ0_URB2D(I,J)= GZ1OZ0_URB
                    U10_URB2D(I,J)   = U10_URB
                    V10_URB2D(I,J)   = V10_URB
                    TH2_URB2D(I,J)   = TH2_URB
                    Q2_URB2D(I,J)    = Q2_URB
                    UST_URB2D(I,J)   = UST_URB
!
                    AKMS_URB2D(I,J)  = KARMAN * UST_URB2D(I,J)/(GZ1OZ0_URB2D(I,J)-PSIM_URB2D(I,J)) 

                 else                                          !KWM   ?????
                    do kroof = 1, num_roof_Layers             !KWM   ?????
                       TRL_URB3D(i,j,kroof) = STC(i,j,kroof)  !KWM   ?????
                       TBL_URB3D(i,j,kroof) = STC(i,j,kroof)  !KWM   ?????
                       TGL_URB3D(i,j,kroof) = STC(i,j,kroof)  !KWM   ?????
                    enddo                                     !KWM   ?????
                    TR_URB2D(i,j) = STC(i,j,1)                !KWM   ?????
                    TB_URB2D(i,j) = STC(i,j,1)                !KWM   ?????
                    TG_URB2D(i,j) = STC(i,j,1)                !KWM   ?????
                    TC_URB2D(i,j) = STC(i,j,1)                !KWM   ?????

                 END IF

              ENDIF                                   ! end of UCM CALL if block
!--------------------------------------
! Urban Part End - urban
!--------------------------------------
#endif

! ***  endif of the land-point
           ENDIF

        ENDDO ILOOP
     ENDDO JLOOP
!-------------------------------------------------------------------
! END of 1-D NOAH processing
!-------------------------------------------------------------------

!------------------------------------------------------------------------
! Call to Output thermodynamic variables
!------------------------------------------------------------------------
!KWM          CALL PRTHM(K,IX,JX,MDT,MINUTE,NSOIL,CHX,FX,GRDFLX,                &
!KWM         CMC,T2,T1,ETAX,ETPX,PRCP1,STC,SMC,DT,HX,RESX)

!------------------------------------------------------------------------
! Call to Output hydrological variables
!------------------------------------------------------------------------
!KWM          CALL PRTHYD(K,IX,JX,NSOIL,MDT,MINUTE,ETAX,ETPX,PRCP1,             &
!KWM        SMC,ZSOIL,RUNOFF1X,RUNOFF2X,RUNOFF3X,DT,EDIRX,ECX,                &
!KWM	ETTX,CMC,SLDPTH,SNOEVPX,SH2OX,ESDX,SNODEP,VEGTYP) 
!FC        QBDRYT,QSTRMVOLT,QBDRY,QSTRMVOL,SFCHEAD,INFXS1,DH,                &
!FC        CH_NET,PDDUM2,PCPDRP,SFCWATR2,QLAKVOL,QLAKVOLT)

! Output for history
     if (output_timestep > 0) then
        if (mod((k-1)*noah_timestep, output_timestep) == 0) then
           call output_hrldas(ounit, version, igrid, output_timestep, llanduse, split_output_count, hgrid, &
                subwindow_xend-subwindow_xstart+1, subwindow_yend-subwindow_ystart+1, &
                iswater, mapproj, lat1, lon1, dx, dy, truelat1, truelat2, cen_lon, &
                nsoil, sldpth, startdate, olddate, &
                vegtyp, soltyp, t1, cmc, stc, smc, sh2ox, soilmx, runoff1x, runoff2x, &
                runoff3x, etax, etakin, ecx, edirx, ettx, albedx, weasd, acrain, acsnom, &
                esnow2d, drip2d, dewfall, snodep, fpar, znt, hx, qfx, grdflx, short, xlong, &
#ifdef _HRLDAS_URBAN_
                ucmcall, & 
           TRL_URB3D, TBL_URB3D, TGL_URB3D, TR_URB2D, TB_URB2D, TG_URB2D, TC_URB2D, &
#endif
           fx, xlai2d)
        endif
     endif

     if (vegtyp(subwindow_xstart,subwindow_ystart)==ISWATER) then
        write(*,'(" ***DATE=", A19)', advance="NO") olddate
     else
     write(*,'(" ***DATE=", A19, 6F10.5)', advance="NO") olddate, &
          stc(subwindow_xstart,subwindow_ystart,1), xlai2d(subwindow_xstart,subwindow_ystart)
     endif


!------------------------------------------------------------------------
! Update the time 
!------------------------------------------------------------------------

     call geth_newdate(newdate, olddate, nint(dt))
     olddate = newdate

!------------------------------------------------------------------------
! End of Time Loop  (do K) 
!------------------------------------------------------------------------

     ! update the timer
     call system_clock(count=clock_count_2, count_rate=clock_rate)
     timing_sum = timing_sum + float(clock_count_2-clock_count_1)/float(clock_rate)
     write(*,'("    Timing: ",f6.2," Cumulative:  ", f10.2 )') &
          float(clock_count_2-clock_count_1)/float(clock_rate), &
          timing_sum
     clock_count_1 = clock_count_2

  ENDDO KLOOP

END program Noah_hrldas_driver

subroutine wrf_message(msg)
  implicit none
  character(len=*), intent(in) :: msg
  print*, msg
end subroutine wrf_message
logical function wrf_dm_on_monitor() result(l)
  l = .TRUE.
  return
end function wrf_dm_on_monitor



!---------------------------------------------------------------------
SUBROUTINE calc_declin ( nowdate, latitude, longitude, cosz, hrang, declin )
  use module_date_utilities
!---------------------------------------------------------------------
   IMPLICIT NONE
!---------------------------------------------------------------------

   REAL, PARAMETER :: DEGRAD = 3.14159265/180.
   REAL, PARAMETER :: DPD    = 360./365.
! !ARGUMENTS:
   character(len=13), intent(in)  :: nowdate
   real,              intent(in)  :: latitude
   real,              intent(in)  :: longitude
   REAL,              intent(out) :: cosz
   REAL,              intent(out) :: hrang
   real,              intent(out) :: DECLIN
   REAl                           :: JULIAN
   REAL                           :: OBECL
   REAL                           :: SINOB
   REAL                           :: SXLONG
   REAL                           :: ARG
   real                           :: tloctim
   integer                        :: iday
   integer                        :: ihour

   call geth_idts(nowdate(1:10), nowdate(1:4)//"-01-01", iday)
   read(nowdate(12:13), *) ihour
   julian = real(iday) + real(ihour)/24.
!
! for short wave radiation

   DECLIN=0.

!-----OBECL : OBLIQUITY = 23.5 DEGREE.
        
   OBECL=23.5*DEGRAD
   SINOB=SIN(OBECL)
        
!-----CALCULATE LONGITUDE OF THE SUN FROM VERNAL EQUINOX:
        
   IF(JULIAN.GE.80.)SXLONG=DPD*(JULIAN-80.)*DEGRAD
   IF(JULIAN.LT.80.)SXLONG=DPD*(JULIAN+285.)*DEGRAD
   ARG=SINOB*SIN(SXLONG)
   DECLIN=ASIN(ARG)

   TLOCTIM = IHOUR + LONGITUDE/15. ! Local time in hours
   tloctim = AMOD(tloctim+24.0, 24.0)
   HRANG=15.*(TLOCTIM-12.)*DEGRAD
   COSZ=SIN(LATITUDE*DEGRAD)*SIN(DECLIN)+COS(LATITUDE*DEGRAD)*COS(DECLIN)*COS(HRANG)

!KWM   write(wrf_err_message,10)DECDEG/DEGRAD
!KWM10 FORMAT(1X,'*** SOLAR DECLINATION ANGLE = ',F6.2,' DEGREES.',' ***')
!KWM   CALL wrf_debug (50, wrf_err_message)

 END SUBROUTINE calc_declin


!-----------------------------------------------------------------
 SUBROUTINE SOIL_VEG_GEN_PARM( MMINLU, MMINSL)
   ! Imported from the WRF code
!-----------------------------------------------------------------
   USE module_sf_noahlsm

   IMPLICIT NONE

   integer :: LUMATCH, IINDEX, LC, NUM_SLOPE
   integer :: ierr
   INTEGER , PARAMETER :: OPEN_OK = 0

   character*4 :: MMINLU, MMINSL
   character*128 :: mess , message
   logical, external :: wrf_dm_on_monitor


!-----SPECIFY VEGETATION RELATED CHARACTERISTICS :
!             ALBBCK: SFC albedo (in percentage)
!                 Z0: Roughness length (m)
!             SHDFAC: Green vegetation fraction (in percentage)
!  Note: The ALBEDO, Z0, and SHDFAC values read from the following table
!          ALBEDO, amd Z0 are specified in LAND-USE TABLE; and SHDFAC is
!          the monthly green vegetation data
!             CMXTBL: MAX CNPY Capacity (m)
!             NROTBL: Rooting depth (layer)
!              RSMIN: Mimimum stomatal resistance (s m-1)
!              RSMAX: Max. stomatal resistance (s m-1)
!                RGL: Parameters used in radiation stress function
!                 HS: Parameter used in vapor pressure deficit functio
!               TOPT: Optimum transpiration air temperature. (K)
!             CMCMAX: Maximum canopy water capacity
!             CFACTR: Parameter used in the canopy inteception calculati
!               SNUP: Threshold snow depth (in water equivalent m) that
!                     implies 100% snow cover
!                LAI: Leaf area index (dimensionless)
!             MAXALB: Upper bound on maximum albedo over deep snow
!
!-----READ IN VEGETAION PROPERTIES FROM VEGPARM.TBL
!


#if _HRLDAS_OFFLINE_
   ! Modis flexibility
   OPEN(19, FILE=MMINLU//'.TBL',FORM='FORMATTED',STATUS='OLD',IOSTAT=ierr)
   IF(ierr .NE. OPEN_OK ) THEN
      WRITE(message,FMT='(A)') &
           'module_sf_noahlsm.F: soil_veg_gen_parm: failure opening '//MMINLU//'.TBL'
      CALL wrf_error_fatal ( message )
   END IF
#else
   OPEN(19, FILE='VEGPARM.TBL',FORM='FORMATTED',STATUS='OLD',IOSTAT=ierr)
   IF(ierr .NE. OPEN_OK ) THEN
      WRITE(message,FMT='(A)') &
           'module_sf_noahlsm.F: soil_veg_gen_parm: failure opening VEGPARM.TBL'
      CALL wrf_error_fatal ( message )
   END IF
#endif

   WRITE ( mess, * ) 'INPUT LANDUSE = ',MMINLU
   CALL wrf_message( mess )

   LUMATCH=0

   READ (19,*)
   READ (19,2000,END=2002)LUTYPE
   READ (19,*)LUCATS,IINDEX
2000 FORMAT (A4)

   IF(LUTYPE.EQ.MMINLU)THEN
      WRITE( mess , * ) 'LANDUSE TYPE = ',LUTYPE,' FOUND',           &
           LUCATS,' CATEGORIES'
      CALL wrf_message( mess )
      LUMATCH=1
   ENDIF

   IF(LUTYPE.EQ.MMINLU)THEN
      DO LC=1,LUCATS
#ifdef _HRLDAS_OFFLINE_
         READ (19,*)IINDEX,ALBMINTBL(LC),ALBMAXTBL(LC),Z0MINTBL(LC),Z0MAXTBL(LC),SHDTBL(LC),   &
              NROTBL(LC),RSTBL(LC),RGLTBL(LC),HSTBL(LC), &
              SNUPTBL(LC),LAIMINTBL(LC),LAIMAXTBL(LC),MAXALB(LC),EMIMINTBL(LC),EMIMAXTBL(LC)
#else
         READ (19,*)IINDEX,ALBTBL(LC),Z0TBL(LC),SHDTBL(LC),   &
              NROTBL(LC),RSTBL(LC),RGLTBL(LC),HSTBL(LC), &
              SNUPTBL(LC),LAITBL(LC),MAXALB(LC)
#endif
#ifdef _HRLDAS_OFFLINE_
              MAXALB(LC) = MAXALB(LC)*0.01
#endif

      ENDDO
!
      READ (19,*)
      READ (19,*)TOPT_DATA
      READ (19,*)
      READ (19,*)CMCMAX_DATA
      READ (19,*)
      READ (19,*)CFACTR_DATA
      READ (19,*)
      READ (19,*)RSMAX_DATA
      READ (19,*)
      READ (19,*)BARE
   ENDIF
!
2002 CONTINUE

   CLOSE (19)

!
!-----READ IN SOIL PROPERTIES FROM SOILPARM.TBL
!
   OPEN(19, FILE='SOILPARM.TBL',FORM='FORMATTED',STATUS='OLD',IOSTAT=ierr)
   IF(ierr .NE. OPEN_OK ) THEN
      WRITE(message,FMT='(A)') &
           'module_sf_noahlsm.F: soil_veg_gen_parm: failure opening SOILPARM.TBL'
      CALL wrf_error_fatal ( message )
   END IF

   MMINSL='STAS'                       !oct2
   WRITE(mess,*) 'INPUT SOIL TEXTURE CLASSIFICAION = ',MMINSL
   CALL wrf_message( mess )

   LUMATCH=0

   READ (19,*)
   READ (19,2000,END=2003)SLTYPE
   READ (19,*)SLCATS,IINDEX
   IF(SLTYPE.EQ.MMINSL)THEN
      WRITE( mess , * ) 'SOIL TEXTURE CLASSIFICATION = ',SLTYPE,' FOUND', &
           SLCATS,' CATEGORIES'
      CALL wrf_message ( mess )
      LUMATCH=1
   ENDIF
   IF(SLTYPE.EQ.MMINSL)THEN
      DO LC=1,SLCATS
         READ (19,*) IINDEX,BB(LC),DRYSMC(LC),F11(LC),MAXSMC(LC),&
              REFSMC(LC),SATPSI(LC),SATDK(LC), SATDW(LC),   &
              WLTSMC(LC), QTZ(LC)
      ENDDO
   ENDIF

2003 CONTINUE

   CLOSE (19)

   IF(LUMATCH.EQ.0)THEN
      CALL wrf_message( 'SOIl TEXTURE IN INPUT FILE DOES NOT ' )
      CALL wrf_message( 'MATCH SOILPARM TABLE'                 )
      CALL wrf_error_fatal ( 'INCONSISTENT OR MISSING SOILPARM FILE' )
   ENDIF

!
!-----READ IN GENERAL PARAMETERS FROM GENPARM.TBL
!
   OPEN(19, FILE='GENPARM.TBL',FORM='FORMATTED',STATUS='OLD',IOSTAT=ierr)
   IF(ierr .NE. OPEN_OK ) THEN
      WRITE(message,FMT='(A)') &
           'module_sf_noahlsm.F: soil_veg_gen_parm: failure opening GENPARM.TBL'
      CALL wrf_error_fatal ( message )
   END IF

   READ (19,*)
   READ (19,*)
   READ (19,*) NUM_SLOPE

   SLPCATS=NUM_SLOPE

   DO LC=1,SLPCATS
      READ (19,*)SLOPE_DATA(LC)
   ENDDO

   READ (19,*)
   READ (19,*)SBETA_DATA
   READ (19,*)
   READ (19,*)FXEXP_DATA
   READ (19,*)
   READ (19,*)CSOIL_DATA
   READ (19,*)
   READ (19,*)SALP_DATA
   READ (19,*)
   READ (19,*)REFDK_DATA
   READ (19,*)
   READ (19,*)REFKDT_DATA
   READ (19,*)
   READ (19,*)FRZK_DATA
   READ (19,*)
   READ (19,*)ZBOT_DATA
   READ (19,*)
   READ (19,*)CZIL_DATA
   READ (19,*)
   READ (19,*)SMLOW_DATA
   READ (19,*)
   READ (19,*)SMHIGH_DATA
   CLOSE (19)


!-----------------------------------------------------------------
 END SUBROUTINE SOIL_VEG_GEN_PARM
!-----------------------------------------------------------------
