module module_hrldas_netcdf_io
  use module_date_utilities
  use netcdf
  implicit none

  logical, parameter :: FATAL = .TRUE.
  logical, parameter :: NOT_FATAL = .FALSE.

  type inputstruct
     character(len=19)             :: read_date
     real, pointer, dimension(:,:) :: t
     real, pointer, dimension(:,:) :: q
     real, pointer, dimension(:,:) :: u
     real, pointer, dimension(:,:) :: v
     real, pointer, dimension(:,:) :: p
     real, pointer, dimension(:,:) :: lw
     real, pointer, dimension(:,:) :: sw
     real, pointer, dimension(:,:) :: pcp
     real, pointer, dimension(:,:) :: fpar
     real, pointer, dimension(:,:) :: lai
  end type inputstruct

contains

  subroutine read_hrldas_hdrinfo(wrfinput_flnm, ix, jx, &
       xstart, xend, ystart, yend,                      &
       iswater, isurban, llanduse, dx, dy, truelat1, truelat2, cen_lon, lat1, lon1, &
       igrid, mapproj)
    ! Return the dimensions of the grid and some map information.
    implicit none
    character(len=*), intent(in)    :: wrfinput_flnm
    integer,          intent(out)   :: mapproj
    integer,          intent(out)   :: igrid
    integer,          intent(out)   :: ix, jx    ! dimensions
    integer,          intent(inout) :: xstart, xend, ystart, yend ! Subwindow definition
    integer,          intent(out)   :: iswater   ! vegetation category corresponding to water bodies
    integer,          intent(out)   :: isurban   ! vegetation category corresponding to urban areas
    character(len=4), intent(out)   :: llanduse  ! Landuse dataset (USGS or MODI)
    real,             intent(out)   :: dx
    real,             intent(out)   :: dy
    real,             intent(out)   :: truelat1
    real,             intent(out)   :: truelat2
    real,             intent(out)   :: cen_lon
    real,             intent(out)   :: lat1
    real,             intent(out)   :: lon1

    integer :: ncid, dimid, varid, ierr
    real, allocatable, dimension(:,:) :: dum2d
    character(len=256) :: units

    ! Open the NetCDF file.
    write(*,'("wrfinput_flnm: ''", A, "''")') trim(wrfinput_flnm)
    ierr = nf90_open(wrfinput_flnm, NF90_NOWRITE, ncid)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO: Problem opening wrfinput file: "//trim(wrfinput_flnm))

    ierr = nf90_inq_dimid(ncid, "west_east", dimid)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding dimension 'west_east'")

    ierr = nf90_inquire_dimension(ncid, dimid, len=ix)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding dimension length for 'west_east'")

    ierr = nf90_inq_dimid(ncid, "south_north", dimid)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding dimension 'south_north'")

    ierr = nf90_inquire_dimension(ncid, dimid, len=jx)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding dimension length for 'south_north'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "DX", dx)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'DX'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "DY", dy)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'DY'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "TRUELAT1", truelat1)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'TRUELAT1'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "TRUELAT2", truelat2)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'TRUELAT2'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "STAND_LON", cen_lon)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'STAND_LON'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "MAP_PROJ", mapproj)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'MAP_PROJ'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "GRID_ID", igrid)
    if (ierr /= 0) then
       ierr = nf90_get_att(ncid, NF90_GLOBAL, "grid_id", igrid)
       call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'GRID_ID' or 'grid_id'")
    endif

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "ISWATER", iswater)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'ISWATER'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "ISURBAN", isurban)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'ISURBAN'")

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "MMINLU", llanduse)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems finding global attribute 'MMINLU'")

    if (xend == 0) then
       xend = ix
    endif
    if (yend == 0) then
       yend = jx
    endif


! This section is for reading the information from the wrfinput file
    allocate(dum2d(xstart:xend,ystart:yend))
    call get_2d_netcdf("XLAT", ncid, dum2d,  units, xstart, xend, ystart, yend, FATAL, ierr)
    lat1 = dum2d(xstart,ystart)

    call get_2d_netcdf("XLONG", ncid, dum2d,  units, xstart, xend, ystart, yend, FATAL, ierr)
    lon1 = dum2d(xstart,ystart)
    deallocate (dum2d)

    ierr = nf90_close(ncid)
    call error_handler(ierr, failure="READ_HRLDAS_HDRINFO:  Problems closing NetCDF file.")

  end subroutine read_hrldas_hdrinfo

  subroutine readland_hrldas(wrfinput_flnm,                            &
       xstart, xend,                                                   &
       ystart, yend,                                                   &
       iswater, vegtyp, soltyp, terrain, tbot_2d, latitude, longitude)
    implicit none
    character(len=*),          intent(in)  :: wrfinput_flnm
    integer,                   intent(in)  :: xstart, xend, ystart, yend
    integer,                   intent(in)  :: iswater
    integer, dimension(xstart:xend,ystart:yend), intent(out) :: vegtyp, soltyp
    real,    dimension(xstart:xend,ystart:yend), intent(out) :: terrain
    real,    dimension(xstart:xend,ystart:yend), intent(out) :: tbot_2d
    real,    dimension(xstart:xend,ystart:yend), intent(out) :: latitude
    real,    dimension(xstart:xend,ystart:yend), intent(out) :: longitude

    character(len=256) :: units
    integer :: ierr
    integer :: ncid
    real, dimension(xstart:xend,ystart:yend) :: xdum

    ! Open the NetCDF file.
    write(*,'("wrfinput_flnm: ''", A, "''")') trim(wrfinput_flnm)
    ierr = nf90_open(wrfinput_flnm, NF90_NOWRITE, ncid)
    if (ierr /= 0) then
       write(*,'("READLAND_HRLDAS:  Problem opening wrfinput file: ''", A, "''")') trim(wrfinput_flnm)
       stop
    endif

    ! Get Latitude (lat)
    call get_2d_netcdf("XLAT", ncid, latitude,  units, xstart, xend, ystart, yend, FATAL, ierr)
    print*, 'latitude(xstart,ystart) = ', latitude(xstart,ystart)

    ! Get Longitude (lon)
    call get_2d_netcdf("XLONG", ncid, longitude, units, xstart, xend, ystart, yend, FATAL, ierr)
    print*, 'longitude(xstart,ystart) = ', longitude(xstart,ystart)

    ! Get Terrain (avg)
    call get_2d_netcdf("HGT", ncid, terrain,   units, xstart, xend, ystart, yend, FATAL, ierr)
    print*, 'terrain(xstart,ystart) = ', terrain(xstart,ystart)

    ! Get Deep layer temperature (TMN)
    call get_2d_netcdf("TMN", ncid, tbot_2d,   units, xstart, xend, ystart, yend, FATAL, ierr)
    print*, 'terrain(xstart,ystart) = ', terrain(xstart,ystart)

    ! Get Dominant Land Use categories (use)
    call get_landuse_netcdf(ncid, xdum ,   units, xstart, xend, ystart, yend)
    vegtyp = nint(xdum)
    print*, 'vegtyp(xstart,ystart) = ', vegtyp(xstart,ystart)

    ! Get Dominant Soil Type categories in the top layer (stl)
    call get_soilcat_netcdf(ncid, xdum ,   units, xstart, xend, ystart, yend)
    soltyp = nint(xdum)
    print*, 'soltyp(xstart,ystart) = ', soltyp(xstart,ystart)

    ! Close the NetCDF file
    ierr = nf90_close(ncid)
    if (ierr /= 0) stop "MODULE_NOAHLSM_HRLDAS_INPUT:  READLAND_HRLDAS:  NF90_CLOSE"

    ! Make sure vegtyp and soltyp are consistent when it comes to water points,
    ! by setting soil category to water when vegetation category is water, and
    ! vice-versa.
    where (vegtyp == ISWATER) soltyp = 14
    where (soltyp == 14) vegtyp = ISWATER

  end subroutine readland_hrldas

  subroutine get_landuse_netcdf(ncid, array, units, xstart, xend, ystart, yend)
    implicit none
    integer, intent(in) :: ncid
    integer, intent(in) :: xstart, xend, ystart, yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: array
    character(len=256), intent(out) :: units
    integer :: iret, varid
    character(len=24), parameter :: name = "IVGTYP"

    units = " "

    iret = nf90_inq_varid(ncid,  trim(name),  varid)
    if (iret /= 0) then
       print*, 'name = "', trim(name)//'"'
       stop "MODULE_NOAHLSM_HRLDAS_INPUT:  get_landuse_netcdf:  nf90_inq_varid"
    endif

    iret = nf90_get_var(ncid, varid, array, (/xstart, ystart/), (/xend-xstart+1, yend-ystart+1/))
    if (iret /= 0) then
       print*, 'name = "', trim(name)//'"'
       stop "MODULE_NOAHLSM_HRLDAS_INPUT:  get_landuse_netcdf:  nf90_get_var"
    endif

  end subroutine get_landuse_netcdf

  subroutine get_soilcat_netcdf(ncid, array, units, xstart, xend, ystart, yend)
    implicit none
    integer, intent(in) :: ncid
    integer, intent(in) :: xstart, xend, ystart, yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: array
    character(len=256), intent(out) :: units
    integer :: iret, varid
    character(len=24), parameter :: name = "ISLTYP"

    units = " "

    iret = nf90_inq_varid(ncid,  trim(name),  varid)
    call error_handler(iret, "Problem finding variable '"//trim(name)//"' in the wrfinput file.")

    iret = nf90_get_var(ncid, varid, array, (/xstart, ystart/), (/xend-xstart+1, yend-ystart+1/))
    call error_handler(iret, "Problem retrieving variable "//trim(name)//" from the wrfinput file.")

  end subroutine get_soilcat_netcdf

  subroutine get_2d_netcdf(name, ncid, array, units, xstart, xend, ystart, yend, &
       fatal_if_error, ierr)
    implicit none
    character(len=*), intent(in) :: name
    integer, intent(in) :: ncid
    integer, intent(in) :: xstart, xend, ystart, yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: array
    character(len=256), intent(out) :: units
    integer :: iret, varid
    ! FATAL_IF_ERROR:  an input code value:
    !      .TRUE. if an error in reading the data should stop the program.
    !      Otherwise the, IERR error flag is set, but the program continues.
    logical, intent(in) :: fatal_if_error 
    integer, intent(out) :: ierr

    units = " "

    iret = nf90_inq_varid(ncid,  name,  varid)
    if (iret /= 0) then
       if (FATAL_IF_ERROR) then
          print*, 'ncid = ', ncid
          call error_handler(iret, "MODULE_HRLDAS_NETCDF_IO:  Problem finding variable '"//trim(name)//"' in NetCDF file.")
       else
          ierr = iret
          return
       endif
    endif

    iret = nf90_get_att(ncid, varid, "units", units)
    if (iret /= 0) units = "units unknown"
!KWM    if (iret /= 0) then
!KWM       if (FATAL_IF_ERROR) then
!KWM          print*, 'name = "', trim(name)//'"'
!KWM          stop "MODULE_NOAHLSM_HRLDAS_INPUT:  get_2d_netcdf:  nf90_get_att"
!KWM       else
!KWM          ierr = iret
!KWM          return
!KWM       endif
!KWM    endif

    iret = nf90_get_var(ncid, varid, values=array, start=(/xstart,ystart/), count=(/xend-xstart+1,yend-ystart+1/))
    if (iret /= 0) then
       if (FATAL_IF_ERROR) then
          print*, 'ncid =', ncid
          call error_handler(iret, "MODULE_HRLDAS_NETCDF_IO:  Problem retrieving variable '"//trim(name)//"' from NetCDF file.")
       else
          ierr = iret
          return
       endif
    endif
!  print*, name//' = ', array(1,1), array(1,jdim), array(idim,jdim), array(idim,1)

    ierr = 0;
  end subroutine get_2d_netcdf

  subroutine get_2d_netcdf_soillevel(name, ncid, array, units, xstart, xend, ystart, yend, &
       fatal_if_error, layer_bottom, layer_top, ierr)
    implicit none
    character(len=*), intent(in) :: name
    integer, intent(in) :: ncid
    integer, intent(in) :: xstart, xend, ystart, yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: array
    character(len=256), intent(out) :: units
    ! FATAL_IF_ERROR:  an input code value:
    !      .TRUE. if an error in reading the data should stop the program.
    !      Otherwise the, IERR error flag is set, but the program continues.
    logical, intent(in) :: fatal_if_error 
    integer, intent(out) :: ierr
    real, intent(out) :: layer_bottom
    real, intent(out) :: layer_top

    integer :: iret, varid

    units = " "

    iret = nf90_inq_varid(ncid,  name,  varid)
    if (iret /= 0) then
       if (FATAL_IF_ERROR) then
          print*, 'name = "', trim(name)//'"'
          stop "MODULE_NOAHLSM_HRLDAS_INPUT:  get_2d_netcdf:  nf90_inq_varid"
       else
          ierr = iret
          return
       endif
    endif

    iret = nf90_get_att(ncid, varid, "units", units)
    if (iret /= 0) units = "units unknown"

    iret = nf90_get_var(ncid, varid, values=array, start=(/xstart,ystart/), count=(/xend-xstart+1,yend-ystart+1/))
    if (iret /= 0) then
       if (FATAL_IF_ERROR) then
          print*, 'name = "', trim(name)//'"'
          print*, 'varid =', varid
          print*, trim(nf90_strerror(iret))
          stop "MODULE_NOAHLSM_HRLDAS_INPUT:  get_2d_netcdf:  nf90_get_var"
       else
          ierr = iret
          return
       endif
    endif

    ! Also get some metadata about the soil layers:
    iret = nf90_get_att(ncid, varid, "layer_top", layer_top)
    call error_handler(iret, "Attempted to find 'layer_top' attribute on variable "//trim(name))

    iret = nf90_get_att(ncid, varid, "layer_bottom", layer_bottom)
    call error_handler(iret, "Attempted to find 'layer_bottom' attribute on variable "//trim(name))

    ierr = 0;
  end subroutine get_2d_netcdf_soillevel

  subroutine readinit_hrldas(netcdf_flnm, xstart, xend, ystart, yend, nsoil, sldpth, target_date, &
       ldasin_version, smc, stc, sh2o, cmc, t1, weasd, snodep)
    implicit none
    character(len=*),                                  intent(in)  :: netcdf_flnm
    integer,                                           intent(in)  :: xstart, xend
    integer,                                           intent(in)  :: ystart, yend
    integer,                                           intent(in)  :: nsoil
    real,    dimension(nsoil),                         intent(in)  :: sldpth
    character(len=*),                                  intent(in)  :: target_date
    
    integer,                                           intent(out) :: ldasin_version
    real,    dimension(xstart:xend,ystart:yend,nsoil), intent(out) :: smc
    real,    dimension(xstart:xend,ystart:yend,nsoil), intent(out) :: stc
    real,    dimension(xstart:xend,ystart:yend,nsoil), intent(out) :: sh2o
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: cmc
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: t1
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: weasd
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: snodep

    character(len=256) :: titlestr
    character(len=256) :: units
    character(len=8) :: name
    character(len=4) :: ldasin_llanduse

    integer :: ierr, ncid, ierr_snodep
    integer :: idx
    real, dimension(100) :: layer_bottom
    real, dimension(100) :: layer_top

    real, dimension(xstart:xend, ystart:yend, 4) :: soildummy

    !
    ! Open the NetCDF LDASIN file.
    !

    write(*,'("netcdf_flnm: ''", A, "''")') trim(netcdf_flnm)
    ierr = nf90_open(netcdf_flnm, NF90_NOWRITE, ncid)
    if (ierr /= 0) then
       write(*,'("Problem opening netcdf file: ''", A, "''")') &
            trim(netcdf_flnm)
       stop
    endif

    !
    ! Check the NetCDF LDASIN file for a version number.
    !

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "TITLE", titlestr)
    if (ierr /= 0) then
       write(*,'("WARNING:  LDASIN file does not have TITLE attribute.")')
       write(*,'("          This probably means that LDASIN files are from an older release.")')
       write(*,'("          I assume you know what you are doing.")')
       ldasin_version = 0
    else
       write(*,'("LDASIN TITLE attribute: ", A)') trim(titlestr)
       ! Pull out the version number, assuming that the version is identified by vYYYYMMDD, and 
       ! based on a search for the string "v20".
       idx = index(trim(titlestr), "v20")
       if (idx <= 0) then
          write(*,'("FATAL:  LDASIN file has a perverse version identifier")')
          !  write(*,'("          I assume you know what you are doing.")')
          stop
       else
          read(titlestr(idx+1:), '(I8)', iostat=ierr) ldasin_version
          if (ierr /= 0) then
             write(*,'("FATAL:  LDASIN file has a perverse version identifier")')
             !  write(*,'("          I assume you know what you are doing.")')
             stop
          endif
       endif
    endif
    write(*, '("ldasin_version = ", I8)') ldasin_version

    ierr = nf90_get_att(ncid, NF90_GLOBAL, "MMINLU", ldasin_llanduse)
    if (ierr /= 0) then
       write(*,'("WARNING:  LDASIN file does not have MMINLU attribute.")')
       write(*,'("          This probably means that LDASIN files are from an older release.")')
       write(*,'("          I assume you know what you are doing.")')
    else
       write(*,'("LDASIN MMNINLU attribute: ", A)') ldasin_llanduse
    endif

    call get_2d_netcdf("CANWAT",     ncid, cmc,     units, xstart, xend, ystart, yend, FATAL, ierr)
    call get_2d_netcdf("SKINTEMP",   ncid, t1,      units, xstart, xend, ystart, yend, FATAL, ierr)
    call get_2d_netcdf("WEASD",      ncid, weasd,   units, xstart, xend, ystart, yend, FATAL, ierr)
    if (trim(units) == "m") then
       ! No conversion necessary
    else if (trim(units) == "mm") then
       ! convert WEASD from mm to m
       weasd = weasd * 1.E-3
    else if (trim(units) == "kg m{-2}") then
       ! convert WEASD from mm to m
       weasd = weasd * 1.E-3
    else if (trim(units) == "kg/m2") then
       ! convert WEASD from mm to m
       weasd = weasd * 1.E-3
    else
       print*, 'units = "'//trim(units)//'"'
       stop "Unrecognized units on WEASD"
    endif
    call get_2d_netcdf("SNODEP",     ncid, weasd,   units, xstart, xend, ystart, yend, NOT_FATAL, ierr_snodep)

    call get_2d_netcdf_soillevel("STEMP_1",    ncid, soildummy(:,:,1), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(1), layer_top(1), ierr)
    call get_2d_netcdf_soillevel("STEMP_2",    ncid, soildummy(:,:,2), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(2), layer_top(2), ierr)
    call get_2d_netcdf_soillevel("STEMP_3",    ncid, soildummy(:,:,3), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(3), layer_top(3), ierr)
    call get_2d_netcdf_soillevel("STEMP_4",    ncid, soildummy(:,:,4), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(4), layer_top(4), ierr)
    write(*, '("layer_bottom(1:4) = ", 4F9.4)') layer_bottom(1:4)
    write(*, '("layer_top(1:4)    = ", 4F9.4)') layer_top(1:4)
    write(*, '("Soil depth = ", 10F12.6)') sldpth

    call init_interp(xstart, xend, ystart, yend, nsoil, sldpth, stc, 4, soildummy, layer_bottom(1:4), layer_top(1:4))

    call get_2d_netcdf_soillevel("SMOIS_1",    ncid, soildummy(:,:,1), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(1), layer_top(1), ierr)
    call get_2d_netcdf_soillevel("SMOIS_2",    ncid, soildummy(:,:,2), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(2), layer_top(2), ierr)
    call get_2d_netcdf_soillevel("SMOIS_3",    ncid, soildummy(:,:,3), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(3), layer_top(3), ierr)
    call get_2d_netcdf_soillevel("SMOIS_4",    ncid, soildummy(:,:,4), units,  xstart, xend, ystart, yend, FATAL, &
         layer_bottom(4), layer_top(4), ierr)

    call init_interp(xstart, xend, ystart, yend, nsoil, sldpth, smc, 4, soildummy, layer_bottom(1:4), layer_top(1:4))


    if (ierr_snodep /= 0) then
       ! Quick assumption regarding snow depth.
       snodep = weasd * 10.
    endif

    sh2o = smc

    ierr = nf90_close(ncid)
  end subroutine readinit_hrldas

  subroutine init_interp(xstart, xend, ystart, yend, nsoil, sldpth, var, nvar, src, layer_bottom, layer_top)
    implicit none
    integer, intent(in)    :: xstart, xend, ystart, yend, nsoil, nvar
    real, dimension(nsoil) :: sldpth ! the thickness of each layer
    real, dimension(xstart:xend, ystart:yend, nsoil), intent(out) :: var
    real, dimension(xstart:xend, ystart:yend, nvar ), intent(in)  :: src
    real, dimension(nvar),                            intent(in)  :: layer_bottom ! The depth from the surface of each layer bottom.
    real, dimension(nvar),                            intent(in)  :: layer_top    ! The depth from the surface of each layer top.
    integer :: i, j, k, kk, ktop, kbottom
    real, dimension(nsoil) :: dst_centerpoint
    real, dimension(nvar)  :: src_centerpoint
    real :: fraction

    do k = 1, nsoil
       if (k==1) then
          dst_centerpoint(k) = sldpth(k)/2.
       else
          dst_centerpoint(k) = sldpth(k)/2. + sum(sldpth(1:k-1))
       endif
       print*, 'k, dst_centerpoint(k) = ', k, dst_centerpoint(k)
    enddo
    print*

    do k = 1, nvar
       src_centerpoint(k) = 0.5*(layer_bottom(k)+layer_top(k))
       print*, 'k, src_centerpoint(k) = ', k, src_centerpoint(k)
    enddo

    KLOOP : do k = 1, nsoil

       if (dst_centerpoint(k) < src_centerpoint(1)) then
          ! If the center of the destination layer is closer to the surface than
          ! the center of the topmost source layer, then simply set the 
          ! value of the destination layer equal to the topmost source layer:
          print'("Shallow destination layer:  Taking destination layer at ",F7.4, " from source layer at ", F7.4)', &
               dst_centerpoint(k), src_centerpoint(1)
          var(:,:,k) = src(:,:,1)
          cycle KLOOP
       endif

       if (dst_centerpoint(k) > src_centerpoint(nvar)) then
          ! If the center of the destination layer is deeper than
          ! the center of the deepest source layer, then simply set the 
          ! value of the destination layer equal to the deepest source layer:
          print'("Deep destination layer:  Taking destination layer at ",F7.4, " from source layer at ", F7.4)', &
               dst_centerpoint(k), src_centerpoint(nvar)
          var(:,:,k) = src(:,:,nvar)
          cycle KLOOP
       endif

       ! Check if the center of the destination layer is "close" to the center
       ! of a source layer.  If so, simply set the value of the destination layer
       ! equal to the value of that close soil layer:
       do kk = 1, nvar
          if (abs(dst_centerpoint(k)-src_centerpoint(kk)) < 0.01) then
             print'("(Near) match for destination layer:  Taking destination layer at ",F7.4, " from source layer at ", F7.4)', &
                  dst_centerpoint(k), src_centerpoint(kk)
             var(:,:,k) = src(:,:,kk)
             cycle KLOOP
          endif
       enddo

       ! Otherwise, do a linear interpolation

       ! Get ktop, the index of the top bracketing layer from the source dataset.
       ! Which from the bottom up, will be the first source level that is closer 
       ! to the surface than the destination level
       ktop = -99999
       TOPLOOP : do kk = nvar,1,-1
          if (src_centerpoint(kk) < dst_centerpoint(k)) then
             ktop = kk
             exit TOPLOOP
          endif
       enddo TOPLOOP
       if (ktop < -99998) stop "ktop problem"


       ! Get kbottom, the index of the bottom bracketing layer from the source dataset.
       ! Which, from the top down, will be the first source level that is deeper than
       ! the destination level
       kbottom = -99999
       BOTTOMLOOP : do kk = 1, nvar
          if ( src_centerpoint(kk) > dst_centerpoint(k) ) then
             kbottom = kk
             exit BOTTOMLOOP
          endif
       enddo BOTTOMLOOP
       if (kbottom < -99998) stop "kbottom problem"

       fraction = (src_centerpoint(kbottom)-dst_centerpoint(k)) / (src_centerpoint(kbottom)-src_centerpoint(ktop))

       ! print '(I2, 1x, 3F7.3, F8.5)', k, src_centerpoint(ktop), dst_centerpoint(k), src_centerpoint(kbottom), fraction

       print '("dst(",I1,") = src(",I1,")*",F8.5," + src(",I1,")*",F8.5)', k, ktop, fraction, kbottom, (1.0-fraction)

       var(:,:,k) = (src(:,:,ktop)*fraction) + (src(:,:,kbottom)*(1.0-fraction))

    enddo KLOOP
     
  end subroutine init_interp


  subroutine READVEG_HRLDAS(flnm, xstart, xend, ystart, yend, target_date, vegtyp, fpar, lai, gvfmin, gvfmax)

    implicit none
    character(len=*),                                  intent(in)  :: flnm
    integer,                                           intent(in)  :: xstart, xend
    integer,                                           intent(in)  :: ystart, yend
    character(len=*),                                  intent(in)  :: target_date
    integer, dimension(xstart:xend,ystart:yend),       intent(in)  :: vegtyp
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: fpar
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: lai
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: gvfmin
    real,    dimension(xstart:xend,ystart:yend),       intent(out) :: gvfmax

    character(len=8)   :: name
    character(len=256) :: units
    integer :: ierr

    integer :: ierr_fpar
    integer :: ierr_lai

    integer :: i, j
    integer :: iret, ncid

    ! Open the NetCDF file.
!KWM    write(*,'("flnm: ''", A, "''")') trim(flnm)
    iret = nf90_open(flnm, NF90_NOWRITE, ncid)
    if (iret /= 0) then
       write(*,'("READVEG_HRLDAS:  Problem opening netcdf file: ''", A, "''")') trim(flnm)
       stop
    endif

    call get_2d_netcdf("VEGFRA",     ncid, fpar,     units, xstart, xend, ystart, yend, NOT_FATAL, ierr_fpar)
    call get_2d_netcdf("LAI",        ncid, lai,      units, xstart, xend, ystart, yend, NOT_FATAL, ierr_lai)

    if (ierr_fpar == 0) then
       ! fpar = fpar * 1.E-2 ! convert from percent to fraction
    else if (ierr_fpar /= 0) then
       ! Get it from tables
       ! print*,' READVEG_HRLDAS:  VEGFRA not found.  Initializing FPAR from table SHDTBL.'
       do i = xstart, xend
          do j = ystart, yend
!KWM             fpar(i,j) = shdtbl(vegtyp(i,j))
          enddo
       enddo
    endif
    if (ierr_lai /= 0) then
       ! Get it from tables
       ! print*,' READVEG_HRLDAS:  LAI not found.  Initializing LAI from table LAITBL.'
       do i = xstart, xend
          do j = ystart, yend
! Fixme for wrfcode input
!             lai(i,j) = laitbl(vegtyp(i,j))
          enddo
       enddo
    endif

    ! Get Minimum Green Vegetation Fraction GVFMIN
    call get_2d_netcdf("GVFMIN", ncid, gvfmin,   units, xstart, xend, ystart, yend, FATAL, ierr)

    ! Get Minimum Green Vegetation Fraction GVFMAX
    call get_2d_netcdf("GVFMAX", ncid, gvfmax,   units, xstart, xend, ystart, yend, FATAL, ierr)

    iret = nf90_close(ncid)
  end subroutine READVEG_HRLDAS

  subroutine READFORC_HRLDAS(flnm_template, forcing_timestep, target_date, xstart, xend, ystart, yend,  &
       t,q,u,v,p,lw,sw,pcp,lai,fpar)
    use kwm_string_utilities
    implicit none

    character(len=*),                   intent(in)  :: flnm_template
    integer,                            intent(in)  :: forcing_timestep
    integer,                            intent(in)  :: xstart, xend
    integer,                            intent(in)  :: ystart, yend
    character(len=19),                  intent(in)  :: target_date ! (YYYY-MM-DD_hh:mm:ss)
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: t
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: q
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: u
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: v
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: p
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: lw
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: sw
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: pcp
    real,             dimension(xstart:xend,ystart:yend), intent(out)   :: lai
    real,             dimension(xstart:xend,ystart:yend), intent(inout) :: fpar

    character(len=256) :: flnm
    character(len=256) :: units
    character(len=256) :: nextflnm
    integer :: ierr
    integer :: ncid

    type(inputstruct) :: lastread = inputstruct("0000-00-00_00:00:00", &
         null(), null(), null(), null(), null(), null(), null(), null(), null(), null() )
    type(inputstruct) :: nextread= inputstruct("0000-00-00_00:00:00", &
         null(), null(), null(), null(), null(), null(), null(), null(), null(), null() )

!KWM    print*, 'target_date        = ', target_date
!KWM    print*, 'lastread%read_date = ', lastread%read_date
!KWM    print*, 'nextread%read_date = ', nextread%read_date

    if (target_date > nextread%read_date ) then
       !
       ! We've advanced beyond the date of the end-bracketing data in memory.
       ! Read the next (later) forcing data, and put the data into the nextread
       ! structure.
       !

       if (nextread%read_date /= "0000-00-00_00:00:00") then
          ! Clear the old lastread data
          call clear_inputstruct(lastread)

          ! Copy nextread to lastread
          lastread = nextread
       
          ! Clear nextread
          call nullify_inputstruct(nextread)
       endif

       ! Guess the next read date (from the last read date and the forcing timestep).
       ! If there is no last date, assume we're at the beginning of our processing
       ! and take the target_date as the first timestep, for which forcing data
       ! must be available.

       if (lastread%read_date == "0000-00-00_00:00:00") then
          nextread%read_date = target_date
       else
          call geth_newdate(nextread%read_date, lastread%read_date, forcing_timestep)
       endif

       ! Build a file name
       flnm = flnm_template
       call strrep(flnm, "<date>", nextread%read_date(1:4)//nextread%read_date(6:7)//nextread%read_date(9:10)//nextread%read_date(12:13))

!KWM       print*, 'new nextread%read_date = ', nextread%read_date
       print*, 'read file:  ', trim(flnm)

       ! Open the NetCDF file.
       ierr = nf90_open(flnm, NF90_NOWRITE, ncid)
       if (ierr /= 0) then
          write(*,'("A)  Problem opening netcdf file: ''", A, "''")') trim(flnm)
          stop
       endif

       ! Allocate space to hold data
       call allocate_inputstruct(nextread, xstart, xend, ystart, yend)

       ! Read the data
       call get_2d_netcdf("T2D",     ncid, nextread%t,     units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("Q2D",     ncid, nextread%q,     units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("U2D",     ncid, nextread%u,     units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("V2D",     ncid, nextread%v,     units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("PSFC",    ncid, nextread%p,     units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("LWDOWN",  ncid, nextread%lw,    units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("SWDOWN",  ncid, nextread%sw,    units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("RAINRATE",ncid, nextread%pcp,   units, xstart, xend, ystart, yend, FATAL, ierr)
       call get_2d_netcdf("VEGFRA",  ncid, nextread%fpar,  units, xstart, xend, ystart, yend, NOT_FATAL, ierr)
       if (ierr /= 0) then
          ! print*, 'VEGFRA not found!'
          ! If we don't find a new VEGFRA, carry over the old one
          if (associated(lastread%fpar)) then
             nextread%fpar = lastread%fpar
          else
             nextread%fpar = fpar
          endif
       endif
       call get_2d_netcdf("LAI",     ncid, nextread%lai,   units, xstart, xend, ystart, yend, NOT_FATAL, ierr)
       if (ierr /= 0) then
          ! print*, 'LAI not found!'
          ! If we don't find a new LAI, carry over the old one
          if (associated(lastread%fpar)) then
             nextread%lai = lastread%lai
          endif
       endif

       ! Close the file
       ierr = nf90_close(ncid)

    endif

    if (target_date == nextread%read_date) then
       !
       ! We have advanced to the later date of our bracketing times for interpolation.
       ! Take that data as is, no interpolation necessary, move that data into the 
       ! lastread structure, and return that data.
       !

       ! Fill the t, q, u, v, ... arrays with data from the nextread structure.
       call copyfrom_inputstruct(nextread, t, q, u, v, p, lw, sw, pcp, fpar, lai, xstart, xend, ystart, yend)

       ! Clear the old lastread data
       call clear_inputstruct(lastread)

       ! Copy nextread to lastread
       lastread = nextread

       ! Set the nextread%read_date field to signal that we need to read
       nextread%read_date = "0000-00-00_00:00:00"

       ! Clear nextread
       call nullify_inputstruct(nextread)

    else if ( ( target_date > lastread%read_date ) .and. ( target_date < nextread%read_date ) ) then

       !
       ! We are at a Noah time step between the lastread data and the available nextread data.
       ! Do temporal interpolation and return the interpolated data.  Keep lastread
       ! and nextread as they were.
       !

       ! Fill the t, q, u, v, ... arrays with data interpolated between lastread and nextread times.
       call interpolate_inputstruct(lastread, nextread, target_date, &
            t, q, u, v, p, lw, sw, pcp, fpar, lai, xstart, xend, ystart, yend)

    else

       print*, 'target_date        = ', target_date
       print*, 'lastread%read_date = ', lastread%read_date
       print*, 'nextread%read_date = ', nextread%read_date

       STOP "We should not be here.  Problem with the logic of READFORC_SHORTER_TIMESTEP"

    endif

  end subroutine READFORC_HRLDAS

  subroutine allocate_inputstruct(instruct, xstart, xend, ystart, yend)
    implicit none
    type(inputstruct)   :: instruct
    integer, intent(in) :: xstart
    integer, intent(in) :: xend
    integer, intent(in) :: ystart
    integer, intent(in) :: yend

    integer :: allostat

    allocate(instruct%t   (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%t"

    allocate(instruct%q   (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%q"

    allocate(instruct%u   (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%u"

    allocate(instruct%v   (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%v"

    allocate(instruct%p   (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%p"

    allocate(instruct%lw  (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%lw"

    allocate(instruct%sw  (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%sw"

    allocate(instruct%pcp (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%pcp"

    allocate(instruct%fpar(xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%fpar"

    allocate(instruct%lai (xstart:xend,ystart:yend), stat=allostat )
    if (allostat/=0) stop "Problem allocating instruct%lai"
  end subroutine allocate_inputstruct

  subroutine copyfrom_inputstruct(instruct, t, q, u, v, p, lw, sw, pcp, fpar, lai, xstart, xend, ystart, yend)
    implicit none
    type(inputstruct), intent(in) :: instruct
    integer,           intent(in) :: xstart, xend, ystart, yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: t, q, u, v, p, lw, sw, pcp, fpar, lai
    t    = instruct%t
    q    = instruct%q
    u    = instruct%u
    v    = instruct%v
    p    = instruct%p
    lw   = instruct%lw
    sw   = instruct%sw
    pcp  = instruct%pcp
    fpar = instruct%fpar
    lai  = instruct%lai
  end subroutine copyfrom_inputstruct

  subroutine interpolate_inputstruct(instructA, instructB, target_date, &
       t, q, u, v, p, lw, sw, pcp, fpar, lai, xstart, xend, ystart, yend)
    implicit none
    type(inputstruct),                        intent(in)  :: instructA, instructB
    character(len=19),                        intent(in)  :: target_date
    integer,                                  intent(in)  :: xstart, xend, ystart, yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: t, q, u, v, p, lw, sw, pcp, fpar, lai

    integer :: idts, idts2
    real    :: fraction

    call geth_idts(target_date, instructA%read_date, idts)
    call geth_idts(instructB%read_date, instructA%read_date, idts2)

    fraction = real(idts2-idts)/real(idts2)
    t  = ( instructA%t  * fraction ) + ( instructB%t  * (1.0-fraction) )
    q  = ( instructA%q  * fraction ) + ( instructB%q  * (1.0-fraction) )
    u  = ( instructA%u  * fraction ) + ( instructB%u  * (1.0-fraction) )
    v  = ( instructA%v  * fraction ) + ( instructB%v  * (1.0-fraction) )
    p  = ( instructA%p  * fraction ) + ( instructB%p  * (1.0-fraction) )
    lw = ( instructA%lw * fraction ) + ( instructB%lw * (1.0-fraction) )
    sw = ( instructA%sw * fraction ) + ( instructB%sw * (1.0-fraction) )
    pcp = instructA%pcp
    fpar = instructA%fpar
    lai  = instructA%lai
  end subroutine interpolate_inputstruct

  subroutine clear_inputstruct(instruct)
    implicit none
    type(inputstruct) :: instruct

    if (associated(instruct%t)) then
       deallocate(instruct%t)
       nullify(instruct%t)
    endif

    if (associated(instruct%q)) then
       deallocate(instruct%q)
       nullify(instruct%q)
    endif

    if (associated(instruct%u)) then
       deallocate(instruct%u)
       nullify(instruct%u)
    endif

    if (associated(instruct%v)) then
       deallocate(instruct%v)
       nullify(instruct%v)
    endif

    if (associated(instruct%p)) then
       deallocate(instruct%p)
       nullify(instruct%p)
    endif

    if (associated(instruct%lw)) then
       deallocate(instruct%lw)
       nullify(instruct%lw)
    endif

    if (associated(instruct%sw)) then
       deallocate(instruct%sw)
       nullify(instruct%sw)
    endif

    if (associated(instruct%pcp)) then
       deallocate(instruct%pcp)
       nullify(instruct%pcp)
    endif

    if (associated(instruct%fpar)) then
       deallocate(instruct%fpar)
       nullify(instruct%fpar)
    endif

    if (associated(instruct%lai)) then
       deallocate(instruct%lai)
       nullify(instruct%lai)
    endif
  end subroutine clear_inputstruct


  subroutine nullify_inputstruct(instruct)
    implicit none
    type(inputstruct) :: instruct

    nullify(instruct%t)
    nullify(instruct%q)
    nullify(instruct%u)
    nullify(instruct%v)
    nullify(instruct%p)
    nullify(instruct%lw)
    nullify(instruct%sw)
    nullify(instruct%pcp)
    nullify(instruct%fpar)
    nullify(instruct%lai)
  end subroutine nullify_inputstruct


  subroutine READSNOW_HRLDAS(flnm,xstart,xend,ystart,yend,target_date,weasd,snodep)
    implicit none

    character(len=*),                                     intent(in)  :: flnm
    integer,                                              intent(in)  :: xstart, xend
    integer,                                              intent(in)  :: ystart, yend
    character(len=*),                                     intent(in)  :: target_date
    real,             dimension(xstart:xend,ystart:yend), intent(out) :: weasd
    real,             dimension(xstart:xend,ystart:yend), intent(out) :: snodep

    character(len=256) :: units
    integer :: ierr
    integer :: ncid

    ! Open the NetCDF file.

    ierr = nf90_open(flnm, NF90_NOWRITE, ncid)
    if (ierr /= 0) then
       write(*,'("READSNOW_HRLDAS:  Problem opening netcdf file: ''", A, "''")') trim(flnm)
       stop
    endif

    call get_2d_netcdf("WEASD",  ncid, weasd,   units, xstart, xend, ystart, yend, FATAL, ierr)

    if (trim(units) == "m") then
       ! No conversion necessary
    else if (trim(units) == "mm") then
       ! convert WEASD from mm to m
       weasd = weasd * 1.E-3
    else if (trim(units) == "kg m{-2}") then
       ! convert WEASD from mm to m
       weasd = weasd * 1.E-3
    else if (trim(units) == "kg/m2") then
       ! convert WEASD from mm to m
       weasd = weasd * 1.E-3
    else
       print*, 'units = "'//trim(units)//'"'
       stop "Unrecognized units on WEASD"
    endif

    call get_2d_netcdf("SNODEP",     ncid, snodep,   units, xstart, xend, ystart, yend, NOT_FATAL, ierr)

    if (ierr /= 0) then
       ! Quick assumption regarding snow depth.
       snodep = weasd * 10.
    endif

    ierr = nf90_close(ncid)

  end subroutine READSNOW_HRLDAS


  subroutine output_hrldas(ounit, version, igrid, output_timestep, llanduse, split_output_count, hgrid, &
       ix, jx, iswater,                                              &
       mapproj, lat1, lon1, dx, dy, truelat1, truelat2, cen_lon,    &
       nsoil, sldpth, startdate, date, &
       vegtyp, soltyp, skintemp, cmc, stc, smc, sh2o, soilmx, sfcrnoff, ugdrnoff, &
       intrflow, sfcevp, etakin, canevp, edirx, ettx, albedx, weasd, acrain, acsnom, &
       esnow2d, drip2d, dewfall, snodep, vegfra, znt, hx, qfx, grdflx, sw, lw, &
#ifdef _HRLDAS_URBAN_
       ucmcall, &
                TRL_URB3D, TBL_URB3D, TGL_URB3D, TR_URB2D, TB_URB2D, TG_URB2D, TC_URB2D, &
#endif
       fx, xlai)
    implicit none

    integer,                                  intent(in) :: ounit
    character(len=*),                         intent(in) :: version
    integer,                                  intent(in) :: igrid
    integer,                                  intent(in) :: output_timestep
    character(len=*),                         intent(in) :: llanduse
    integer,                                  intent(in) :: split_output_count
    character,                                intent(in) :: hgrid
    integer,                                  intent(in) :: ix
    integer,                                  intent(in) :: jx
    integer,                                  intent(in) :: iswater
    integer,                                  intent(in) :: mapproj
    real,                                     intent(in) :: lat1
    real,                                     intent(in) :: lon1
    real,                                     intent(in) :: dx
    real,                                     intent(in) :: dy
    real,                                     intent(in) :: truelat1
    real,                                     intent(in) :: truelat2
    real,                                     intent(in) :: cen_lon
    integer,                                  intent(in) :: nsoil
    real,             dimension(nsoil),       intent(in) :: sldpth
    character(len=19),                        intent(in) :: startdate
    character(len=19),                        intent(in) :: date
    integer,          dimension(ix,jx),       intent(in) :: vegtyp
    integer,          dimension(ix,jx),       intent(in) :: soltyp
    real,             dimension(ix,jx),       intent(in) :: skintemp
    real,             dimension(ix,jx),       intent(in) :: cmc
    real,             dimension(ix,jx,nsoil), intent(in) :: stc
    real,             dimension(ix,jx,nsoil), intent(in) :: smc
    real,             dimension(ix,jx,nsoil), intent(in) :: sh2o
    real,             dimension(ix,jx),       intent(in) :: soilmx
    real,             dimension(ix,jx),       intent(in) :: ugdrnoff
    real,             dimension(ix,jx),       intent(in) :: sfcrnoff
    real,             dimension(ix,jx),       intent(in) :: intrflow
    real,             dimension(ix,jx),       intent(in) :: sfcevp
    real,             dimension(ix,jx),       intent(in) :: etakin
    real,             dimension(ix,jx),       intent(in) :: canevp
    real,             dimension(ix,jx),       intent(in) :: edirx
    real,             dimension(ix,jx),       intent(in) :: ettx
    real,             dimension(ix,jx),       intent(in) :: albedx
    real,             dimension(ix,jx),       intent(in) :: weasd
    real,             dimension(ix,jx),       intent(in) :: acrain
    real,             dimension(ix,jx),       intent(in) :: acsnom
    real,             dimension(ix,jx),       intent(in) :: esnow2d
    real,             dimension(ix,jx),       intent(in) :: drip2d
    real,             dimension(ix,jx),       intent(in) :: snodep
    real,             dimension(ix,jx),       intent(in) :: vegfra
    real,             dimension(ix,jx),       intent(in) :: znt
    real,             dimension(ix,jx),       intent(in) :: hx
    real,             dimension(ix,jx),       intent(in) :: qfx
    real,             dimension(ix,jx),       intent(in) :: grdflx
    real,             dimension(ix,jx),       intent(in) :: sw
    real,             dimension(ix,jx),       intent(in) :: lw
    real,             dimension(ix,jx),       intent(in) :: fx
    real,             dimension(ix,jx),       intent(in) :: xlai
    real,             dimension(ix,jx),       intent(in) :: dewfall
#ifdef _HRLDAS_URBAN_
    integer,                                  intent(in) :: ucmcall
    real,             dimension(ix,jx,nsoil), intent(in) :: trl_urb3d
    real,             dimension(ix,jx,nsoil), intent(in) :: tbl_urb3d
    real,             dimension(ix,jx,nsoil), intent(in) :: tgl_urb3d
    real,             dimension(ix,jx),       intent(in) :: tr_urb2d
    real,             dimension(ix,jx),       intent(in) :: tb_urb2d
    real,             dimension(ix,jx),       intent(in) :: tg_urb2d
    real,             dimension(ix,jx),       intent(in) :: tc_urb2d
#endif

    integer :: output_count = 0
    integer, save :: ncid

    integer :: dimid_ix, dimid_jx, dimid_times, dimid_datelen, varid, n
    integer :: dimid_dum, dimid_layers
    integer :: iret
    character(len=256) :: output_flnm
    character(len=19)  :: date19

    real, dimension(ix,jx) :: xdum

    if (output_count == 0) then
       if (mod(output_timestep,3600) == 0) then
          write(output_flnm, '(A10,".LDASOUT_DOMAIN",I1)') date(1:4)//date(6:7)//date(9:10)//date(12:13), igrid
       elseif (mod(output_timestep,60) == 0) then
          write(output_flnm, '(A12,".LDASOUT_DOMAIN",I1)') date(1:4)//date(6:7)//date(9:10)//date(12:13)//date(15:16), igrid
       else
          write(output_flnm, '(A14,".LDASOUT_DOMAIN",I1)') date(1:4)//date(6:7)//date(9:10)//date(12:13)//date(15:16)//date(18:19), igrid
       endif
       print*, 'output_flnm = "'//trim(output_flnm)//'"'
       iret = nf90_create(trim(output_flnm), NF90_CLOBBER, ncid)
       if (iret /= 0) stop "Problem nf_create"

       iret = nf90_def_dim(ncid, "Time", NF90_UNLIMITED, dimid_times)
       iret = nf90_def_dim(ncid, "DateStrLen", 19, dimid_datelen)
       iret = nf90_def_dim(ncid, "west_east", ix, dimid_ix)
       iret = nf90_def_dim(ncid, "south_north", jx, dimid_jx)
       iret = nf90_def_dim(ncid, "west_east_stag", ix+1, dimid_dum)
       iret = nf90_def_dim(ncid, "south_north_stag", jx+1, dimid_dum)
       iret = nf90_def_dim(ncid, "soil_layers_stag", nsoil, dimid_layers)

       iret = nf90_put_att(ncid, NF90_GLOBAL, "TITLE", "OUTPUT FROM HRLDAS "//version)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "missing_value", -1.E33)

       ! TODO:  Add Grid information   (should look more-or-less like wrfout files)
       ! TODO:  Add Units information  (should look more-or-less like wrfout files)

       date19(1:19) = "0000-00-00_00:00:00"
       date19(1:len_trim(startdate)) = startdate

       iret = nf90_put_att(ncid, NF90_GLOBAL, "START_DATE", date19)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "MAP_PROJ", mapproj)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "LAT1", lat1)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "LON1", lon1)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "DX", dx)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "DY", dy)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "TRUELAT1", truelat1)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "TRUELAT2", truelat2)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "STAND_LON", cen_lon)
       iret = nf90_put_att(ncid, NF90_GLOBAL, "MMINLU", llanduse)

!
! Done with dimensions and global attributes.
! Now define and describe all our NetCDF output variables.
!

       iret = nf90_def_var(ncid,  "Times",  NF90_CHAR, (/dimid_datelen,dimid_times/), varid)

!
!      _PARGS_ is the result of laziness.  
!      I simply don't want to keep repeating these four arguments in all of the subroutine calls below.
!
#define _PARGS_ ncid,dimid_ix,dimid_jx,dimid_times

       call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "SOIL_T"  , "soil temperature",     "K")
       call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "SOIL_M"  , "volumetric soil moisture", "m3 m-3")
       call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "SOIL_W"  , "liquid volumetric soil moisture", "m3 m-3")
       call make_var_att_2d(_PARGS_, NF90_FLOAT,               "SOIL_MX" , "total column soil moisture", "mm")

       call make_var_att_2d(_PARGS_, NF90_INT, "IVGTYP" , "Dominant vegetation category"       , "dimensionless category"          )
       call make_var_att_2d(_PARGS_, NF90_INT, "ISLTYP" , "Dominant soil category"             , "dimensionless category"          )

       call make_var_att_2d(_PARGS_, NF90_FLOAT, "SKINTEMP", "Skin temperature"                   , "K"                  )
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "CANWAT"  , "Canopy water content"               , "mm"                 )
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "XLAI"    , "Leaf area index"                    , "dimensionless")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "HFX"     , "Upward surface sensible heat flux", "W m-2")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "QFX"     , "Upward surface latent heat flux", "W m-2")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "GRDFLX"  , "Ground heat flux at surface", "W m-2")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "SW"      , "Downward shortwave radiation flux at surface", "W m-2")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "LW"      , "Downward longwave radiation flux at surface", "W m-2")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "FDOWN"   , "Radiation forcing at the surface", "W m-2")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "ALBEDX"  , "Albedo -- What kind? (I.e., including what effects?)", "fraction")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "VEGFRA"  , "Green vegetation fraction", "fraction")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "Z0"      , "Roughness length", "m")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "SNODEP"  , "Snow depth", "m")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "UGDRNOFF", "Accumulated underground runoff", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "SFCRNOFF", "Accumulatetd surface runoff", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "INTRFLOW", "Accumulated interflow runoff", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "SFCEVP"  , "Accumulated evaporation from surface", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "ETAKIN"  , "Evapotranspiration", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "CANEVP"  , "Accumulated canopy evaporation", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "EDIRX"   , "Accumulated direct soil evaporation", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "ETTX"    , "Accumulated plant transpiration", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "WEASD"   , "Water equivalent accumulated snow depth", "m")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "ACRAIN"  , "Accumulated precipitation", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "ACSNOM"  , "Accumulated snow melt", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "ESNOW"   , "Accumulated evaporation of snow", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "DRIP"    , "Accumulated canopy drip", "mm")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "DEWFALL" , "Accumulated dewfall", "mm")
#ifdef _HRLDAS_URBAN_
       if (ucmcall == 1) then
          call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "TRL"  , "Roof temperature",     "K")
          call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "TBL"  , "Wall temperature",     "K")
          call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "TGL"  , "Road temperature",     "K")
          call make_var_att_2d(_PARGS_, NF90_FLOAT, "TR"    , "Roof layer temperature", "K")
          call make_var_att_2d(_PARGS_, NF90_FLOAT, "TB"    , "Wall layer temperature", "K")
          call make_var_att_2d(_PARGS_, NF90_FLOAT, "TG"    , "Road layer temperature", "K")
          call make_var_att_2d(_PARGS_, NF90_FLOAT, "TC"    , "Urban canopy temperature", "K")
       endif
#endif

#undef _PARGS_

       iret = nf90_enddef(ncid)
    endif
    output_count = output_count + 1

!
! Done defining and describing all our NetCDF output variables.
! Now actually put the data for each variable to the NetCDF file.
!

    date19(1:19) = "0000-00-00_00:00:00"
    date19(1:len_trim(date)) = date
    iret = nf90_inq_varid(ncid, "Times", varid)
    if (iret /= 0) then
       print*, 'nf90_inq_varid: ', trim(nf90_strerror(iret))
       stop
    endif
    iret = nf90_put_var(ncid, varid, date, (/1,output_count/), (/19,1/))
    if (iret /= 0) then
       print*, 'nf90_put_var:', trim(nf90_strerror(iret))
       stop
    endif

!
!   _PARGS_ is the result of laziness.  
!   I simply don't want to keep repeating these five arguments in all of the subroutine calls below.
!
#define _PARGS_ ncid,output_count,vegtyp,iswater,ix,jx
    call put_var_3d(_PARGS_, nsoil, "SOIL_T", stc)
    call put_var_3d(_PARGS_, nsoil, "SOIL_M", smc)
    call put_var_3d(_PARGS_, nsoil, "SOIL_W", sh2o)
    call put_var_2d(_PARGS_,        "SOIL_MX", soilmx)

    call put_var_int(_PARGS_, "IVGTYP" ,  vegtyp)
    call put_var_int(_PARGS_, "ISLTYP" ,  soltyp)

    call put_var_2d(_PARGS_, "SKINTEMP" , skintemp)
    xdum = cmc*1.E3 ! Converting from m to mm 
    call put_var_2d(_PARGS_, "CANWAT"   , xdum    )  !cmc
    call put_var_2d(_PARGS_, "XLAI"     , xlai    )
    call put_var_2d(_PARGS_, "HFX"      , hx      )
    call put_var_2d(_PARGS_, "QFX"      , qfx     )
    call put_var_2d(_PARGS_, "GRDFLX"   , grdflx  )
    call put_var_2d(_PARGS_, "SW"       , sw      )
    call put_var_2d(_PARGS_, "LW"       , lw      )
    call put_var_2d(_PARGS_, "FDOWN"    , fx      )
    call put_var_2d(_PARGS_, "ALBEDX"   , albedx  )
    call put_var_2d(_PARGS_, "VEGFRA"   , vegfra  )
    call put_var_2d(_PARGS_, "Z0"       , znt     )
    call put_var_2d(_PARGS_, "SNODEP"   , snodep  )
    call put_var_2d(_PARGS_, "UGDRNOFF" , ugdrnoff)
    call put_var_2d(_PARGS_, "SFCRNOFF" , sfcrnoff)
    call put_var_2d(_PARGS_, "INTRFLOW" , intrflow)
    call put_var_2d(_PARGS_, "SFCEVP"   , sfcevp  )
    call put_var_2d(_PARGS_, "ETAKIN"   , etakin  )
    call put_var_2d(_PARGS_, "CANEVP"   , canevp  )
    call put_var_2d(_PARGS_, "EDIRX"    , edirx   )
    call put_var_2d(_PARGS_, "ETTX"     , ettx    )
    call put_var_2d(_PARGS_, "WEASD"    , weasd   )
    call put_var_2d(_PARGS_, "ACRAIN"   , acrain  )
    call put_var_2d(_PARGS_, "ACSNOM"   , acsnom  )
    call put_var_2d(_PARGS_, "ESNOW"    , esnow2d )
    call put_var_2d(_PARGS_, "DRIP"     , drip2d  )
    call put_var_2d(_PARGS_, "DEWFALL"  , dewfall )
#ifdef _HRLDAS_URBAN_
    if (ucmcall == 1) then
       call put_var_3d(_PARGS_, nsoil, "TRL", trl_urb3d)
       call put_var_3d(_PARGS_, nsoil, "TBL", tbl_urb3d)
       call put_var_3d(_PARGS_, nsoil, "TGL", tgl_urb3d)
       call put_var_2d(_PARGS_, "TR"     , tr_urb2d  )
       call put_var_2d(_PARGS_, "TB"     , tb_urb2d  )
       call put_var_2d(_PARGS_, "TG"     , tg_urb2d  )
       call put_var_2d(_PARGS_, "TC"     , tc_urb2d  )
    endif
#endif
#undef _PARGS_

    iret = nf90_sync(ncid)

    if (output_count == split_output_count) then
       output_count =0 
       iret = nf90_close(ncid)
    endif

  end subroutine output_hrldas

  subroutine make_var_att_2d(ncid, dimid_ix, dimid_jx, dimid_times, itype, varname, vardesc, varunits)
    implicit none
    integer,          intent(in) :: ncid
    character(len=*), intent(in) :: varname
    character(len=*), intent(in) :: vardesc
    character(len=*), intent(in) :: varunits
    integer,          intent(in) :: dimid_ix
    integer,          intent(in) :: dimid_jx
    integer,          intent(in) :: dimid_times
    integer,          intent(in) :: itype
    integer :: iret
    integer :: varid

    iret = nf90_def_var(ncid,  varname,   itype, (/dimid_ix,dimid_jx,dimid_times/), varid)
    if (iret /= 0) then
       print*, 'varname = ', varname
       stop "Failure defining variable."
    endif

    iret = nf90_put_att(ncid, varid, "MemoryOrder", "XY ")
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       stop "Failure adding MemoryOrder attribute"
    endif

    iret = nf90_put_att(ncid, varid, "description", vardesc)
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       print*,'vardesc = "'//vardesc//'"'
       stop "Failure adding description attribute"
    endif

    iret = nf90_put_att(ncid, varid, "units", varunits)
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       print*,'vardesc = "'//vardesc//'"'
       print*,'varunits = "'//varunits//'"'
       stop "Failure adding units attribute"
    endif

    iret = nf90_put_att(ncid, varid, "stagger", " ")
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       stop "Failure adding stagger attribute"
    endif

  end subroutine make_var_att_2d

  subroutine make_var_att_3d(ncid, dimid_ix, dimid_jx, dimid_times, itype, dimid_layers, varname, vardesc, varunits)
    implicit none
    integer,          intent(in) :: ncid
    character(len=*), intent(in) :: varname
    character(len=*), intent(in) :: vardesc
    character(len=*), intent(in) :: varunits
    integer,          intent(in) :: dimid_ix
    integer,          intent(in) :: dimid_jx
    integer,          intent(in) :: dimid_times
    integer,          intent(in) :: dimid_layers
    integer,          intent(in) :: itype
    integer :: iret
    integer :: varid

    iret = nf90_def_var(ncid,  varname, itype, (/dimid_ix,dimid_jx,dimid_layers,dimid_times/), varid)
    if (iret /= 0) then
       print*, 'varname = ', varname
       stop "Failure defining variable."
    endif

    iret = nf90_put_att(ncid, varid, "MemoryOrder", "XYZ")
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       stop "Failure adding MemoryOrder attribute"
    endif

    iret = nf90_put_att(ncid, varid, "description", vardesc)
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       print*,'vardesc = "'//vardesc//'"'
       stop "Failure adding description attribute"
    endif

    iret = nf90_put_att(ncid, varid, "units", varunits)
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       print*,'vardesc = "'//vardesc//'"'
       print*,'varunits = "'//varunits//'"'
       stop "Failure adding units attribute"
    endif

    iret = nf90_put_att(ncid, varid, "stagger", "Z")
    if (iret /= 0) then
       print*,'varname = "'//varname//'"'
       stop "Failure adding stagger attribute"
    endif

  end subroutine make_var_att_3d

  subroutine put_var_2d(ncid, output_count, vegtyp, iswater, ix, jx, varname, vardata)
    implicit none
    integer,                   intent(in) :: ncid
    integer,                   intent(in) :: output_count
    character(len=*),          intent(in) :: varname
    integer,                   intent(in) :: ix
    integer,                   intent(in) :: jx
    integer, dimension(ix,jx), intent(in) :: vegtyp
    integer,                   intent(in) :: iswater
    real,    dimension(ix,jx), intent(in) :: vardata

    real,    dimension(ix,jx)             :: xdum
    integer                               :: iret
    integer                               :: varid

    xdum = vardata
    where (vegtyp(:,:) == ISWATER) xdum = -1.E33

    iret = nf90_inq_varid(ncid,  varname, varid)
    if (iret /= 0) then
       print*, 'varname = "'//varname//'"'
       stop "Subroutine PUT_VAR_2D:  Problem finding variable id."
    endif

    iret = nf90_put_var(ncid, varid, xdum, (/1,1,output_count/), (/ix,jx,1/))
    if (iret /= 0) then
       print*, 'varname = "'//varname//'"'
       stop "Subroutine PUT_VAR_2D:  Problem putting variable to NetCDF file."
    endif

  end subroutine put_var_2d


  subroutine put_var_int(ncid, output_count, vegtyp, iswater, ix, jx, varname, vardata)
    implicit none
    integer,                   intent(in) :: ncid
    integer,                   intent(in) :: output_count
    character(len=*),          intent(in) :: varname
    integer,                   intent(in) :: ix
    integer,                   intent(in) :: jx
    integer, dimension(ix,jx), intent(in) :: vegtyp
    integer,                   intent(in) :: iswater
    integer, dimension(ix,jx), intent(in) :: vardata

    integer                               :: iret
    integer                               :: varid

    iret = nf90_inq_varid(ncid,  varname, varid)
    if (iret /= 0) then
       print*, 'varname = "'//varname//'"'
       stop "Subroutine PUT_VAR_INT:  Problem finding variable id."
    endif

    iret = nf90_put_var(ncid, varid, vardata, (/1,1,output_count/), (/ix,jx,1/))
    if (iret /= 0) then
       print*, 'varname = "'//varname//'"'
       stop "Subroutine PUT_VAR_INT:  Problem putting variable to NetCDF file."
    endif

  end subroutine put_var_int

  subroutine put_var_3d(ncid, output_count, vegtyp, iswater, ix, jx, nsoil, varname, vardata)
    implicit none
    integer,                         intent(in) :: ncid
    integer,                         intent(in) :: output_count
    character(len=*),                intent(in) :: varname
    integer,                         intent(in) :: ix
    integer,                         intent(in) :: jx
    integer,                         intent(in) :: nsoil
    integer, dimension(ix,jx),       intent(in) :: vegtyp
    integer,                         intent(in) :: iswater
    real,    dimension(ix,jx,nsoil), intent(in) :: vardata

    real,    dimension(ix,jx,nsoil)             :: xdum
    integer                                     :: iret
    integer                                     :: varid
    integer                                     :: n

    xdum = vardata
    do n = 1, nsoil
       where (vegtyp(:,:) == ISWATER) xdum(:,:,n) = -1.E33
    enddo

    iret = nf90_inq_varid(ncid,  varname, varid)
    if (iret /= 0) then
       print*, 'varname = "'//varname//'"'
       stop "Subroutine PUT_VAR_3D:  Problem finding variable id."
    endif

    iret = nf90_put_var(ncid, varid, xdum, (/1,1,1,output_count/), (/ix,jx,nsoil,1/))
    if (iret /= 0) then
       print*, 'varname = "'//varname//'"'
       stop "Subroutine PUT_VAR_3D:  Problem putting variable to NetCDF file."
    endif

  end subroutine put_var_3d

  subroutine write_restart(version, igrid, llanduse, olddate, startdate,                      &
       ix, jx, nsoil, dx, dy, truelat1, truelat2, mapproj, lat1, lon1, cen_lon,               &
       iswater, vegtyp,                                                                       &
#ifdef _HRLDAS_URBAN_
       ucmcall, &
                ts_urb2d, tr_urb2d, tb_urb2d, tg_urb2d, tc_urb2d, qc_urb2d, uc_urb2d,         &
                trl_urb3d, tbl_urb3d, tgl_urb3d, &
                xxxr_urb2d, xxxb_urb2d, xxxg_urb2d, xxxc_urb2d,                               &
                frc_urb2d, utype_urb2d,                                                       &
#endif
       stc, smc, sh2o, qfx, snodep, weasd, fx, etp, skintemp, eta, etakin, cmc, grdflx, &
       ch, runoff1, runoff2, runoff3, edir, ec, ett, rc, hfx, smcmax, &
       smcref, smcwlt, albed, acrain, acsnom, esnow2d, drip2d, dewfall, soilmx, &
       fpar, znt)

    implicit none

    character(len=*),                intent(in) :: version
    integer,                         intent(in) :: igrid
    character(len=*),                intent(in) :: llanduse
    character(len=*),                intent(in) :: olddate
    character(len=*),                intent(in) :: startdate
    integer,                         intent(in) :: ix, jx, nsoil
    real,                            intent(in) :: dx, dy
    real,                            intent(in) :: truelat1, truelat2
    integer,                         intent(in) :: mapproj
    real,                            intent(in) :: lat1, lon1, cen_lon
    integer,                         intent(in) :: iswater
    integer, dimension(ix,jx),       intent(in) :: vegtyp
#ifdef _HRLDAS_URBAN_
    integer,                         intent(in) :: ucmcall
    real,    dimension(ix,jx),       intent(in) :: ts_urb2d, tr_urb2d, tb_urb2d, tg_urb2d, tc_urb2d, qc_urb2d
    real,    dimension(ix,jx),       intent(in) :: uc_urb2d, xxxr_urb2d, xxxb_urb2d, xxxg_urb2d, xxxc_urb2d
    real,    dimension(ix,jx,nsoil), intent(in) :: trl_urb3d, tbl_urb3d, tgl_urb3d
    real,    dimension(ix,jx),       intent(in) :: frc_urb2d
    integer, dimension(ix,jx),       intent(in) :: utype_urb2d
#endif
    real,    dimension(ix,jx,nsoil), intent(in) :: stc, smc, sh2o
    real,    dimension(ix,jx),       intent(in) :: qfx, snodep, weasd, fx, etp, skintemp, eta, etakin
    real,    dimension(ix,jx),       intent(in) :: cmc, grdflx, ch, runoff1, runoff2, runoff3
    real,    dimension(ix,jx),       intent(in) :: edir, ec, ett, rc, hfx, smcmax, smcref
    real,    dimension(ix,jx),       intent(in) :: smcwlt, albed, acrain, acsnom, esnow2d, drip2d
    real,    dimension(ix,jx),       intent(in) :: dewfall, soilmx, fpar, znt
    integer, parameter :: iunit = 156
    character(len=1) :: hgrid
    integer :: ncid
    character(len=256) :: output_flnm
    integer :: ierr
    integer :: varid
    integer :: dimid_times, dimid_datelen, dimid_ix, dimid_jx, dimid_dum, dimid_layers
    character(len=19) :: date19

    write(output_flnm, '("RESTART.",A10,"_DOMAIN",I1)') olddate(1:4)//olddate(6:7)//olddate(9:10)//olddate(12:13), igrid
    print*, 'output_flnm = "'//trim(output_flnm)//'"'
    ierr = nf90_create(trim(output_flnm), NF90_CLOBBER, ncid)
    if (ierr /= 0) stop "Problem nf_create"

    ierr = nf90_def_dim(ncid, "Time", NF90_UNLIMITED, dimid_times)
    ierr = nf90_def_dim(ncid, "DateStrLen", 19, dimid_datelen)
    ierr = nf90_def_dim(ncid, "west_east", ix, dimid_ix)
    ierr = nf90_def_dim(ncid, "south_north", jx, dimid_jx)
    ierr = nf90_def_dim(ncid, "west_east_stag", ix+1, dimid_dum)
    ierr = nf90_def_dim(ncid, "south_north_stag", jx+1, dimid_dum)
    ierr = nf90_def_dim(ncid, "soil_layers_stag", nsoil, dimid_layers)

    ierr = nf90_put_att(ncid, NF90_GLOBAL, "TITLE", "RESTART FILE FROM HRLDAS "//version)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "missing_value", -1.E33)

    date19(1:19) = "0000-00-00_00:00:00"
    date19(1:len_trim(startdate)) = startdate

    ierr = nf90_put_att(ncid, NF90_GLOBAL, "START_DATE", date19)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "MAP_PROJ", mapproj)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "LAT1", lat1)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "LON1", lon1)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "DX", dx)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "DY", dy)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "TRUELAT1", truelat1)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "TRUELAT2", truelat2)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "STAND_LON", cen_lon)
    ierr = nf90_put_att(ncid, NF90_GLOBAL, "MMINLU", llanduse)

!
! Done with dimensions and global attributes.
! Now define and describe all our NetCDF restart variables.
!

    ierr = nf90_def_var(ncid,  "Times",  NF90_CHAR, (/dimid_datelen,dimid_times/), varid)

!
!      _PARGS_ is the result of laziness.  
!      I simply don't want to keep repeating these four arguments in all of the subroutine calls below.
!
#define _PARGS_ ncid,dimid_ix,dimid_jx,dimid_times

    call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "SOIL_T"  , "soil temperature",     "K")
    call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "SOIL_M"  , "volumetric soil moisture", "m3 m-3")
    call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "SOIL_W"  , "liquid volumetric soil moisture", "m3 m-3")

    call make_var_att_2d(_PARGS_, NF90_FLOAT, "QFX"     , "Upward surface latent heat flux", "W m-2")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "SNODEP"  , "Snow depth", "m")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "WEASD"   , "Water equivalent accumulated snow depth", "m")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "FDOWN"   , "Radiation forcing at the surface", "W m-2")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ETP"     , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "SKINTEMP", "Skin temperature"                   , "K"                  )
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ETA"     , "Evapotranspiration??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ETAKIN"  , "Evapotranspiration", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "CANWAT"  , "Canopy water content"               , "mm"                 )
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "GRDFLX"  , "Ground heat flux at surface", "W m-2")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "CH"      , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "RUNOFF1" , "Runoff 1", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "RUNOFF2" , "Runoff 2", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "RUNOFF3" , "Runoff 3", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "EDIR"    , "Accumulated direct soil evaporation", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "EC"      , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ETT"     , "Accumulated plant transpiration", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "RC"      , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "HFX"     , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "SMCMAX"  , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "SMCREF"  , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "SMCWLT"  , "??", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ALBED"   , "Albedo -- What kind? (I.e., including what effects?)", "fraction")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ACRAIN"  , "Accumulated precipitation", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ACSNOM"  , "Accumulated snow melt", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ESNOW"   , "Accumulated evaporation of snow", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "DRIP"    , "Accumulated canopy drip", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "DEWFALL" , "Accumulated dewfall", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "SOIL_MX" , "total column soil moisture", "mm")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "FPAR"    , "Fraction of Photosythentically Active Radiation", "??")
    call make_var_att_2d(_PARGS_, NF90_FLOAT, "ZNT"     , "??", "??")

#ifdef _HRLDAS_URBAN_
    if (ucmcall == 1) then
       call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "TRL_URB3D"  , "??",     "K")
       call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "TBL_URB3D"  , "??",     "K")
       call make_var_att_3d(_PARGS_, NF90_FLOAT, dimid_layers, "TGL_URB3D"  , "??",     "K")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "TC_URB2D"   , "??",     "K")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "QC_URB2D"   , "??",     "?")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "UC_URB2D"   , "??",     "?")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "TS_URB2D"   , "??",     "K")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "TR_URB2D"   , "??",     "K")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "TB_URB2D"   , "??",     "K")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "TG_URB2D"   , "??",     "K")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "XXXR_URB2D" , "??",     "?")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "XXXB_URB2D" , "??",     "?")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "XXXG_URB2D" , "??",     "?")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "XXXC_URB2D" , "??",     "?")
       call make_var_att_2d(_PARGS_, NF90_FLOAT, "FRC_URB2D"  , "??",     "?") 
       call make_var_att_2d(_PARGS_, NF90_INT, "UTYPE_URB2D", "??",     "?")
    endif
#endif

#undef _PARGS_

    ierr = nf90_enddef(ncid)

!
! Done defining and describing all our NetCDF restart variables.
! Now actually put the data for each variable into the NetCDF file.
!

    date19(1:19) = "0000-00-00_00:00:00"
    date19(1:len_trim(olddate)) = olddate
    ierr = nf90_inq_varid(ncid, "Times", varid)
    if (ierr /= 0) then
       print*, 'nf90_inq_varid: ', trim(nf90_strerror(ierr))
       stop
    endif
    ierr = nf90_put_var(ncid, varid, olddate, (/1,1/), (/19,1/))
    if (ierr /= 0) then
       print*, 'nf90_put_var:', trim(nf90_strerror(ierr))
       stop
    endif

!
!   _PARGS_ is the result of laziness.  
!   I simply don't want to keep repeating these five arguments in all of the subroutine calls below.
!
#define _PARGS_ ncid,1,vegtyp,iswater,ix,jx
    call put_var_3d(_PARGS_, nsoil, "SOIL_T", stc)
    call put_var_3d(_PARGS_, nsoil, "SOIL_M", smc)
    call put_var_3d(_PARGS_, nsoil, "SOIL_W", sh2o)

    call put_var_2d(_PARGS_, "QFX"      , qfx     )
    call put_var_2d(_PARGS_, "SNODEP"   , snodep  )
    call put_var_2d(_PARGS_, "WEASD"    , weasd   )
    call put_var_2d(_PARGS_, "FDOWN"    , fx      )
    call put_var_2d(_PARGS_, "ETP"      , etp     )
    call put_var_2d(_PARGS_, "SKINTEMP" , skintemp)
    call put_var_2d(_PARGS_, "ETA"      , eta     )
    call put_var_2d(_PARGS_, "ETAKIN"   , etakin  )
    call put_var_2d(_PARGS_, "CANWAT"   , cmc     )
    call put_var_2d(_PARGS_, "GRDFLX"   , grdflx  )
    call put_var_2d(_PARGS_, "CH"       , ch      )
    call put_var_2d(_PARGS_, "RUNOFF1"  , runoff1 )
    call put_var_2d(_PARGS_, "RUNOFF2"  , runoff2 )
    call put_var_2d(_PARGS_, "RUNOFF3"  , runoff3 )
    call put_var_2d(_PARGS_, "EDIR"     , edir    )
    call put_var_2d(_PARGS_, "EC"       , ec      )
    call put_var_2d(_PARGS_, "ETT"      , ett     )
    call put_var_2d(_PARGS_, "RC"       , rc      )
    call put_var_2d(_PARGS_, "HFX"      , hfx     )
    call put_var_2d(_PARGS_, "SMCMAX"   , smcmax  )
    call put_var_2d(_PARGS_, "SMCREF"   , smcref  )
    call put_var_2d(_PARGS_, "SMCWLT"   , smcwlt  )
    call put_var_2d(_PARGS_, "ALBED"    , albed   )
    call put_var_2d(_PARGS_, "ACRAIN"   , acrain  )
    call put_var_2d(_PARGS_, "ACSNOM"   , acsnom  )
    call put_var_2d(_PARGS_, "ESNOW"    , esnow2d )
    call put_var_2d(_PARGS_, "DRIP"     , drip2d  )
    call put_var_2d(_PARGS_, "DEWFALL"  , dewfall )
    call put_var_2d(_PARGS_, "SOIL_MX"  , soilmx  )
    call put_var_2d(_PARGS_, "FPAR"     , fpar    )
    call put_var_2d(_PARGS_, "ZNT"      , znt     )
#ifdef _HRLDAS_URBAN_
    if (ucmcall == 1) then
       call put_var_2d(_PARGS_, "TS_URB2D"    ,   ts_urb2d   )
       call put_var_2d(_PARGS_, "TR_URB2D"    ,   tr_urb2d   )
       call put_var_2d(_PARGS_, "TB_URB2D"    ,   tb_urb2d   )
       call put_var_2d(_PARGS_, "TG_URB2D"    ,   tg_urb2d   )
       call put_var_3d(_PARGS_, nsoil, "TRL_URB3D"    ,   trl_urb3d   )
       call put_var_3d(_PARGS_, nsoil, "TBL_URB3D"    ,   tbl_urb3d   )
       call put_var_3d(_PARGS_, nsoil, "TGL_URB3D"    ,   tgl_urb3d   )
       call put_var_2d(_PARGS_, "TC_URB2D"    ,   tc_urb2d   )
       call put_var_2d(_PARGS_, "QC_URB2D"    ,   qc_urb2d   )
       call put_var_2d(_PARGS_, "UC_URB2D"    ,   uc_urb2d   )
       call put_var_2d(_PARGS_, "XXXR_URB2D"  ,  xxxr_urb2d  )
       call put_var_2d(_PARGS_, "XXXB_URB2D"  ,  xxxb_urb2d  )
       call put_var_2d(_PARGS_, "XXXG_URB2D"  ,  xxxg_urb2d  )
       call put_var_2d(_PARGS_, "XXXC_URB2D"  ,  xxxc_urb2d  )
       call put_var_2d(_PARGS_, "FRC_URB2D"   ,  frc_urb2d   )
       call put_var_int(_PARGS_, "UTYPE_URB2D" , utype_urb2d )
    endif
#endif
#undef _PARGS_

    ierr = nf90_close(ncid)
    call error_handler(ierr, "WRITE_RESTART:  nf90_close")


  end subroutine write_restart

!-----------------------------------------------------------------------------------------
!-----------------------------------------------------------------------------------------

  subroutine add_to_restart(olddate, igrid, iswater, ix, jx, vegtyp, name, units, description, array)
    implicit none
    character(len=*),                  intent(in) :: olddate
    integer,                           intent(in) :: igrid
    integer,                           intent(in) :: iswater
    integer,                           intent(in) :: ix
    integer,                           intent(in) :: jx
    integer,         dimension(ix,jx), intent(in) :: vegtyp
    character(len=*),                  intent(in) :: name
    character(len=*),                  intent(in) :: units
    character(len=*),                  intent(in) :: description
    real,            dimension(ix,jx), intent(in) :: array

    character(len=256) :: output_flnm
    integer :: ncid
    integer :: ierr
    integer :: dimid_ix
    integer :: dimid_jx
    integer :: dimid_times

    write(output_flnm, '("RESTART.",A10,"_DOMAIN",I1)') olddate(1:4)//olddate(6:7)//olddate(9:10)//olddate(12:13), igrid
    ierr = nf90_open(trim(output_flnm), NF90_WRITE, ncid)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_open")

    ierr = nf90_inq_dimid(ncid, "west_east", dimid_ix)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_inq_dimid for 'west_east'")

    ierr = nf90_inq_dimid(ncid, "south_north", dimid_jx)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_inq_dimid for 'south_north'")

    ierr = nf90_inq_dimid(ncid, "Time", dimid_times)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_inq_dimid for 'Time'")

    ierr = nf90_redef(ncid)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_redef")

    call make_var_att_2d(ncid, dimid_ix, dimid_jx, dimid_times, NF90_FLOAT, name, description, units)

    ierr = nf90_enddef(ncid)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_enddef")

    call put_var_2d(ncid, 1, vegtyp, iswater, ix, jx, name, array)

    ierr = nf90_close(ncid)
    call error_handler(ierr, "ADD_TO_RESTART:  nf90_close")

  end subroutine add_to_restart

!-----------------------------------------------------------------------------------------
!-----------------------------------------------------------------------------------------

  subroutine read_restart(igrid, olddate, ix, jx, nsoil, &
#ifdef _HRLDAS_URBAN_
    ucmcall,                                                                      &
         ts_urb2d, tr_urb2d, tb_urb2d, tg_urb2d, tc_urb2d, qc_urb2d, uc_urb2d,         &
         trl_urb3d, tbl_urb3d, tgl_urb3d,                                              &
         xxxr_urb2d, xxxb_urb2d, xxxg_urb2d, xxxc_urb2d,                               &
         frc_urb2d, utype_urb2d,                                                       &
#endif
    stc, smc, sh2o, qfx, snodep, weasd, fx, etp, t1, etax, etakin, cmc, grdflx, &
         ch, runoff1, runoff2, runoff3, edirx, ecx, ettx, rcx, hfx, smcmax1,        &
         smcref1, smcwlt1, albedx, acrain, acsnom, esnow2d, drip2d, dewfall, soilmx)

    implicit none

    integer,                      intent(in)  :: igrid
    character(len=*),             intent(in)  :: olddate
    integer,                      intent(in)  :: ix, jx, nsoil
#ifdef _HRLDAS_URBAN_
    integer,                      intent(in)  :: ucmcall
    real, dimension(ix,jx),       intent(out) :: ts_urb2d, tr_urb2d, tb_urb2d, tg_urb2d, tc_urb2d, qc_urb2d
    real, dimension(ix,jx),       intent(out) :: uc_urb2d, xxxr_urb2d, xxxb_urb2d, xxxg_urb2d, xxxc_urb2d
    real, dimension(ix,jx,nsoil), intent(out) :: trl_urb3d, tbl_urb3d, tgl_urb3d
    real, dimension(ix,jx),       intent(out) :: frc_urb2d
    integer, dimension(ix,jx),    intent(out) :: utype_urb2d
#endif
    real, dimension(ix,jx,nsoil), intent(out) :: stc, smc, sh2o
    real, dimension(ix,jx),       intent(out) :: qfx, snodep, weasd, fx, etp, t1, etax, etakin
    real, dimension(ix,jx),       intent(out) :: cmc, grdflx, ch, runoff1, runoff2, runoff3
    real, dimension(ix,jx),       intent(out) :: edirx, ecx, ettx, rcx, hfx, smcmax1, smcref1
    real, dimension(ix,jx),       intent(out) :: smcwlt1, albedx, acrain, acsnom, esnow2d, drip2d
    real, dimension(ix,jx),       intent(out) :: dewfall, soilmx

    integer, parameter :: iunit = 156
    character(len=1) :: hgrid
    character(len=256) :: restart_filename
    integer :: ierr
    integer :: ncid
    integer :: varid
    character(len=256) :: titlestr
    integer :: restart_version
    integer :: idx

    write(hgrid,'(I1)') igrid
    restart_filename = "RESTART."//olddate(1:4)//olddate(6:7)//olddate(9:10)//olddate(12:13)//"_DOMAIN"//trim(hgrid)

    ierr = nf90_open(trim(restart_filename), NF90_NOWRITE, ncid)

    if (ierr == NF90_ENOTNC) then

    else
       if (ierr /= NF90_NOERR) then
          write(*,*)
          write(*,'(" ***** Restart problem ***************************************")')
          write(*,'(" ***** ")')
          write(*,'(" *****        You have requested a restart from date ", A)') olddate(1:13)
          write(*,'(" *****        but there was a problem in accessing the file ''", A, "''")') trim(restart_filename)
          write(*,'(" ***** ")')
       endif
       call error_handler(ierr, " trying to open restart file "//restart_filename)

       ierr = nf90_get_att(ncid, NF90_GLOBAL, "TITLE", titlestr)
       if (ierr /= 0) then
          write(*,'("WARNING:  RESTART file does not have TITLE attribute.")')
          write(*,'("          This probably means that LDASIN files are from an older release,")')
          write(*,'("          And are very likely incompatible with the current code.")')
          write(*,'("          I assume you know what you are doing.")')
          restart_version = 0
       else
          write(*,'("RESTART TITLE attribute: ", A)') trim(titlestr)
          ! Pull out the version number, assuming that the version is identified by vYYYYMMDD, and 
          ! based on a search for the string "v20".
          idx = index(trim(titlestr), "v20")
          if (idx <= 0) then
             write(*,'("FATAL:  RESTART file has a perverse version identifier")')
             !  write(*,'("          I assume you know what you are doing.")')
             stop
          else
             read(titlestr(idx+1:), '(I8)', iostat=ierr) restart_version
             if (ierr /= 0) then
                write(*,'("FATAL:  RESTART file has a perverse version identifier")')
                !  write(*,'("          I assume you know what you are doing.")')
                stop
             endif
          endif
       endif
       write(*, '("ldasin_version = ", I8)') restart_version

! Older RESTART files are incompatible if you use the Urban Canopy Model.
#if defined _HRLDAS_URBAN_
       if (ucmcall == 1) then
          if (restart_version < 20070712) then
             write(*,'(/,"  *****************************************************************************  ")')
             write(*,'("  *****  You are using a restart file created prior to version v20070712.")')
             write(*,'("  *****  That version of restart files has problems with Urban Canopy Model.")')
             write(*,'("  *****  Update your code and use newer restart files.")')
             write(*,'("  *****************************************************************************  ",/)')
             stop
          endif
       endif
#endif


       ierr = nf90_inq_varid(ncid, "SOIL_T", varid)
       call error_handler(ierr, "Problem finding SOIL_T varid")
       ierr = nf90_get_var(ncid, varid, stc)
       call error_handler(ierr, "Problem finding SOIL_T")

       ierr = nf90_inq_varid(ncid, "SOIL_M", varid)
       call error_handler(ierr, "Problem finding SOIL_M varid")
       ierr = nf90_get_var(ncid, varid, smc)
       call error_handler(ierr, "Problem finding SOIL_M")

       ierr = nf90_inq_varid(ncid, "SOIL_W", varid)
       call error_handler(ierr, "Problem finding SOIL_W varid")
       ierr = nf90_get_var(ncid, varid, sh2o)
       call error_handler(ierr, "Problem finding SOIL_W")

       ierr = nf90_inq_varid(ncid, "QFX", varid)
       call error_handler(ierr, "Problem finding QFX varid")
       ierr = nf90_get_var(ncid, varid, qfx)
       call error_handler(ierr, "Problem finding QFX")

       ierr = nf90_inq_varid(ncid, "SNODEP", varid)
       call error_handler(ierr, "Problem finding SNODEP varid")
       ierr = nf90_get_var(ncid, varid, snodep)
       call error_handler(ierr, "Problem finding SNODEP")

       ierr = nf90_inq_varid(ncid, "WEASD", varid)
       call error_handler(ierr, "Problem finding WEASD varid")
       ierr = nf90_get_var(ncid, varid, weasd)
       call error_handler(ierr, "Problem finding WEASD")

       ierr = nf90_inq_varid(ncid, "FDOWN", varid)
       call error_handler(ierr, "Problem finding FDOWN varid")
       ierr = nf90_get_var(ncid, varid, fx)
       call error_handler(ierr, "Problem finding FDOWN")

       ierr = nf90_inq_varid(ncid, "ETP", varid)
       call error_handler(ierr, "Problem finding ETP varid")
       ierr = nf90_get_var(ncid, varid, etp)
       call error_handler(ierr, "Problem finding ETP")

       ierr = nf90_inq_varid(ncid, "SKINTEMP", varid)
       call error_handler(ierr, "Problem finding SKINTEMP varid")
       ierr = nf90_get_var(ncid, varid, t1)
       call error_handler(ierr, "Problem finding SKINTEMP")

       ierr = nf90_inq_varid(ncid, "ETA", varid)
       call error_handler(ierr, "Problem finding ETA varid")
       ierr = nf90_get_var(ncid, varid, etax)
       call error_handler(ierr, "Problem finding ETA")

       ierr = nf90_inq_varid(ncid, "ETAKIN", varid)
       call error_handler(ierr, "Problem finding ETAKIN varid")
       ierr = nf90_get_var(ncid, varid, etakin)
       call error_handler(ierr, "Problem finding ETAKIN")

       ierr = nf90_inq_varid(ncid, "CANWAT", varid)
       call error_handler(ierr, "Problem finding CMC varid")
       ierr = nf90_get_var(ncid, varid, cmc)
       call error_handler(ierr, "Problem finding CANWAT")

       ierr = nf90_inq_varid(ncid, "GRDFLX", varid)
       call error_handler(ierr, "Problem finding GRDFLX varid")
       ierr = nf90_get_var(ncid, varid, grdflx)
       call error_handler(ierr, "Problem finding GRDFLX")

       ierr = nf90_inq_varid(ncid, "CH", varid)
       call error_handler(ierr, "Problem finding CH varid")
       ierr = nf90_get_var(ncid, varid, CH)
       call error_handler(ierr, "Problem finding CH")

       ierr = nf90_inq_varid(ncid, "RUNOFF1", varid)
       call error_handler(ierr, "Problem finding RUNOFF1 varid")
       ierr = nf90_get_var(ncid, varid, runoff1)
       call error_handler(ierr, "Problem finding RUNOFF1")

       ierr = nf90_inq_varid(ncid, "RUNOFF2", varid)
       call error_handler(ierr, "Problem finding RUNOFF2 varid")
       ierr = nf90_get_var(ncid, varid, runoff2)
       call error_handler(ierr, "Problem finding RUNOFF2")

       ierr = nf90_inq_varid(ncid, "RUNOFF3", varid)
       call error_handler(ierr, "Problem finding RUNOFF3 varid")
       ierr = nf90_get_var(ncid, varid, runoff3)
       call error_handler(ierr, "Problem finding RUNOFF3")

       ierr = nf90_inq_varid(ncid, "EDIR", varid)
       call error_handler(ierr, "Problem finding EDIR varid")
       ierr = nf90_get_var(ncid, varid, edirx)
       call error_handler(ierr, "Problem finding EDIR")

       ierr = nf90_inq_varid(ncid, "EC", varid)
       call error_handler(ierr, "Problem finding EC varid")
       ierr = nf90_get_var(ncid, varid, ecx)
       call error_handler(ierr, "Problem finding EC")

       ierr = nf90_inq_varid(ncid, "ETT", varid)
       call error_handler(ierr, "Problem finding ETT varid")
       ierr = nf90_get_var(ncid, varid, ettx)
       call error_handler(ierr, "Problem finding ETT")

       ierr = nf90_inq_varid(ncid, "RC", varid)
       call error_handler(ierr, "Problem finding RC varid")
       ierr = nf90_get_var(ncid, varid, rcx)
       call error_handler(ierr, "Problem finding RC")

       ierr = nf90_inq_varid(ncid, "HFX", varid)
       call error_handler(ierr, "Problem finding HFX varid")
       ierr = nf90_get_var(ncid, varid, hfx)
       call error_handler(ierr, "Problem finding HFX")

       ierr = nf90_inq_varid(ncid, "SMCMAX", varid)
       call error_handler(ierr, "Problem finding SMCMAX varid")
       ierr = nf90_get_var(ncid, varid, smcmax1)
       call error_handler(ierr, "Problem finding SMCMAX1")

       ierr = nf90_inq_varid(ncid, "SMCREF", varid)
       call error_handler(ierr, "Problem finding SMCREF varid")
       ierr = nf90_get_var(ncid, varid, smcref1)
       call error_handler(ierr, "Problem finding SMCREF")

       ierr = nf90_inq_varid(ncid, "SMCWLT", varid)
       call error_handler(ierr, "Problem finding SMCWLT varid")
       ierr = nf90_get_var(ncid, varid, smcwlt1)
       call error_handler(ierr, "Problem finding SMCWLT")

       ierr = nf90_inq_varid(ncid, "ALBED", varid)
       call error_handler(ierr, "Problem finding ALBED varid")
       ierr = nf90_get_var(ncid, varid, albedx)
       call error_handler(ierr, "Problem finding ALBED")

       ierr = nf90_inq_varid(ncid, "ACRAIN", varid)
       call error_handler(ierr, "Problem finding ACRAIN varid")
       ierr = nf90_get_var(ncid, varid, acrain)
       call error_handler(ierr, "Problem finding ACRAIN")

       ierr = nf90_inq_varid(ncid, "ACSNOM", varid)
       call error_handler(ierr, "Problem finding ACSNOM varid")
       ierr = nf90_get_var(ncid, varid, acsnom)
       call error_handler(ierr, "Problem finding ACSNOM")

       ierr = nf90_inq_varid(ncid, "ESNOW", varid)
       call error_handler(ierr, "Problem finding ESNOW varid")
       ierr = nf90_get_var(ncid, varid, esnow2d)
       call error_handler(ierr, "Problem finding ESNOW")

       ierr = nf90_inq_varid(ncid, "DRIP", varid)
       call error_handler(ierr, "Problem finding DRIP varid")
       ierr = nf90_get_var(ncid, varid, drip2d)
       call error_handler(ierr, "Problem finding DRIP")

       ierr = nf90_inq_varid(ncid, "DEWFALL", varid)
       call error_handler(ierr, "Problem finding DEWFALL varid")
       ierr = nf90_get_var(ncid, varid, dewfall)
       call error_handler(ierr, "Problem finding DEWFALL")

       ierr = nf90_inq_varid(ncid, "SOIL_MX", varid)
       call error_handler(ierr, "Problem finding SOIL_MX varid")
       ierr = nf90_get_var(ncid, varid, soilmx)
       call error_handler(ierr, "Problem finding SOIL_MX")

#if defined _HRLDAS_URBAN_
       if (ucmcall == 1) then
          ierr = nf90_inq_varid(ncid, "TS_URB2D", varid)
          call error_handler(ierr, "Problem finding TS_URB2D varid", "Found TS_URB2D varid")
          ierr = nf90_get_var(ncid, varid, ts_urb2d)
          call error_handler(ierr, "Problem finding TS_URB2D", "Found TS_URB2D")

          ierr = nf90_inq_varid(ncid, "TR_URB2D", varid)
          call error_handler(ierr, "Problem finding TR_URB2D varid", "Found TR_URB2D varid")
          ierr = nf90_get_var(ncid, varid, tr_urb2d)
          call error_handler(ierr, "Problem finding TR_URB2D", "Found TR_URB2D")

          ierr = nf90_inq_varid(ncid, "TB_URB2D", varid)
          call error_handler(ierr, "Problem finding TB_URB2D varid", "Found TB_URB2D varid")
          ierr = nf90_get_var(ncid, varid, tb_urb2d)
          call error_handler(ierr, "Problem finding TB_URB2D", "Found TB_URB2D")

          ierr = nf90_inq_varid(ncid, "TG_URB2D", varid)
          call error_handler(ierr, "Problem finding TG_URB2D varid", "Found TG_URB2D varid")
          ierr = nf90_get_var(ncid, varid, tg_urb2d)
          call error_handler(ierr, "Problem finding TG_URB2D", "Found TG_URB2D")

          ierr = nf90_inq_varid(ncid, "TC_URB2D", varid)
          call error_handler(ierr, "Problem finding TC_URB2D varid", "Found TC_URB2D varid")
          ierr = nf90_get_var(ncid, varid, tc_urb2d)
          call error_handler(ierr, "Problem finding TC_URB2D", "Found TC_URB2D")

          ierr = nf90_inq_varid(ncid, "QC_URB2D", varid)
          call error_handler(ierr, "Problem finding QC_URB2D varid", "Found QC_URB2D varid")
          ierr = nf90_get_var(ncid, varid, qc_urb2d)
          call error_handler(ierr, "Problem finding QC_URB2D", "Found QC_URB2D")

          ierr = nf90_inq_varid(ncid, "UC_URB2D", varid)
          call error_handler(ierr, "Problem finding UC_URB2D varid", "Found UC_URB2D varid")
          ierr = nf90_get_var(ncid, varid, uc_urb2d)
          call error_handler(ierr, "Problem finding UC_URB2D", "Found UC_URB2D")

          ierr = nf90_inq_varid(ncid, "TRL_URB3D", varid)
          call error_handler(ierr, "Problem finding TRL_URB3D varid", "Found TRL_URB3D varid")
          ierr = nf90_get_var(ncid, varid, trl_urb3d)
          call error_handler(ierr, "Problem finding TRL_URB3D", "Found TRL_URB3D")

          ierr = nf90_inq_varid(ncid, "TBL_URB3D", varid)
          call error_handler(ierr, "Problem finding TBL_URB3D varid", "Found TBL_URB3D varid")
          ierr = nf90_get_var(ncid, varid, tbl_urb3d)
          call error_handler(ierr, "Problem finding TBL_URB3D", "Found TBL_URB3D")

          ierr = nf90_inq_varid(ncid, "TGL_URB3D", varid)
          call error_handler(ierr, "Problem finding TGL_URB3D varid", "Found TGL_URB3D varid")
          ierr = nf90_get_var(ncid, varid, tgl_urb3d)
          call error_handler(ierr, "Problem finding TGL_URB3D", "Found TGL_URB3D")

          ierr = nf90_inq_varid(ncid, "XXXR_URB2D", varid)
          call error_handler(ierr, "Problem finding XXXR_URB2D varid", "Found XXXR_URB2D varid")
          ierr = nf90_get_var(ncid, varid, xxxr_urb2d)
          call error_handler(ierr, "Problem finding XXXR_URB2D", "Found XXXR_URB2D")

          ierr = nf90_inq_varid(ncid, "XXXB_URB2D", varid)
          call error_handler(ierr, "Problem finding XXXB_URB2D varid", "Found XXXB_URB2D varid")
          ierr = nf90_get_var(ncid, varid, xxxb_urb2d)
          call error_handler(ierr, "Problem finding XXXB_URB2D", "Found XXXB_URB2D")

          ierr = nf90_inq_varid(ncid, "XXXG_URB2D", varid)
          call error_handler(ierr, "Problem finding XXXG_URB2D varid", "Found XXXG_URB2D varid")
          ierr = nf90_get_var(ncid, varid, xxxg_urb2d)
          call error_handler(ierr, "Problem finding XXXG_URB2D", "Found XXXG_URB2D")

          ierr = nf90_inq_varid(ncid, "XXXC_URB2D", varid)
          call error_handler(ierr, "Problem finding XXXC_URB2D varid", "Found XXXC_URB2D varid")
          ierr = nf90_get_var(ncid, varid, xxxc_urb2d)
          call error_handler(ierr, "Problem finding XXXC_URB2D", "Found XXXC_URB2D")

          ierr = nf90_inq_varid(ncid, "FRC_URB2D", varid)
          call error_handler(ierr, "Problem finding FRC_URB2D varid", "Found FRC_URB2D varid")
          ierr = nf90_get_var(ncid, varid, frc_urb2d)
          call error_handler(ierr, "Problem finding FRC_URB2D", "Found FRC_URB2D")

          ierr = nf90_inq_varid(ncid, "UTYPE_URB2D", varid)
          call error_handler(ierr, "Problem finding UTYPE_URB2D varid", "Found UTYPE_URB2D varid")
          ierr = nf90_get_var(ncid, varid, utype_urb2d)
          call error_handler(ierr, "Problem finding UTYPE_URB2D", "Found UTYPE_URB2D")

       endif
#endif
       ierr = nf90_close(ncid)
       call error_handler(ierr, "Problem closing restart file")
    endif

  end subroutine read_restart

!-----------------------------------------------------------------------------------------
!-----------------------------------------------------------------------------------------

  subroutine get_from_restart(olddate, igrid, ix, jx, name, array, return_error)
    implicit none
    character(len=*),                   intent(in) :: olddate
    integer,                            intent(in) :: igrid
    integer,                            intent(in) :: ix
    integer,                            intent(in) :: jx
    character(len=*),                   intent(in)  :: name
    real,             dimension(ix,jx), intent(out) :: array
    integer,          optional,         intent(out) :: return_error

    character(len=256) :: restart_filename
    integer :: ierr
    integer :: ncid
    integer :: varid

    write(restart_filename, '("RESTART.",A10,"_DOMAIN",I1)') olddate(1:4)//olddate(6:7)//olddate(9:10)//olddate(12:13), igrid

    ierr = nf90_open(trim(restart_filename), NF90_NOWRITE, ncid)
    call error_handler(ierr, "Problem opening restart file '"//trim(restart_filename)//"'")

    if (present(return_error)) then
       ierr = nf90_inq_varid(ncid, name, varid)
       if (ierr == NF90_NOERR) then
          return_error = 0
          call error_handler(ierr, "Problem finding variable in restart file '"//trim(name)//"'")

          ierr = nf90_get_var(ncid, varid, array)
          call error_handler(ierr, "Problem finding variable in restart file: '"//trim(name)//"'")
       else
          return_error = 1
          write(*,'("Did not find optional variable ''",A,"'' in restart file ''", A, "''")') trim(name), trim(restart_filename)
       endif
    else
       ierr = nf90_inq_varid(ncid, name, varid)
       call error_handler(ierr, "Problem finding required variable in restart file: '"//trim(name)//"'")

       ierr = nf90_get_var(ncid, varid, array)
       call error_handler(ierr, "Problem finding variable in restart file: '"//trim(name)//"'")
    endif

    ierr = nf90_close(ncid)
    call error_handler(ierr, "Problem closing restart file")
    
  end subroutine get_from_restart

!-----------------------------------------------------------------------------------------
!-----------------------------------------------------------------------------------------

  subroutine error_handler(status, failure, success)
    !
    ! Check the error flag from a NetCDF function call, and print appropriate
    ! error message.
    !
    implicit none
    integer,                    intent(in) :: status
    character(len=*), optional, intent(in) :: failure
    character(len=*), optional, intent(in) :: success

    if (status .ne. NF90_NOERR) then
       write(*,'(/,A)') nf90_strerror(status)
       if (present(failure)) then
          write(*,'(/," ***** ", A,/)') failure
       endif
       stop 'Stopped'
    endif

    if (present(success)) then
       write(*,'(A)') success
    endif

  end subroutine error_handler


  subroutine read_additional(flnm_template, hdate, name, xstart, xend, ystart, yend, array, ierr)
    use kwm_string_utilities
    implicit none
    character(len=*),                         intent(in)  :: flnm_template
    character(len=*),                         intent(in)  :: hdate
    character(len=*),                         intent(in)  :: name
    integer,                                  intent(in)  :: xstart
    integer,                                  intent(in)  :: xend
    integer,                                  intent(in)  :: ystart
    integer,                                  intent(in)  :: yend
    real, dimension(xstart:xend,ystart:yend), intent(out) :: array
    integer,                                  intent(out) :: ierr

    character(len=256) :: flnm
    integer :: jday
    character(len=3) :: hjday
    integer :: ncid
    integer :: varid
    logical :: lexist

    call geth_idts(hdate(1:10), hdate(1:4)//"-01-01", jday)
    jday = jday + 1
    write(hjday,'(I3.3)') jday

    flnm = flnm_template

    call strrep(flnm, "<YYYY>", hdate(1:4))
    call strrep(flnm, "<MM>", hdate(6:7))
    call strrep(flnm, "<DD>", hdate(9:10))
    call strrep(flnm, "<HH>", hdate(12:13))
    call strrep(flnm, "<JDAY>", hjday)

    inquire(file=trim(flnm), exist=lexist)
    if (.not. lexist) then
       ierr = 1
       return
    endif

    write(*, '("Additional flnm = ''",A,"''")') trim(flnm)

    ierr = nf90_open(trim(flnm), NF90_NOWRITE, ncid)
    call error_handler(ierr, failure="READ_ADDITIONAL: Problem opening additional file: "//trim(flnm))

    ierr = nf90_inq_varid(ncid,  name,  varid)
    call error_handler(ierr, failure="READ_ADDITIONAL: Problem finding variable: "//name)

    ierr = nf90_get_var(ncid, varid, array, start=(/xstart,ystart/), count=(/xend-xstart+1,yend-ystart+1/))
    call error_handler(ierr, failure="READ_ADDITIONAL: Problem getting variable: "//name)

    ierr = nf90_close(ncid)
    call error_handler(ierr, failure="READ_ADDITIONAL:  Problem closing file:  "//trim(flnm))

  end subroutine read_additional


 end module module_hrldas_netcdf_io
